//  Time-stamp: <11 Jan 2007 11:21 GMT DNAProcess.cpp km>
/*
 * routines to read and write sequences and to call bases given a 
 * list of intensities in different colours.
 *
 * Copyright (c) Solexa 2004, 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <stdlib.h>
#include <string.h>
#include <math.h>
#else
#include <cstdlib>
#include <cstring>
#include <cmath>
#endif
#include <cstdio>
#include <cassert>
#include <vector>
#include "DNACodes.h"
#include "DNAProcess.h"
#include "Stats.h"
#include "Message.h"

//---------------------------------------------------------------------------
#define MIN(x,y) ((x) < (y) ? (x) : (y))

void write_quality_file(const char filename[], float *prob,
                        int nobj, int nprob, int ncol)
{
    FILE *probfile = fopen(filename,"wt");
    if (probfile) {
        char *buffer = new char [nprob * 20];
        int n, i;
        for (n = 0; n < nobj; n++) {
            int idx = 0;
            for (i = 0; i < nprob; i++) {
                idx += snprintf(buffer+idx, 19, "%4.0f", prob[n*nprob+i]);
                if (i == nprob-1)
                    continue;
                if (i % ncol == ncol-1)
                    idx += sprintf(buffer+idx, "\t");
                else
                    idx += sprintf(buffer+idx, " ");
            }
            fprintf(probfile, "%s\n", buffer);
        }
        delete [] buffer;
        fclose(probfile);
    }
}

int write_quality_file(const char filename[], std::vector<float * > prob,
                       int nobj, int ncol)
{
    int code = 0;
    int nprob = int(prob.size());
    FILE *probfile = fopen(filename, "wt");
    if (probfile) {
        int buffersize = nprob * ncol * 8;
        char *buffer = new char [buffersize];
        int n, i, k;
        for (n = 0; n < nobj; n++) {
            int idx = 0;
            for (i = 0; i < nprob; i++) {
                for (k = 0; k < ncol; k++) {
                    idx += snprintf(buffer+idx, 7, "%4.0f", prob[i][n*ncol+k]);
                    if (k != ncol-1)
                        idx += sprintf(buffer+idx, " ");
                }
                if (i != nprob-1)
                    idx += sprintf(buffer+idx, "\t");
            }
            sprintf(buffer+idx, "\n");
            if (fprintf(probfile, buffer) < (int)strlen(buffer)) {
                error("Failed writing to %s.", filename);
                code = 2;
            }
        }
        delete [] buffer;
        if (fclose(probfile)) {
            error("Failed to close %s.", filename);
            return 3;
        }
        return code;
    }
    else {
        error("Failed to open %s for writing.", filename);
        return 1;
    }
}

int write_sequence_file(const char filename[],
                         unsigned char *sequence_array,
                         int nobj, int ncycle,
                         int lane, int tile, st_fposition *positions)
{
    FILE *seqfile = fopen(filename,"wt");
    if (!seqfile) {
        error("Failed to open %s for writing.", filename);
        return 1;
    }
    char *buffer = new char[21+ncycle+1];
    int i, n;
    int error_code = 0;

    for (n = 0; n < nobj; n++) {
        int idx = 0;
        if (positions != NULL) {
            idx = snprintf(buffer, 20, "%i\t%i\t%.0f\t%.0f\t", 
                           lane + 1, tile + 1, 
                           int(round(positions[n].x * 100.)) / 100.,
                           int(round(positions[n].y * 100.)) / 100.); 
        }
        for (i = 0; i < ncycle; i++)
            buffer[idx+i] = sequence_array[i*nobj+n];
        sprintf(buffer+idx+ncycle, "\n");
        if (fprintf(seqfile, "%s", buffer) < int(strlen(buffer))) {
            error_code = 2;
            error("Failed to write to %s.", filename);
        };
    }
    delete [] buffer;

    if (fclose(seqfile)) {
        error("Failed to close %s.", filename);
        error_code = 3;
    }
    return error_code;
}

void read_sequence_file(const char filename[],
                        unsigned char *sequence_array,
                        int nobj, int ncycle)
{
    assert(ncycle <= 100);
    int nread = 0;
    int i;
    FILE *seqfile = fopen(filename, "r");

    for (i = 0; i < nobj*ncycle; i++) {
        sequence_array[i] = '.';
    }

    if (seqfile) {
        char buffer[200+1];
        char seqbuffer[100+1];
        while (fgets(buffer, 200, seqfile) && (nread < nobj)) {
            if (buffer[0] != '#') {
                sscanf(buffer, "%100s", seqbuffer);
                for (i = 0; i < MIN(ncycle, (int)strlen(seqbuffer)); i++)
                    sequence_array[i*nobj+nread] = seqbuffer[i];
                nread++;
            }
        }

        fclose(seqfile);
    }
}


//---------------------------------------------------------------------------

/*
 * apply matrix rescaling to each row of the intensities array,
 * i.e. to the colour channels each object has been observed in
 */
void rescale_intensities(float *intensities, float *rescaled_intensities,
                         int nobj, int ncol, float *matrix)
{
    assert(nobj >= 0);
    assert(ncol > 0);
    if (matrix != NULL) {
        int i, j, k;
        for (i = 0; i < nobj; i++) 
            for (j = 0; j < ncol; j++) {
                rescaled_intensities[i*ncol+j] = 0.;
                for (k = 0; k < ncol; k++)
                    rescaled_intensities[i*ncol+j] += 
                            matrix[j*ncol+k] * intensities[i*ncol+k];
            }
    }
    else
        memcpy(rescaled_intensities, intensities, nobj*ncol*sizeof(float));
}

/*
 * call bases for each cycle and molecule/colony in a list.
 * uses the intensities and noise rms in each colour to determine
 * the most likely sequence.
 * INPUT
 *         intensity  [nobj * ncol] array of intensities for
 *                    all molecules in the observed colours
 *         sigma      [nobj * ncol] array of the background noise
 *                    dispersions at the molecule positions
 *         nobj       number of molecules/colonies for which a base
 *                    is to be determined
 *         ncol       number of colours observed
 *         bases      [ncol] array of bases for each colour
 * RETURNS
 *         sequence   [nobj] array containing a base code for each 
 *                    molecule/colony. the array has to be allocated
 *                    by the caller      
 */
void call_bases(float *intensity, float *sigma, int nobj, int ncol,
                unsigned char *bases, unsigned char *sequence)
{
    assert(ncol > 0);
    assert(nobj >= 0);
    if (nobj == 0)
        return;
    
    int i, j;

    unsigned short *codes = new unsigned short [ncol*2];

    /* 
     * define binary codes for each cycle to help lookup of ambiguity codes
     * in arrays; see definition of codes in DNAProcess.h.
     */
    for (j = 0; j < ncol; j++) {
        codes[j] = 0; codes[ncol+j] = 0;

        for (i = 0; i < 4; i++)
            if (BaseUpperCase[i] == bases[j]) {
                codes[j] = 1<<(i*2); codes[ncol+j] = 1<<(i*2+1);
                break;
            }
    }

    /* call bases for each molecule */
    for (i = 0; i < nobj; i++) {
        unsigned short code = 0;

        float max_intensity = 0.f;
        for (j = 0; j < ncol; j++)
            if (intensity[i*ncol+j] > max_intensity) {
                max_intensity = intensity[i*ncol+j];
                code = codes[ncol+j];
            }

        sequence[i] = DNA_2bit_base(code);
    }
    delete [] codes;
} 
