/*
 * Copyright (c) Solexa   2005-2007
 *               Illumina 2007
 * Authors: Klaus Maisinger
 *
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

/*
 * The core routines of the base-caller are found in this file.
 */  

#include <vector>
#include <cassert>

#include "BaseCallConfig.h"
#include "PhasingModel.h"
#include "ProcessBaseCall.h"
#include "ProbabilityIntegration.h"
#include "DNAProcess.h"

#include "LinAlg.h"
#include "Stats.h"
#include "Tools.h"
#include "IntensityData.h"


ProcessBaseCall::ProcessBaseCall(const int ncycles_)
{ 
    ncycles = ncycles_;
}

ProcessBaseCall::~ProcessBaseCall()
{ 
}

void ProcessBaseCall::setPhasingModel(const PhasingModel &pm)
{
    phasing = pm;
}

/*
 * go through intensity values  and make sure that each of the 4 values
 * for a cluster is non-zero (hasn't shifted off the image); otherwise
 * set to zero (no base-call).
 */
void validate_signal(float *signal, float *noise, int nobj, int ncol)
{
    int k, i;
    for (k = 0; k < nobj; k++) {
        if (min(&noise[k*ncol], ncol) <= 0.) {
             for (i = 0; i < ncol; i++) {
                 signal[k*ncol+i] = 0.;
                 noise[k*ncol+i] = 0.;
            }
        }
    }
}

/*
 * go through intensity values from all cycles used for phasing correction
 * and make sure that every single one of them is set to zero (no base-call)
 * if the current cycle is zero (cluster has shifted off the image).
 */
void validate_signals(std::vector<float * > signal,
                      std::vector<float * > noise,
                      int nobj, int ncol, int index)
{
    int k, l, i;
    int nw = int(signal.size());
    for (k = 0; k < nobj; k++) {
        if (min(&noise[index][k*ncol], ncol) <= 0.)
            for (l = 0; l < nw; l++)
               for (i = 0; i < ncol; i++) {
                   signal[l][k*ncol+i] = 0.;
                   noise [l][k*ncol+i] = 0.;
               }
    }
}

/* 
 * estimate the average signal per cycle; this can be used to normalise
 * the intensities
 */
std::vector<float> estimateAverageSignals(const std::vector<float * > &signal,
                                          const std::vector<float * > &noise,
                                          int nobj, int ncol)
{
    assert(signal.size() == noise.size());
    int k, l, i;
    std::vector<int>   counts(signal.size());
    std::vector<float> average(signal.size());
    for (l = 0; l < int(counts.size()); l++) {
        average[l] = 0.;
        counts[l]  = 0;
    }
    for (k = 0; k < nobj; k++) {
        for (l = 0; l < int(counts.size()); l++)
            if (min(&noise[l][k*ncol], ncol) > 0.) {
                counts[l]  ++;
                for (i = 0; i < ncol; i++)
                    average[l] += signal[l][k*ncol+i];
            }
    }
    for (l = 0; l < int(counts.size()); l++)
        if (counts[l] > 0)
            average[l] /= counts[l];
    return average;
}   


/*
 * Calculate the noise covariance matrix based on the matrix and 
 * phasing corrections.
 */
void calculateCovariance(float *covariance, int ndim, int k,
                         const std::vector<PhasingWindow> &window,
                         const std::vector<float * > &matrices,
                         const std::vector<float * > &allnoise)
{
    int i, j, l, m, i2, j2, k2;
    k2 = k * ndim;
    initialise_array(covariance, ndim*ndim);
    for (l = 0; l < int(window.size()); l++) {
        float window2 = window[l].value * window[l].value;
        float *noise  = allnoise[l];
        float *matrix = matrices[l];
        for (i = 0; i < ndim; i++) {
            i2 = i*ndim;
            for (j = 0; j < ndim; j++) {
                j2 = j * ndim;
                for (m = 0; m < ndim; m++)
                    covariance[i*ndim+j] += window2
                                * matrix[i2+m] * matrix[j2+m] 
                                * noise[k2+m] * noise[k2+m];
            }
        }
    }
}
                         

/*
 * Estimate relative base probabilities:
 * - For each cluster, obtain the noise covariance matrix of the 
 *   matrix-corrected intensities. The intensity covariance matrix
 *   is assumed to be diagonal with elements given by the error estimates
 *   obtained from the image analysis. Then apply matrix correction 
 *   to these values to obtain the covariance of the matrix-corrected data.
 * - Integrate the likelihood over the 4-d volumes defined by the 4 possible
 *   base calls (I_A >= I_X, X=C,G,T; I_C >= I_X, X=A,G,T; etc.)
 */
void estimateProbabilities(const std::vector<float * > &allsignal,
                           const std::vector<float * > &allnoise,
                           const std::vector<float * > &matrices,
                           const std::vector<PhasingWindow> &window,
                           float *signal,
                           float *probabilities, int nobj, int ncol)
{
    int i, k;
    float *covariance = new float [ncol*ncol];

    ProbabilityIntegration *integrator 
                   = new ProbabilityIntegration(mc_samples, ncol, max_score);
    for (k = 0; k < nobj; k++) {
        /* 
         * calculate noise covariance matrix
         */
        calculateCovariance(covariance, ncol, k,
                            window, matrices, allnoise);

        /*
         * Decompose covariance matrix and integrate 4d signal space
         * to obtain relative base probabilities
         */
        float *original = new float [ncol*ncol];
        for (i = 0; i < ncol*ncol; i++)
            original[i] = covariance[i];
            if (max(covariance, ncol*ncol) > 0.)
                if (cholesky_decompose(covariance, ncol)) {
                    char msg[10000];
                    printf("Covariance\n");
                    snprint_matrix(msg, sizeof(msg), original, ncol, ncol);
                    printf("%s\n", msg); 
                }
        delete [] original;

        integrator->integrate(&signal[k*ncol], ncol,
                              covariance, &probabilities[k*ncol]);
    }
    delete [] covariance;
    delete integrator;
}


/*
 * Implementation of a phasing correction.
 * Identifies a set of cycles that the current cycle has leaked signal to,
 * calculates an inverse "window function" (a filter to reverse the effect
 * of phasing) for all cycles that contribute significantly, loads the
 * corresponding intensity files and applies the filter correction.
 */
void ProcessBaseCall::applyPhasingCorrection(std::vector<float * > intensities,
                                             std::vector<float * > noise,
                                             std::vector<float * > matrices,
                                             int nobj, int ncol, int cycle,
                                             int *is_deblocked,
                                             float *res_signal,
                                             float *corr_signal, 
                                             float *probabilities)
{
    int ncycles = int(intensities.size());
    assert(ncycles > cycle);

    /* 
     * Construct the "phasing correction", a function that tells us 
     * in what ratios to add signals from different incorporation cycles.
     * generate a list of all relevant cycles.
     */
    std::vector<PhasingWindow> window = phasing.getPhasingCorrection(cycle+1, ncycles,
                                                                    is_deblocked);
    int window_index = 0;                   // current cycle
    int l;
    for (l = 0; l < int(window.size()); l++)
        if (cycle == window[l].cycle)
            window_index = l;

#if defined(DEBUG_BASECALL)
    printf("Window %2i %2i %f %f\n", 
           ncycles, nwindow, phasing.phasing_rate, phasing.prephasing_rate);
    for (l = 0; l < nwindow; l++)
        printf("%2i %2i %g\n", l, window_cycles[l], window[l]);
#endif // defined(DEBUG_BASECALL)

    /*
     * Now construct a signal vector containing the signal from all cycles
     * that have a non-negligible contribution to the phasing/prephasing.
     */
    std::vector<float *> allsignal;
    std::vector<float *> allnoise;
    std::vector<float *> allmatrices;

    for (l = 0; l < int(window.size()); l++) {
        float *this_signal    = new float [nobj*ncol];
        float *this_noise    = new float [nobj*ncol];

        memcpy(this_noise, noise[window[l].cycle], nobj*ncol*sizeof(float));
        rescale_intensities(intensities[window[l].cycle], this_signal,
                            nobj, ncol, matrices[window[l].cycle]);
        // verify that all objects have been remapped
        validate_signal(this_signal, this_noise, nobj, ncol);
        allsignal.push_back(this_signal);
        allnoise.push_back(this_noise);
        allmatrices.push_back(matrices[window[l].cycle]);
        if (window[l].cycle == cycle)
            memcpy(res_signal, this_signal, nobj*ncol*sizeof(float));
    } 

    /*
     * check that every cluster is remapped in every cycle the phasing
     * correction is applied to (i.e. the cluster hasn't shifted out of the
     * image frame.
     */
    validate_signals(allsignal, allnoise, nobj, ncol, window_index);

#if 1
    /*
     * rescale window function by average signal in each cycle to take
     * out effects of signal decay and possible focus issues.
     */
    std::vector<float> average_signals =
        estimateAverageSignals(allsignal, allnoise, nobj, ncol); 
    for (l = 0; l < int(window.size()); l++)
#if 0
        if ((average_signals[l] > 0.) &&
            (average_signals[l] > average_signals[window_index]))
#else
        if (average_signals[l] > 0.)
#endif
            window[l].value *= (average_signals[window_index] / average_signals[l]);
#endif // #if

    /* 
     * calculate signal by applying the phasing correction
     */
    int i;
    for (i = 0; i < nobj*ncol; i++)
        corr_signal[i] = 0.f;
    for (l = 0; l < int(window.size()); l++) {
        for (i = 0; i < nobj*ncol; i++)
            corr_signal[i] += window[l].value * allsignal[l][i];
    }

    /*
     * compute confidence estimates on base calls
     */
    estimateProbabilities(allsignal, allnoise, allmatrices, window,
                          corr_signal, probabilities, nobj, ncol);

    for (l = 0; l < int(allsignal.size()); l++) {
        delete [] allsignal[l];
        delete [] allnoise[l];
    }

}

/*
 * The base caller chooses the dominant base based on the matrix-corrected,
 * phasing-corrected intensities.
 * It attempts to estimate probabilities by performing a Monte-Carlo
 * integration of the posterior distribution obtained from the
 * noise covariance.
 */
void ProcessBaseCall::call(std::vector<float * > intensities,
                           std::vector<float * > noise,
                           std::vector<float * > matrices,
                           int nobj, int ncol, int cycle,
                           unsigned char *bases, int *is_deblocked,
                           unsigned char *sequence, float *probabilities,
                           float *signal, float *matrix_signal)
{
    /*
     * estimate the relative base probabilities
     */
    applyPhasingCorrection(intensities, noise, matrices,
                           nobj, ncol, cycle, is_deblocked,
                           matrix_signal, signal, probabilities);

    /*
     * assign bases (based on highest intensity; due to sampling errors
     * this may be different from the most likely base; but of course,
     * that is a rare case and indicates a very poor base call.
     */
    call_bases(signal, noise[cycle], nobj, ncol, bases, sequence+cycle*nobj);
}
