/*
 * Copyright (c) Solexa 2005-2007
 * Authors: Klaus Maisinger, Lorenzo Milazzo
 *
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
  */

//
// CVS - $Id: run_base_caller.cpp,v 1.26 2007/11/09 10:51:11 km Exp $
// Main program for the base-caller ("Bustard").


#include <string>
#include <vector>
#include <cassert>
#include <cstdio>

#include "BaseCallConfig.h"
#include "ParameterTypes.h"
#include "BaseCallParamsXMLReader.h"
#include "PhasingModel.h"
#include "ProcessBaseCall.h"

#include "DNACodes.h"
#include "ImageXMLReader.h"
#include "IntensityData.h"
#include "XMLSAXParser.h"

#include "Logger.h"
#include "Matrix.h"
#include "Tiling.h"
#include "DNAProcess.h"
#include "Message.h"
#include "Tools.h"


/*
 * parse the QC XML file for information on bases, exposure times and
 * deblocked cycles.
 */
void read_xml_info(const char filename[],
                   unsigned char bases[], float *exposure_times, 
                   int ncol, int *is_deblocked)
{
    int i;
    assert(ncol <= N_COLOURS);
    for (i = 0; i < ncol; i++) {
        exposure_times[i] = 0.;
        bases[i] = BaseUpperCase[i];
        *is_deblocked = 0;
    }
        
    FILE *fh = fopen(filename, "r");
    if (!fh)
        return;

    ImageXMLReader *handler = new ImageXMLReader();
    XMLSAXParser *sax_parser = new XMLSAXParser();
    sax_parser->setContentHandler(handler);
    sax_parser->parse(fh);
    fclose(fh);
    
    int ncol2 = handler->getCount();
    if (ncol2 != ncol)
        error("XML file %s has only %i images.", filename, ncol2);

    assert((ncol2 == ncol) || (ncol2 == 0));
    for (i = 0; i < ncol2; i++) {
        bases[i] = handler->getBases()[i];
        exposure_times[i] = handler->getExposureTimes()[i];
        *is_deblocked = handler->getDeblocked()[0];
    }
    delete sax_parser;
    delete handler;
}
/*
 * parse the parameter XML file for options
 */
void read_xml_parameters(const char filename[],
                         struct BasecallParameters *params)
{
    FILE *fh = fopen(filename, "r");
    if (!fh) {
        error("Failed to open paramter file %s.", filename);
        return;
    }

    BaseCallParamsXMLReader *handler    = new BaseCallParamsXMLReader();
    XMLSAXParser            *sax_parser = new XMLSAXParser();
    sax_parser->setContentHandler(handler);
    sax_parser->parse(fh);
    fclose(fh);
    handler->getParameters(params);
    delete sax_parser;
    delete handler;
}

ResponseMatrix *get_response_matrix(std::vector<ResponseMatrix * > response_matrices,
                                    std::vector<int> start_cycles, 
                                    int current_cycle)
{
    int i;
    int maximum = 0;
    ResponseMatrix *response_matrix = NULL;
    for (i = 0; i < int(start_cycles.size()); i++) {
        if ((current_cycle >= start_cycles[i]) && 
            (response_matrices[i] != NULL) &&
            (start_cycles[i] >= maximum)) {
            maximum = start_cycles[i];
            response_matrix = response_matrices[i];            
        }
    }

    return response_matrix;
}

/*
 * set up the frequency response matrix
 */
std::vector <float * > get_matrices(std::vector<ResponseMatrix * > response_matrices,
                                    std::vector<int> start_cycles,
                                    int ncol, int ncycles,
                                    unsigned char *bases, 
                                    std::vector<float * > exposure,
                                    Logger *logger)
{
    std::vector <float * > matrices;
    int l;
    for (l = 0; l < ncycles; l++) {
        float *matrix = NULL;
        ResponseMatrix *response_matrix = get_response_matrix(response_matrices, 
                                                              start_cycles, l);
        if (response_matrix) {
            response_matrix->set_power_correction();
            {
                char message[250];
                int pos = 0;
                matrix = response_matrix->get_matrix(ncol, bases, exposure[l]);
                pos = sprintf(message, "Matrix: ");
                snprint_matrix(message+pos, 250-pos, matrix, ncol, ncol);
                logger->info(message);
                delete [] matrix;
            }
            matrix = response_matrix->get_inverse_matrix(ncol, bases, exposure[l]);
        }
        else 
            matrix = identity_matrix(ncol);
        matrices.push_back(matrix);
    }
    return matrices;
}                                      

void call(ProcessBaseCall                      * base_caller, 
          Tiling                               * intensity_name,
          Tiling                               * sequence_name,
          int                                    ncycles,
          const struct RunInfo                 & runInfo,
          const struct TileInfo                & tileInfo,
          st_fposition                         * positions,
          int                                    npositions,
          unsigned char                        * bases, 
          const std::vector<float * >          & exposure,
          int                                  * is_deblocked, 
          const std::vector<ResponseMatrix * > & response_matrices,
          const std::vector<int>               & start_matrix_cycles,
          Logger                               * logger,
          const std::string                    & savesuffix,
          int                                    compression)
{
    int i;
    char msg[256];
    int nobj, ncol, errorno;
    snprintf(msg, sizeof(msg), "Reading intensity file %s.",
             intensity_name->getIndexFilename(INTENSITY_FILE, 0).c_str());
    logger->info(msg);
    std::vector<float * > intensities 
        = read_signal_array(intensity_name->getIndexFilename(INTENSITY_FILE,
                                                             0).c_str(),
                            &nobj, &ncol, &errorno, compression);
    assert(npositions == nobj);
    assert((int(intensities.size()) >= ncycles) ||
           (int(intensities.size()) == 0));
    snprintf(msg, sizeof(msg), "Reading noise     file %s.",
             intensity_name->getIndexFilename(NOISE_FILE, 0).c_str());
    logger->info(msg);
    std::vector<float * > noise 
        = read_signal_array(intensity_name->getIndexFilename(NOISE_FILE,
                                                             0).c_str(),
                            &nobj, &ncol, &errorno, compression);
    assert((int(noise.size()) >= ncycles) ||
           (int(noise.size()) == 0));
    assert(int(noise.size()) == int(intensities.size()));
    if (ncol == 0)
        ncol = runInfo.NumChannels;
    if (intensities.size() == 0)
        for (i = 0; i < ncycles; i++) {
            intensities.push_back(0);
            noise.push_back(0);
        }
        
    for (i = int(intensities.size()) - 1; i >= ncycles; i--) {
        delete [] intensities[i];
        delete [] noise[i];
        intensities.pop_back();
        noise.pop_back();
    }

    std::vector<float * > matrices = get_matrices(response_matrices, 
                                                  start_matrix_cycles, 
                                                  ncol, ncycles,
                                                  bases, exposure, logger);

    std::vector<float * > probabilities(ncycles);
    std::vector<float * > matrix_signal(ncycles);
    std::vector<float * > signal(ncycles);
    logger->info("Starting to base-call...");
    unsigned char *sequence = new unsigned char [nobj * ncycles];
    int cycle;
    for (cycle = 0; cycle < ncycles; cycle++) {
        probabilities[cycle] = new float [nobj*ncol];
        matrix_signal[cycle] = new float [nobj*ncol];
        signal[cycle]        = new float [nobj*ncol];
        
        sprintf(msg, "Calling cycle %i...", cycle+1);
        logger->info(msg);
        base_caller->call(intensities, noise, matrices, nobj, ncol, cycle, 
                          bases, is_deblocked, 
                          sequence, probabilities[cycle], 
                          signal[cycle], matrix_signal[cycle]);
    }


    /* save corrected intensities and probabilities to file */
    errorno = 0;
    if (!errorno) {
        snprintf(msg, sizeof(msg), "Writing %s.",
                 sequence_name->getIndexFilename(SIGNAL2_FILE, 0).c_str());
        logger->info(msg);
        errorno = write_signal_array(sequence_name->getIndexFilename(SIGNAL2_FILE,
                                                                   0).c_str(),
                                   signal, nobj, ncol, 0,
                                   tileInfo.Lane, tileInfo.TileIndex,
                                   (Position *)positions);
    }                                   
    if (!errorno) {
        snprintf(msg, sizeof(msg), "Writing %s.",
                 sequence_name->getIndexFilename(PROBABILITY_FILE, 0).c_str());
        logger->info(msg);
        errorno = write_quality_file(sequence_name->getIndexFilename(PROBABILITY_FILE,
                                                           0).c_str(),
                                   probabilities, nobj, ncol);
    }
    if (!errorno) {
        std::string filename = sequence_name->getIndexFilename(SEQUENCE_FILE, 0);
        snprintf(msg, sizeof(msg), "Writing %s.",
                 filename.c_str());
        logger->info(msg);
        filename += savesuffix;
        errorno = write_sequence_file(filename.c_str(),
                                    sequence, nobj, ncycles,
                                    tileInfo.Lane, tileInfo.TileIndex,
                                    positions);
    }
    if (errorno) {
        error("Cannot recover from IO error! Exiting...");
        perror("Error: ");
        exit(1);
    }

    delete [] sequence;
    int l;
    for (l = 0; l < ncycles; l++) {
        delete [] intensities[l];
        delete [] noise[l];
        delete [] matrices[l];
        delete [] matrix_signal[l];
        delete [] signal[l];
        delete [] probabilities[l];
    }
}

/*
 * set up and run the base-caller
 */
void run_base_caller(const char path_image_analysis[],
                     const char path_basecaller[], 
                     const char path_intensities[], 
                     const char sample[],
                     const struct RunInfo            & runInfo,
                     const struct TileInfo           & tileInfo,
                     int                               ncycles,
                     const std::vector<BasecallParameters> & params,
                     const std::vector<std::string>  & matrices, 
                     const std::vector<int>          & starts_matrices,
                     const std::string               & logfile,
                     const std::string               & savesuffix,
                     const int                         compression)
{
    assert(ncycles > 0);
    assert(runInfo.NumChannels > 0);
    assert(runInfo.NumChannels <= N_COLOURS);

    /* set up logger */
    Logger *logger = new Logger(logfile.c_str());
    logger->add_handler(new LogHandler());
    char msg[256];
    sprintf(msg, "%s%s", software, software_version);
    logger->set_source(msg); 
    sprintf(msg, "> `Bustard` - base-caller");
    logger->info(msg);
    sprintf(msg, "Base-caller path: %s", path_basecaller);
    logger->info(msg);
    sprintf(msg, "Log file: %s", logfile.c_str());
    logger->info(msg);
    if (compression == SIGNAL_COMPRESSION_GZIP)
        logger->info("Using gzip compressed intensity files.");
    if (compression == SIGNAL_COMPRESSION_BZIP2)
        logger->info("Using bzip2 compressed intensity files.");
    
    
    // load frequency response matrix
    std::vector<ResponseMatrix*> response_matrices;
    response_matrices.reserve(matrices.size());

    int i;
    for (i = 0; i < (int)matrices.size(); i++) {
        response_matrices.push_back(new ResponseMatrix(runInfo.NumChannels));
        int num = response_matrices.at(i)->load(matrices.at(i).c_str());
        if (num) {
            sprintf(msg, "Successfully loaded %ix%i matrix from %s", 
                    num, num, matrices.at(i).c_str());
        }
        else {
            sprintf(msg, "Could not find matrix file %s", matrices.at(i).c_str());
            delete response_matrices.at(i);
            response_matrices.at(i) = NULL;
        }
        logger->info(msg);
    }

    for (i = 0; i < static_cast<int>(params.size()); i++) {
        sprintf(msg, "Using phasing=%6.4f, prephasing=%6.4f from cycle %2i", 
                params[i].PhasingRate, params[i].PrephasingRate,
                params[i].FirstCycle + 1);
        logger->info(msg);
    }

    // initialise objects for creation of filenames
    std::string path = std::string(path_image_analysis) + std::string(sample);
    Tiling *sequencer_name = new Tiling(path.c_str());
    sequencer_name->setLane(tileInfo.Lane, tileInfo.TileIndex, runInfo.NumTiles);

    // initialise objects for creation of filenames
    path = std::string(path_basecaller) + std::string(sample);
    Tiling *base_caller_name = new Tiling(path.c_str());
    base_caller_name->setLane(tileInfo.Lane, tileInfo.TileIndex, runInfo.NumTiles);

    // initialise objects for creation of filenames
    path = std::string(path_intensities) + std::string(sample);
    Tiling *intensity_name = new Tiling(path.c_str());
    intensity_name->setLane(tileInfo.Lane, tileInfo.TileIndex, runInfo.NumTiles);

    ProcessBaseCall *base_caller = new ProcessBaseCall();
    base_caller->setCycleNumber(ncycles);
    PhasingModel phasing(params);
    base_caller->setPhasingModel(phasing);

    /* 
     * load intensity files and call base-caller
     */
    {
        unsigned char bases[4] = { 'A', 'C', 'G', 'T' };
        int tile  = 0;
        int ncol  = runInfo.NumChannels;
        int nread = 0;
        int *is_deblocked = new int [ncycles];
        int i;

        /* read cluster positions */
        st_fposition *referencepositions;
        read_positions(sequencer_name->getIndexFilename(POSITION_FILE,
                                                        tile).c_str(),
                       & referencepositions, & nread);
        unsigned char *allbases     = new unsigned char [ runInfo.NumChannels * ncycles ];
        float         *allexposures = new float         [ runInfo.NumChannels * ncycles ];
        int cycle;
        for (cycle = 0; cycle < ncycles; cycle++) {
            for (i = 0; i < runInfo.NumChannels; i++) {
            	allbases    [cycle*runInfo.NumChannels+i] = bases[i];
            	allexposures[cycle*runInfo.NumChannels+i] = 0.;
            }
           /* read exposure times and block info from XML file */
            read_xml_info(sequencer_name->getIndexFilename(QC_FILE,
                                                           tile, cycle).c_str(),
                          allbases+cycle*runInfo.NumChannels,
                          allexposures+cycle*runInfo.NumChannels, 
                          runInfo.NumChannels,
                          &is_deblocked[cycle]);
        }
        
        sprintf(msg, "Intensity file: %s",   
                intensity_name->getIndexFilename(INTENSITY_FILE, tile).c_str());
        logger->info(msg);
        sprintf(msg, "Noise file:     %s",   
                intensity_name->getIndexFilename(NOISE_FILE, tile).c_str());
       logger->info(msg);
            
        std::vector <float * > exposures;
        for (cycle = 0; cycle < ncycles; cycle++) {
            float *exposure_times = new float [ runInfo.NumChannels ];

            /* read intensity values */
            for (i = 0; i < ncol; i++) {
                int baseidx = get_base_index(allbases[cycle*runInfo.NumChannels+i],
                                             bases, ncol);
                assert((baseidx >= 0) && (baseidx < runInfo.NumChannels)); 
                exposure_times[baseidx] = allexposures[cycle*runInfo.NumChannels+i]; 
            }
            int offset = sprintf(msg, "Exposure:");
            for (i = 0; i < runInfo.NumChannels; i++)
                offset += sprintf(msg+offset, "  %c %6.3f",
                                  BaseUpperCase[i], exposure_times[i]); 
            logger->info(msg);
            exposures.push_back(exposure_times);
        }     
        /* call base-caller */
        call(base_caller, intensity_name, base_caller_name,
             ncycles, runInfo, tileInfo, referencepositions, nread,
             (unsigned char *)BaseUpperCase, exposures,
             is_deblocked, response_matrices, starts_matrices, logger,
             savesuffix, compression);

        for (cycle = 0; cycle < ncycles; cycle++)
            delete [] exposures[cycle];
        for (i = 0; i < int(response_matrices.size()); i++)
            if (response_matrices.at(i))
                delete response_matrices.at(i);
        delete [] allexposures;
        delete [] allbases;
        delete [] referencepositions;
        delete [] is_deblocked;
    }

    delete base_caller;
    delete logger;
    delete base_caller_name;
    delete sequencer_name;
    delete intensity_name;
}


int main(int argc, char *argv[]) 
{
    std::string path_prefix                 = "tempseq";
    std::string path_prefix2                = "tempseq";
    std::string path_prefix3                = "../tempseq";
    std::string sample                      = "s";
    char        tmp_filename[FILENAME_MAX] = "../tempseq";
    int         ncycles = 0;
    int         i, restart;
    std::string logfile;
    int         compression = SIGNAL_COMPRESSION_NONE;
    std::string savesuffix    = "";

    std::vector<int>         starts_matrix;  // cycles for new matrix (0 indexed)
    std::vector<std::string> matrices;

    std::vector<BasecallParameters> params;
    params.resize(1);
    
    TileInfo tileInfo;
    tileInfo.TileIndex     = 0;
    tileInfo.Lane         = 0;
    
    RunInfo runInfo;
    runInfo.NumTiles      = 1;
    runInfo.NumChannels   = N_COLOURS;

    for (i = 1; i < argc; i++) {
        if (argv[i][0] == '-')
            switch (argv[i][1]) {
            case 'c':
                {
                    char comp[20];
                    sscanf(argv[i], "-compression=%s", comp);
                    if (!strcmp(comp, "gzip"))
                        compression = SIGNAL_COMPRESSION_GZIP;
                    if (!strcmp(comp, "bzip2"))
                        compression = SIGNAL_COMPRESSION_BZIP2;
                } break;
            case 't': 
                switch (argv[i][2]) {
                case 'i': 
                    if ((sscanf(argv[i], "-tile=%i,%i", 
                               &(tileInfo.Lane), &(tileInfo.TileIndex)) == 2)
                        || (sscanf(argv[i], "-tile=%[^_]_%d_%d",
                                   tmp_filename, &(tileInfo.Lane), 
                                   &(tileInfo.TileIndex)) == 3)) {
                        sample = tmp_filename;
                        tileInfo.Lane --; tileInfo.TileIndex --;
                    }
                    break;
                } break;
            case 'm': 
                switch (argv[i][2]) {
                case 'a':
                    switch (argv[i][3]) {
                    case 't':
                        if (sscanf(argv[i], "-matrix=%s", tmp_filename)) {
                            matrices.push_back(tmp_filename);
                            starts_matrix.push_back(0);
                        }
                        break;
                    } 
                    break;
                }
                break;
            case 'p': 
                if (sscanf(argv[i], "-phasing=%f",    &(params[0].PhasingRate)))
                    break;
                if (sscanf(argv[i], "-prephasing=%f", &(params[0].PrephasingRate)))
                    break;
                if (sscanf(argv[i], "-prefix=%s", tmp_filename) == 1) {
                    path_prefix = tmp_filename;
                    if (path_prefix2 == "tempseq") 
                        path_prefix2 = tmp_filename;
                    break;
                }
                if (sscanf(argv[i], "-prefix2=%s", tmp_filename)) {
                    path_prefix2 = tmp_filename;
                    break;
                }
                if (sscanf(argv[i], "-prefix3=%s", tmp_filename)) {
                    path_prefix3 = tmp_filename;
                    break;
                }
                if (sscanf(argv[i], "-params=%s", tmp_filename)) {
                    read_xml_parameters(tmp_filename, &(params[0]));
                }
                break;
            case 's': 
                switch (argv[i][2]) {
                case 'a': 
                    switch(argv[i][3]) {
                    case 'm':
                        if (sscanf(argv[i], "-sample=%s", tmp_filename))
                            sample = tmp_filename;
                        break;
                    case 'v':
                        if (sscanf(argv[i], "-savesuffix=%s", tmp_filename))
                            savesuffix = tmp_filename;
                        break;
                    } break;
                } break;
            case 'n': 
                switch (argv[i][2]) {
                case 'c': sscanf(argv[i], "-nc=%i", &ncycles); break;
                case 'e':
                    if (strlen(argv[i]) > 5) {
                        switch (argv[i][5]) {
                        case 'm':
                            if (sscanf(argv[i], "-new-matrix=%i:%[^:]",
                                       &restart, tmp_filename)) {
                                matrices.push_back(tmp_filename);
                                starts_matrix.push_back(restart-1);
                            }
                            break;
                       case 'p': 
                            if (sscanf(argv[i], "-new-params=%i:%[^:]",
                                       &restart, tmp_filename) == 2) {
                                BasecallParameters param;
                                read_xml_parameters(tmp_filename, &param);
                                param.FirstCycle = restart - 1;
                                params.push_back(param);
                            }
                            break;
                       } break;
                    }
                }
            case 'l':
            	{
            	    char path[FILENAME_MAX];
                    sscanf(argv[i], "-logfile=%s", path);
                    logfile = path;
            	} break;
            }
    }

    if ((argc < 2)) {
        printf("Usage: run_base_caller [options]\n"
               "Options:\n"
           "  -matrix=<filename>\n     set the path for matrix file (single matrix)\n"
           "  -new-matrix=<filename>,<cycle>\n     set the path for matrix file and the cycle at which\n     the matrix is applied (multiple matrices)\n"
           "  -new-params=<filename>,<cycle>\n     set cycle at which the phasing/prephasing effect is reset (1 indexed)\n"
           "  -phasing=<x>\n     set phasing rate (between 0.0 and 0.1)\n"
           "  -prephasing=<x>\n     set prephasing rate (between 0.0 and 0.1)\n"
           "  -params=<filename>\n     read options from parameter XML file\n"
           "  -prefix=<path>\n     set path for image analysis output files\n"
           "  -prefix2=<path>\n     set prefix for output files\n"
           "  -prefix3=<path>\n     set prefix for input intensity files\n"
           "  -sample=<text>\n     set sample id (must not contain underscores)\n"
           "  -tile=<lane>,<tile>\n     set lane/tile (1 indexed)\n"
           "  -tile=<sample>_<lane>_<tile>\n     set sample/lane/tile (1 indexed)\n"
           "  -nc=<n>\n     set number of cycles sequenced\n"
           "  -logfile=<filename>\n     set the path for log file\n");
        exit(0);
    }

    for (i = 0; i < static_cast<int>(params.size()); i++) {
        assert((params[i].PhasingRate >= 0.)    && (params[i].PhasingRate < 0.1));
        assert((params[i].PrephasingRate >= 0.) && (params[i].PrephasingRate < 0.1));
    }

    if (logfile.length() == 0)
        logfile = path_prefix2 + std::string(software) + std::string(".log");

    run_base_caller(path_prefix.c_str(),  path_prefix2.c_str(),
                    path_prefix3.c_str(), sample.c_str(),
                    runInfo, tileInfo, ncycles,
                    params, matrices, starts_matrix,
                    logfile, savesuffix, compression);

    return 0;
}
