//  Time-stamp: <26 Jun 2006 16:54 BST DNACodes.cpp km>
/*
 * map a binary number to a DNA base or ambiguity code.
 * the number consists of 4 different segments corresponding to 4 bases:
 * - either 4 1-bit segments, using 4 bits altogether, each bit encoding
 *   the presence of absence of a base
 * - or 4 2-bit segments (using 8 bits altogether), each bit encoding
 *   the intensity of a given base on a 3-value scale (absent, low 
 *   intensity, high intensity). The corresponding ambiguity codes are 
 *   non-standard and have been used by Andrew Pike at Solexa.
 *
 * Copyright (c) Solexa 2003, 2004, 2005, 2006
 * Authors: Andrew Pike, Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */


#if defined(__BORLANDC__)
#include <ctype.h>
#else
#include <cctype>
#endif
#include "DNACodes.h"

//---------------------------------------------------------------------------

/*
 * return IUPAC ambiguity code for a 4-bit number, where each bit
 * indicates the presence or absence of a particular base in the
 * order ACGT.
 * 'N' is returned for an invalid number.
 */
unsigned char DNA_1bit_base(unsigned short code)
{
    unsigned char result[16] =
        {
            '.', 'A', 'C', 'M',       // ...., ...A, ..C., ..CA
            'G', 'R', 'S', 'V',       // .G.., .G.A, .GC., .GCA
            'T', 'W', 'Y', 'H',       // T..., T..A, T.C., T.CA
            'K', 'D', 'B', 'X' };     // TG.., TG.A, TGC., TGCA
    if (code < 16)
        return result[code];
    else
        return 'N';
}     

/*
 * return IUPAC ambiguity code for an 8-bit number. 
 * each 2-bit segment of the 8-bit value encodes a base and can only have 
 * 3 valid values (0, 1, 2) with the following meaning:
 * 0 - no base present, 1 - base detected with low significance,
 * 2 - base detected with high significance.
 * if the number is invalid, 'N' is returned.
 */
unsigned char DNA_2bit_base(unsigned short code)
{
    unsigned char result[256] =  {
    /*         ..  .a  .A  ..  c.  ca  cA  ..  C.  Ca  CA  ..  ..  ..  ..  ..*/
    /* .. */  '.','a','A','.','c','m','E','.','C','e','M','.','.','.','.','.',
    /* .g */  'g','r','F','.','s','V','V','.','L','V','V','.','.','.','.','.',
    /* .G */  'G','f','R','.','l','V','V','.','S','V','V','.','.','.','.','.',
    /* .. */  '.','.','.','.','.','.','.','.','.','.','.','.','.','.','.','.',
    /* t. */  't','w','J','.','y','H','H','.','P','H','H','.','.','.','.','.',
    /* tg */  'k','D','D','.','B','X','X','.','B','X','X','.','.','.','.','.',
    /* tG */  'Q','D','D','.','B','X','X','.','B','X','X','.','.','.','.','.',
    /* .. */  '.','.','.','.','.','.','.','.','.','.','.','.','.','.','.','.',
    /* T. */  'T','j','W','.','p','H','H','.','Y','H','H','.','.','.','.','.',
    /* Tg */  'q','D','D','.','B','X','X','.','B','X','X','.','.','.','.','.',
    /* TG */  'K','D','D','.','B','X','X','.','B','X','X','.','.','.','.','.',
    /* .. */  '.','.','.','.','.','.','.','.','.','.','.','.','.','.','.','.',
    /* .. */  '.','.','.','.','.','.','.','.','.','.','.','.','.','.','.','.',
    /* .. */  '.','.','.','.','.','.','.','.','.','.','.','.','.','.','.','.',
    /* .. */  '.','.','.','.','.','.','.','.','.','.','.','.','.','.','.','.',
    /* .. */  '.','.','.','.','.','.','.','.','.','.','.','.','.','.','.','.',
    };
    if (code >= 256)
        return 'N';
    if (code == 0)
        return '.';
    if (result[code] == '.')
        return 'N';
    return result[code];
}     

/*
 * given a base code <base>, return the index this base would have in the
 * lookup array <BaseUpperCase>.
 * If an additional array <bases> of <nbase> base codes is specified,
 * discard all bases that do not appear in <bases> from <BaseUpperCase>
 * before returning the index.
 * 
 * Example:
 * BaseUpperCase = {"A", "C", "G", "T"}
 * 
 * get_base_index('C') == 1
 * get_base_index('G') == 2
 * get_base_index('G', "AG", 2) == 1
 * get_base_index('G', "GA", 2) == 1
 */
int get_base_index(unsigned char base, unsigned char *bases, int nbase)
{
    unsigned char upperbase = toupper(base);
    int result = -1;
    int i, j;
    for (i = 0; i < 4; i++)
        if (upperbase == BaseUpperCase[i])
            result = i;
    if (nbase > 0) {
        int less = 0;
        for (i = 0; i < result; i++) {
            int match = 0;
            for (j = 0; j < nbase; j++)
                if (toupper(bases[j]) == BaseUpperCase[i])
                    match = 1;
            if (!match)
                less++;
        }
        result -= less;
    }
    return result;
}
