/*
 * Copyright (c) 2006 Solexa
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */
#ifndef INCLUDED_ELAND_OUTER_H
#define INCLUDED_ELAND_OUTER_H


#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fstream>
#include <iostream>
#include <vector>
#include <deque>
#include <string>
#include <cmath>
#include <algorithm>
#include <map>
#include <dirent.h>
#include <pthread.h>

#include <sys/resource.h>

#include "GlobalUtilities.h"

/*
PROJECT: ELAND (Efficient Local Alignment of Nucleotide Data)
MODULE: ELAND.cpp
AUTHOR: A. J. Cox

Copyright (c) 2003-2006 Solexa Limited. All rights reserved.

*/

// Identifier to embed in source and binary files

static char* eland_outer_h_id="@(#) $Id: ELAND_outer.h,v 1.11 2007/10/03 15:51:09 tc Exp $";

// ELAND_outer.h

// Contains interface functionality from ELAND
// - specifically, stuff to do with scoring and storage of alignments


// *** User #defines - change these to modify program behaviour ***


// Better to change OLIGO_LENGTH by setting in environment then
// building ELAND as "make -e eland" (may need to "make clean" first).
// Definition below will then be overridden.
#ifndef OLIGO_LENGTH
#define OLIGO_LENGTH 25
#endif

#define MAX_HASH_BITS 25
#define NUM_THREADS 1

//#define DONT_SEARCH_REVERSE_STRAND
// if next line is uncommented, only do the first pass out of three
// This means all single error matches are found but only some
// double error matches
//#define ONE_ERROR_PER_OLIGO

// if next line is uncommented, interpret N characters as deletions
// as well as missing bases when doing alignments - deprecated, see documents
//#ifdef ALLOW_N_TO_BE_DELETION



//#define DEBUG xx

// *** end of user #defines, don't mess with anything from here on! ***


// Program-wide compile-time constants
enum
{
  // hashShift is the number of bits to shift for a given OLIGO_LENGTH
  // so that the number of bases stored in each of the two unsigned ints
  // is either equal (if OLIGO_LENGTH is even) or else (if OLIGO_LENGTH is
  // odd) differs only by 1.
  hashShift = ((2*numBitsPerWord/numBitsPerBase)-OLIGO_LENGTH+1)&(~1),

  // an oligo is divided into four fragments
  fragmentsPerOligo = 4,
  //  basesPerFragment = (OLIGO_LENGTH/fragmentsPerOligo),
  //  bitsPerFragment = 2*basesPerFragment,
  
  // oligo is partitioned into a prefix and suffix, each containing two
  // fragments
  fragmentsPerPartition = 2,

  // there are 6 ways of partitioning the 4 fragments into a prefix and 
  // a suffix (not 24 as we are not concerned about the ordering of the 
  // fragments within the prefix and suffix). Two of these partitionings are
  // searched on each pass through the genome, hence 3 passes required in all
  numPasses = 3,

  // if more than this many matches of a given quality have already been found
  // for an oligo then don't look for any more, e.g. if we have more than 3 
  // matches with 2 errors, then only single error or exact match errors 
  // will be looked for on subsequent passes
  maxNumBestMatches = 3,

  // length (in bits) of largest suffix that will be needed for this oligo length 
  // length 20 => fragments = 5 5 5 5 => largest suffix =5+5=10 bases = 20 bits
  // length 21 => fragments = 5 6 5 5 => largest suffix =5+6=11 bases = 22 bits
  // length 22 => fragments = 5 6 5 6 => largest suffix =6+6=12 bases = 24 bits
  // length 23 => fragments = 6 5 6 6 => largest suffix =6+6=12 bases = 24 bits
  // length 24 => fragments = 6 6 6 6 => largest suffix =6+6=12 bases = 24 bits
  maxPrefixLength = OLIGO_LENGTH
  +((OLIGO_LENGTH%fragmentsPerOligo)!=0)
  +((OLIGO_LENGTH%fragmentsPerOligo)==2),

  maxOligoLength = 2*maxBasesPerWord,

  // Oligos are split into fragments as follows:
  // Oligo Length  Fragment Lengths: A B C D
  // 20                              5 5 5 5
  // 21                              5 6 5 5
  // 22                              5 6 5 6
  // 23                              6 5 6 6 
  // 24                              6 6 6 6 
  // 25                              6 7 6 6 ... etc.
  fragLengthA=((OLIGO_LENGTH/fragmentsPerOligo) 
	       + ((OLIGO_LENGTH%fragmentsPerOligo)==3)),
  fragLengthB=((OLIGO_LENGTH/fragmentsPerOligo)
	       + (    ((OLIGO_LENGTH%fragmentsPerOligo)==1)
		      ||((OLIGO_LENGTH%fragmentsPerOligo)==2))),
  fragLengthC=((OLIGO_LENGTH/fragmentsPerOligo)
	       + ((OLIGO_LENGTH%fragmentsPerOligo)==3)),
  fragLengthD=((OLIGO_LENGTH/fragmentsPerOligo)
	       + (    ((OLIGO_LENGTH%fragmentsPerOligo)==2)
		      ||((OLIGO_LENGTH%fragmentsPerOligo)==3))),
  prefixLength=fragLengthC+fragLengthD,
  suffixLength=fragLengthA+fragLengthB

};

// Software operates in two modes: 
// 1. Standard mode, each oligo is split into prefix and suffix, prefix used to
// make a look up table
// 2. "Split prefix" mode: prefix is too large to be used to make a look up table
// so "hide" some bits in the TableEntry

const bool useSplitPrefix(maxPrefixLength>MAX_HASH_BITS);
const int maxNumOligos((1<<(MAX_HASH_BITS-1))-1);
//const int prefixBitsToHide(maxPrefixLength-MAX_HASH_BITS);
// next two should be OligoNumbers really?
const Word splitPrefixMaskHigh( ((Word)~0) << MAX_HASH_BITS );
const Word splitPrefixMaskLow(~splitPrefixMaskHigh);


const Word prefixMask
(((uint)prefixLength==(uint)maxBasesPerWord)
 ? 0
 : (((Word)~0)<<(prefixLength*numBitsPerBase)) );
const Word suffixMask
(((uint)suffixLength==(uint)maxBasesPerWord)
 ? 0
 : (((Word)~0)<<(suffixLength*numBitsPerBase)) );


// code for checking compile time assertions

template<bool> struct CompileTimeAssertionFailure;
template<> struct CompileTimeAssertionFailure<true> {};
#define STATIC_ASSERT(exp)  CompileTimeAssertionFailure<(exp)!=0>()

// SanityCheck
// Sole purpose of this class is to fail to compile if any of the
// STATIC_ASSERTs fail at compile time
class SanityCheck
{
private:
  SanityCheck()
  {
    assert(1==0);
    STATIC_ASSERT(OLIGO_LENGTH<=maxOligoLength);
    //    STATIC_ASSERT(OLIGO_LENGTH==fragLengthA+fragLengthB+
    //		  fragLengthC+fragLengthD);
    //   STATIC_ASSERT((fragMaskA&fragMaskB)==0);
    //   STATIC_ASSERT((fragMaskC&fragMaskD)==0);
  } // ~ctor
}; // ~class SanityCheck


// The way in which an oligo matches to the genome data is described by
// one instance of MatchPosition 
// one instance of MatchDescriptor
// The match as described by these two quantities may be categorised as
// being in one of seven states:
// 
// 1: NM - No Match for this oligo
// 2: UE - Unique Exact match for this oligo
// 3: U1 - Unique match for this oligo with a single base substitution error
// 4: U2 - Unique match for this oligo, having 2 base substitution errors
// 5: RE - Unique Exact match for this oligo
// 6: R1 - Unique match for this oligo with a single base substitution error
// 7: R2 - Unique match for this oligo, having 2 base substitution errors


// MatchPosition stores a position in the genome (chromosome number + position)
// as a single unsigned int
// Capacity of an unsigned int is split into blocks according to the
// value of the most significant byte
// 0 : no match
// 1 - 239 : reserved for sequence data
// 240 - 255 : repeat
// blockRepeat + x : this oligo is the same as oligo number x in the batch
// ~0 (all bits 1) : match not attempted - oligo failed quality control (too many Ns etc)


typedef uint MatchPosition;

const MatchPosition noMatch(0);
const int blockShift(24);
const MatchPosition blockMask(((MatchPosition)0xFF)<<blockShift);
const MatchPosition blockRepeat(((MatchPosition)0xF0)<<blockShift);
const MatchPosition blockSize(((MatchPosition)1)<<blockShift);
const MatchPosition qualityFailed(~(MatchPosition)0);
const MatchPosition repeatMasked(qualityFailed-1);

// MatchDescriptor contains the remaining information necessary to describe
// a match between an oligo and the genome
//
// errorType:
//  bit 7: 1 = reverse complement match (ignored if MatchPosition is a repeat)
//  bit 2-5: pass at which match found, needed to convert error positions 
//   in the suffix to positions in the oligo. 
//  bit 0-1: number of errors in best match found so far (if zero, need to
//  check MatchPosition is non zero to distinguish from the case where no 
//  match has yet been found)
//
// registers r[0], r[1], r[2]: 
//  The roles of these registers vary according to the current state of the
//  match. The table below gives the contents of the three registers for each 
//  of the seven states, and how the state responds upon receipt of each of 
//  the three types of match.
//
//  Key:
//  'E'=exact match, '1'=match with 1 error, '2'=match with 2 errors
//  e0, e1 = positions of errors in suffix
//  #E, #1, #2 = counts of E,1,2 matches
//  xx = don't care
//  Copy= replace current match details, ->XX transition to state XX 
// 
// State match r0 r1 r2 | E               | 1               | 2
// ---------------------|-----------------|-----------------|----------------
// NM    noM    0  0  0 | Copy, r0++ ->UE | Copy, r1++ ->U1 | Copy, r2++ ->U2
// UE    mPos   1 #1 #2 |       r0++ ->RE |       r1++      |       r2++     
// U1    mPos  e0  1 #2 | Copy, r0++ ->UE |       r1++ ->R1 |       r2++ 
// U2    mPos  e0 e1  1 | Copy, r0++ ->UE | Copy, r1++ ->U1 |       r2++ ->R2
// RE    xx    #E #1 #2 |       r0++      |       r1++      |       r2++
// R1    xx    xx #1 #2 | Copy, r0++ ->UE |       r1++      |       r2++
// R2    xx    xx xx #2 | Copy, r0++ ->UE | Copy, r1++ ->U1 |       r2++

struct MatchDescriptor
{
  uchar errorType;
  uchar r[3];
}; // ~struct MatchDescriptor


union SuffixType
{
  Word ui;
  uchar uc[sizeof(Word)];
}; // ~union SuffixType

typedef unsigned short MaskTableEntry;

typedef uint OligoNumber;
const OligoNumber isReverseOligo
( ((OligoNumber)1) << (MAX_HASH_BITS-1) );
//const OligoNumber isReverseOligo
//( ((OligoNumber)1) << ((numBitsPerByte*sizeof(OligoNumber))-1) );

// Each Partition HashTable consists of a look up table of pointers into a list
// of table entries. Templatized because < and == operators need to change
// in split prefix mode
template <bool useSplitPrefix> struct TableEntry
{
  SuffixType suffix;
  MaskTableEntry mask;
  OligoNumber position;
  inline bool operator<(const TableEntry<useSplitPrefix>& rhs) const;
  inline bool operator==(const TableEntry<useSplitPrefix>& rhs) const;
  inline bool operator!=(const TableEntry<useSplitPrefix>& rhs) const;

} __attribute__ ((packed)); // ~struct TableEntry


typedef ushort FragmentErrorType;

#ifdef OLD_FRAGMENT_ERROR_DEFS
const int errorBits(6);
const FragmentErrorType incrementDist
( ((FragmentErrorType)1) << (2*errorBits) );
const FragmentErrorType errorMask1( (((FragmentErrorType)1)<<errorBits)-1 );
const FragmentErrorType errorMask2( errorMask1<<errorBits );
const FragmentErrorType distanceMask(~(errorMask1|errorMask2));
const FragmentErrorType moreThanTwoErrors
( ((FragmentErrorType)3) << (2*errorBits) );
const FragmentErrorType twoErrors
( ((FragmentErrorType)2) << (2*errorBits) );
const FragmentErrorType oneError
( ((FragmentErrorType)1) << (2*errorBits) );
#endif

// Constants to facilitate storing two error descriptions in an
// unsigned short. 
//
// Storage scheme is (bits numbered starting at from least significant):
// Bits 1-6: Position of 1st error
// (so this storage scheme only works for <=2^6=64 base reads)
// Bits 7-8: XOR result of 1st error 
// (can combine this with original base to infer what the error was)
// Bits 9-14: Position of 2nd error (if present, else zero)
// Bits 15-16: XOR of 2nd error (if present, else zero)
// 
// If more than 2 errors, all bits are set to 1 
// If more than 1 error, value of FragmentErrorType is greater than 255
// If no errors at all, all bits are off!

const int errorPosBits__(6);
const int errorTypeBits__(2);
const int errorBits__(errorPosBits__+errorTypeBits__);
const FragmentErrorType moreThanTwoErrors__(~((FragmentErrorType)0));
const FragmentErrorType errorPosMask1__(0x3F);  // binary 00111111
const FragmentErrorType errorTypeMask1__(0xC0); // binary 11000000
const FragmentErrorType errorInfoMask1__(errorPosMask1__|errorTypeMask1__);

const FragmentErrorType errorPosMask2__(errorPosMask1__<<errorBits__);
const FragmentErrorType errorTypeMask2__(errorTypeMask1__<<errorBits__);
const FragmentErrorType errorInfoMask2__(errorPosMask2__|errorTypeMask2__);
const FragmentErrorType oneError__(errorPosMask1__|errorTypeMask1__);
const FragmentErrorType noErrors__(0);

// These are equivalent uchar definitions for pulling the error info
// out of a uchar in a MatchDescriptor prior to printing
const uchar ucharErrorPosMask(0x3F);
const uchar ucharErrorTypeMask(0xC0);

// These codes are used to store information about the Ns used in a match
const Word noNonIncorps(0);
// Any Ns that are in the oligo (possibly none) are all non-detections 

const Word nonIncorpFirstN(0x55555555); 
// this is 101010... in binary
// first N is a non-incorporation, second (if present) is a non-detection

const Word nonIncorpSecondN(0xAAAAAAAA);
// this is 010101... in binary
// second N is a non-incorporation, first is a non-detection

const Word nonIncorpBothNs(0xFFFFFFFF);
// this is 111111... in binary
// both Ns are non-incorporations



// class SuffixScoreTable
// This stores the scoring information for fast inexact matching of suffixes
// suffixScore_ - gives match score based on XOR of two suffixes
// quickByte_ - filters out words based on examining single bytes
struct SuffixScoreTable
{
  SuffixScoreTable();
  //  typedef bool ByteTable[numPossibleChars];
  //  typedef ByteTable ByteTableSet[numPossibleChars];

  void buildScoreTable( vector<FragmentErrorType>& score, const int fragLength, const int offset );

  //  vector<SuffixErrorType> suffixScore_;
  vector<FragmentErrorType> scoreFragA_;
  vector<FragmentErrorType> scoreFragB_;
  vector<FragmentErrorType> scoreFragC_;
  vector<FragmentErrorType> scoreFragD_;

  //  ByteTableSet quickByte_;
}; // ~struct SuffixScoreTable


// class ContigNameFinder
// If required, this converts a match position into a contig name + position 
// in contig, else returns a null string and leaves position unchanged

class ContigNameFinder
{
 public:
 static ContigNameFinder* getContigNameFinder
   ( const string& directoryName, const string& chromName );
 virtual void operator() ( const char*& pContigName, uint& outputPos )=0;
 virtual ~ContigNameFinder() {}
};

class ContigNameFinderNull : public ContigNameFinder
{
 public:
  ContigNameFinderNull( void ): nullContigName('\0') {}
  virtual void operator() ( const char*& pContigName, uint& outputPos )
  {
    pContigName=&nullContigName;
  } // ~operator()   
 private:
  const char nullContigName;
};

class ContigNameFinderIndex : public ContigNameFinder
{
 public:
  ContigNameFinderIndex( FILE* pIndex )
  {
    const uint bufSize(2048);
    char buf[bufSize];
    uint offset;

    int numItems;
    while( (numItems=fscanf( pIndex, "%u\t%s\n", &offset, buf))==2)
    {
      offsets_.push_back(offset);
      names_.push_back('/'+(string)(buf+1));
      cerr << "Parsed offset=" << offsets_.back() << ", name="
      	   << names_.back() << endl;
    } // ~while
    if (numItems!=EOF)
    {
	cerr << "Problem parsing line in index file :" 
	     << " " <<buf << endl;
	exit (1);
    } // ~if
    fclose(pIndex);
  } // ~ctor

  virtual void operator() ( const char*& pContigName, uint& outputPos )
  {
    vector<uint>::iterator i
      (lower_bound(offsets_.begin(),offsets_.end(),outputPos));
    --i;
    outputPos-=*i;
    pContigName=names_[i-offsets_.begin()].c_str();
    // TBD
  } // ~operator()   
 private:
  vector<uint> offsets_;
  vector<string> names_;

};






// class MatchPositionTranslator
// This converts a match position into chromosome name + position
// and optionally a contig name
typedef MatchPosition SubtractTable[numPossibleChars];
typedef int ChromTable[numPossibleChars];




class MatchPositionTranslator
{
public:
  MatchPositionTranslator( const vector<string>& chromNames,
			   const vector<MatchPosition>& blockStarts,
			   const string& directoryName ) :
    chromNames_(chromNames)
  {
    int thisChrom(0);
    uint j;
    for (vector<MatchPosition>::const_iterator i(blockStarts.begin());
	 i!=blockStarts.end()-1;++i)
    {
      for ( j = *i; j != *(i+1); j+=blockSize)
      {
	subtractTable_[j>>blockShift]=*i;
	chromTable_[j>>blockShift]=thisChrom;
      } // ~for j
      ++thisChrom;
    } // ~for i

    // First entry of chromNames is null so need to miss it out
    // and add a null entry to getContigName_ to match
    vector<string>::const_iterator i(chromNames.begin());
    i++;
    getContigName_.push_back( new ContigNameFinderNull() ); 

    for (;i!=chromNames.end(); ++i)
    {
      getContigName_.push_back
	(ContigNameFinder::getContigNameFinder(directoryName, *i));
    } // ~for
  } // ~ctor

  ~MatchPositionTranslator()
  {
    for (vector<ContigNameFinder*>::iterator i(getContigName_.begin());
	 i!=getContigName_.end();++i)
      delete *i;
  } // ~dtor

  void operator()( const MatchPosition originalPos,
		 const char*& chromName,
		 const char*& contigName,
		 MatchPosition& outputPos)
  {
		 
	MatchPosition thisBlock = originalPos >> blockShift;
	outputPos=originalPos-subtractTable_[thisBlock];
	int thisChrom = chromTable_[thisBlock];
	chromName=chromNames_[thisChrom].c_str();
	contigName=chromName;
	assert((uint)thisChrom<getContigName_.size());
	(*getContigName_[thisChrom])( contigName, outputPos );
  } // ~operator()
 private:
  const vector<string>& chromNames_;
  ChromTable chromTable_;
  SubtractTable subtractTable_;
  vector<ContigNameFinder*> getContigName_;
}; // class MatchPositionTranslator





// struct MatchTable: this stores all the match information accumulated
// for each oligo during the search
class MatchTable
{
public:
  static MatchTable* getMatchTable( const char* outputFileName );

  MatchTable( const char* outputFileName ) :
    pOut_(fopen(outputFileName, "w"))
  {
#if (NUM_THREADS>1)
    pthread_mutex_init(&mutex_, NULL);
#endif
    if (pOut_==NULL) 
    {
      cerr << "Error in MatchTable: could not open file "
           << outputFileName << endl;
      exit (1);
    } // ~if
  } // ~ctor
  virtual ~MatchTable() 
  { 
    fclose (pOut_); 
  } // ~dtor

  // addMatch: process a match
  // oligoNum - number of the oligo found
  // matchPos - where it was found in the genome
  // matchType - number and position of any errors in the match
  // partitionNumber - which pass of the genome the match was found on
  virtual void addMatch( const TableEntry<useSplitPrefix>& entry, 
			const Word thisMask,
			const MatchPosition matchPos,
			const FragmentErrorType errorLow,
			const FragmentErrorType errorHigh );
  //			const int numErrors );
  //			const int nInfo );

  // print: print the search results to standard output at the end of a run
  virtual void print( OligoSource& oligos, 
		      MatchPositionTranslator& getMatchPos,
		      const vector<string>& chromNames,
		      const vector<MatchPosition>& blockStarts,
		      const SuffixScoreTable& scoreTable,
		      int oligoLength )=0;

  virtual void resize( int n )
  {
    matchPosition_.resize(n);
    matchType_.resize(n);
  } // ~void resize( int n )

  virtual void setQualityFailed__( int i )
  {
    if (matchPosition_.size()!=0)
    {
      assert((int)matchPosition_.size()>i);
      matchPosition_[i]=qualityFailed;
    }
  } // ~setQuality__

  virtual void setRepeatMasked__( int i )
  {
    matchPosition_[i]=repeatMasked;
  } // ~setRepeatMask__

  virtual bool isRepeatMasked__( int i ) const
  {
    return (matchPosition_[i]==repeatMasked);
  } // ~isRepeatMasked__

  virtual bool hasQualityFailed__( int i ) const
  {
    return (matchPosition_[i]==qualityFailed);
  } // ~hasQualityFailed__

  //  virtual bool hasQualityFailed__( int i ) const
  //  {
  //   return (matchPosition_[i]==qualityFailed);
  //  } // ~isRepeat__
  
  virtual bool isInterested__( int oligoNum, int PASS, bool hasNs );


  virtual void setSameAs__( OligoNumber newOligo, OligoNumber existingOligo )
  {
    if (  (matchPosition_[newOligo]<blockRepeat)
          && (matchPosition_[existingOligo]!=blockRepeat+newOligo) )
    {
      matchPosition_[newOligo] = blockRepeat+existingOligo;
    } // ~if
  } // ~setSameAs__
  size_t size( void ) const
  {
    assert(matchPosition_.size()==matchType_.size());
    return matchPosition_.size();
  } // ~size_t size( void ) const

  // count Ns: count the number of heading, trailing and internal Ns in
  // an oligo - used by adjustMatchPos;
  void countNs( const char* pOligo, 
		int& headSize, 
		int& tailSize,
		int& numInternalNs )
  {
    headSize=0;
    tailSize=0;
    numInternalNs=0;
    bool inHead(true);
    for (int i(0); i<OLIGO_LENGTH; ++i)
    {
      //      if ((pOligo[i]=='n')||(pOligo[i]=='N'))
      if (isBlank(pOligo[i]))
      {
	headSize+=(inHead==true);
	++tailSize;
	++numInternalNs;
      } // ~if
      else 
      {
	inHead=false;
	tailSize=0;
      } // ~else
    } // ~for
    numInternalNs-=headSize;
    numInternalNs-=tailSize;
  } // ~countNs

  // adjustMatchPos: make small corrections to the match position to correct
  // for various quirks of the matching process (mainly to do with 
  // incorporation errors and reverse complements)
  void adjustMatchPos( const char* pOligo, 
		       const char dirChar,
		       char& firstN,
		       char& secondN,
		       MatchPosition& matchPos);

protected:
  vector<MatchPosition> matchPosition_;
  vector<MatchDescriptor> matchType_;

  FILE* pOut_;
#if (NUM_THREADS>1)
  pthread_mutex_t mutex_;
#endif

}; // ~struct MatchTable




// MatchTableVerbose: produce verbose match output
class MatchTableVerbose : public MatchTable
{
public:
  MatchTableVerbose( const char* outputFileName ) :
    MatchTable( outputFileName ) {} // ~ctor

  virtual void print( OligoSource& oligos, 
		      MatchPositionTranslator& getMatchPos,
		      const vector<string>& chromNames,
		      const vector<MatchPosition>& blockStarts,
		      const SuffixScoreTable& scoreTable,
		      int oligoLength);
}; // ~class MatchTableVerbose


// MatchTableAssembly : print output in a format compatible with the 
// assembly module
class MatchTableAssembly : public MatchTable
{
public:
  MatchTableAssembly( const char* outputFileName ) :
    MatchTable( outputFileName ) {} // ~ctor

  virtual void print( OligoSource& oligos, 
		      MatchPositionTranslator& getMatchPos,
		      const vector<string>& chromNames,
		      const vector<MatchPosition>& blockStarts,
		      const SuffixScoreTable& scoreTable,
		      int oligoLength);

  // parseErrorInfo: parses the information stored in the error char
  // - bottom 6 bits are the error position
  // - using the top 2 bits and the original read it works 
  //   out what the erroneous base should have been
  void parseErrorInfo( const char* oligoBuf, 
                       const char errorInfo, 
                       const char dirChar, 
                       uint& errorPos,
                       uchar& shouldBe );


}; // ~class MatchTableAssembly



struct MultiMatch
{
  MultiMatch( MatchPosition pos, uchar type ) : 
    pos_(pos), type_(type) {}
  MatchPosition pos_;
  uchar type_;
  bool operator==( const MultiMatch& rhs ) const
  {
    return(pos_==rhs.pos_);
  } // ~bool operator==( const MultiMatch& rhs )

  bool operator<( const MultiMatch& rhs ) const
  {
    return(pos_<rhs.pos_);
  } // ~bool operator==( const MultiMatch& rhs )

  bool operator!=( const MultiMatch& rhs ) const
  {
    return(pos_!=rhs.pos_);
  } // ~bool operator!=( const MultiMatch& rhs )



} __attribute__ ((packed)); // ~struct MultiMatch

//static bool lessThanMultiMatch
//( const MultiMatch& lhs, const MultiMatch& rhs )
//{
//  return (lhs.pos_<rhs.pos_);
//} // ~lessThanMultiMatch

// MatchTableMulti : save multiple matches for each oligo in a temporary file
class MatchTableMulti : public MatchTable
{
public:
  enum
  {
    maxNumMatchesDefault=10
    //  maxNumMatchesOneError=10,
    //  maxNumMatchesTwoErrors=10
  };

  //  MatchTableMulti( const char* outputFileName, 
  //		   const char* tmpFilePrefix=NULL );

  // initialize does some setup that is common to all constructora
  void initialize( const char* tmpFilePrefix=NULL);


  // Specify separate max numbers of 0,1,2-error matches
  MatchTableMulti( const char* outputFileName,
		   const int maxNumMatchesExact,
		   const int maxNumMatchesOneError,
		   const int maxNumMatchesTwoErrors,
		   const char* tmpFilePrefix=NULL) :
    MatchTable( outputFileName ),
    maxNumMatchesExact_(maxNumMatchesExact),
    maxNumMatchesOneError_(maxNumMatchesOneError), 
    maxNumMatchesTwoErrors_(maxNumMatchesTwoErrors),
    matchesStored_(0)
  {
    initialize( tmpFilePrefix ); 
  } // ~ctor


  // Use default number of matches per read
  MatchTableMulti (const char* outputFileName,
		   const char* tmpFilePrefix=NULL) :
    MatchTable( outputFileName ),
    maxNumMatchesExact_(maxNumMatchesDefault),
    maxNumMatchesOneError_(maxNumMatchesDefault), 
    maxNumMatchesTwoErrors_(maxNumMatchesDefault), 
    matchesStored_(0)
  {
    initialize( tmpFilePrefix );
  } // ~ctor
    

  // Store same number of matches for each type of match
  MatchTableMulti (const char* outputFileName,
		   const int maxNumMatches,
		   const char* tmpFilePrefix=NULL) :
    MatchTable( outputFileName ),
    maxNumMatchesExact_(maxNumMatches),
    maxNumMatchesOneError_(maxNumMatches), 
    maxNumMatchesTwoErrors_(maxNumMatches), 
    matchesStored_(0)
  {
    initialize( tmpFilePrefix );
  } // ~ctor

		   
  ~MatchTableMulti();

  virtual void addMatch( const TableEntry<useSplitPrefix>& entry, 
			 const Word thisMask,
			 const MatchPosition matchPos,
			 const FragmentErrorType errorLow,
			 const FragmentErrorType errorHigh );

  virtual void print( OligoSource& oligos, 
		      MatchPositionTranslator& getMatchPos,
		      const vector<string>& chromNames,
		      const vector<MatchPosition>& blockStarts,
		      const SuffixScoreTable& scoreTable,
		      int oligoLength);



  const int maxNumMatchesExact_;
  const int maxNumMatchesOneError_;
  const int maxNumMatchesTwoErrors_;

  uint matchesStored_;
  FILE* pOligoNum_;
  FILE* pMatchType_;
  string tmpFileNameOligoNum_;
  string tmpFileNameMatchType_;
  vector<vector<MultiMatch> > multiMatch_;

  //  vector<vector<MatchPosition> > multiPos_;
  //  vector<vector<uchar> > multiType_;


}; // ~class MatchTableMulti





#endif // #ifndef INCLUDED_ELAND_OUTER_H
