#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fstream>
#include <iostream>
#include <vector>
#include <map>
#include <cmath>
#include <algorithm>
#include <dirent.h>
#include <pthread.h>
#include "GlobalUtilities.h"
#include <errno.h>


// Copyright (c) 2007 Solexa
// This source file is covered by the "Solexa Public Source License"
// agreement and bound by the terms therein.

// PROJECT: ELAND (Efficient Local Alignment of Nucleotide Databases)
// MODULE: Unsquash.cpp
// AUTHOR: A. J. Cox

// Description: grab sequence fragments from a directory of sequence files
// that have been 'squashed' by SquashGenome.cpp into the 2-bits-per-base
// format used by ELAND.
// For convenience, text files in ELAND output format are used as input.


using namespace std;

typedef unsigned char uchar;

// bufSize: maximum size of an input line
const uint bufSize(2048);
// batchSize: this many requests will be stored then sorted and fulfilled
// in order of file name and position
const uint batchSize(262144);

// maxReadLengthELAND: longest read length allowable by ELAND
const uint maxReadLengthELAND(32);

#ifdef XXX
// StringIndex: assigns an index number to each different string seen.
// This index number can be used to retrieve the corresponding name. 
// map<std::string,int> will
// do pretty much the same job but I want to avoid having to instantiate a 
// std::string object for each line of input parsed. Std::strings are just
// used as convenient stores for const char* strings.
struct StringIndex
{
  StringIndex() 
  { 
    // Need to make sure the vector is not reallocated as all the pointers
    // then become invalid. Assuming each entry refers to a different file
    // is the most conservative assumption.
    names_.reserve(batchSize); 
  } // ~ctor
  struct LessThanString
  {
    bool operator()( const char* a, const char* b ) const
    {
      return (strcmp(a, b)<0);
    } // ~bool lessThanGenomeFrag
  };

  vector<string> names_;
  map<const char*, int, LessThanString > index_;
  
  int getIndex(const char* name)
  {
    if (index_.find(name)==index_.end())
    { // add name to index if not already seen
      names_.push_back(string(name));
      index_[names_.back().c_str()]=names_.size()-1;
    } // ~if
    return index_[name];
  } // ~get 
}; // ~struct StringIndex
#endif



struct ContigIndex
{
  ContigIndex( const string& squashDirName, 
	       const char* name )
  { 
    string fileName(squashDirName
		    +(string)"/"
		    +(string)name
		    +(string)".idx");
    cerr << "Making index " << fileName << endl;
    FILE* pIndex=fopen(fileName.c_str(),"r");
    if (pIndex==NULL)
    {
      cerr << "Could not find index file " << fileName << endl;
      exit (1);
    } // ~if
    const uint bufSize(2048);
    char buf[bufSize];
    uint offset;
    vector<uint> offsets;

    int numItems;
    while( (numItems=fscanf( pIndex, "%u\t%s\n", &offset, buf))==2)
    {
      offsets.push_back(offset);
      names_.push_back((string)(buf+1));
      cerr << "Parsed offset=" << offsets.back() << ", name="
      	   << names_.back() << endl;
    } // ~while
    if (numItems!=EOF)
    {
	cerr << "Problem parsing line in index file :" 
	     << " " <<buf << endl;
	exit (1);
    } // ~if
    fclose(pIndex);

    // populate map at end, avoids any problems with out of date pointers
    for (uint i(0);i<offsets.size();++i)
    {
      index_[names_[i].c_str()]=offsets[i];
    } // ~for
  } // ~ctor


  void adjustPos (const char* contigName, uint& chromPos )
  {
    map<const char*, int, LessThanString >::iterator 
      i(index_.find(contigName));
    if (i!=index_.end())
    {
      chromPos+=i->second;
    } // ~if
    else
    {
      cerr << "Could not find entry for " << contigName << endl;
      exit (1);
    } // ~else
  } // ~adjustPos

  struct LessThanString
  {
    bool operator()( const char* a, const char* b ) const
    {
      return (strcmp(a, b)<0);
    } // ~bool lessThanGenomeFrag
  };

  vector<string> names_;
  map<const char*, int, LessThanString > index_;
  
  //  int getIndex(const char* name)
  //  {
  //   if (index_.find(name)==index_.end())
  //  { // add name to index if not already seen
  //   names_.push_back(string(name));
  //  index_[names_.back().c_str()]=names_.size()-1;
  //  } // ~if
  //   return index_[name];
  //  } // ~get 
}; // ~struct ContigIndex



// StringIndex: assigns an index number to each different string seen.
// This index number can be used to retrieve the corresponding name. 
// map<std::string,int> will
// do pretty much the same job but I want to avoid having to instantiate a 
// std::string object for each line of input parsed. Std::strings are just
// used as convenient stores for const char* strings.
// TC 1.8.7 - modified to distinguish between 'chromName' and
// 'chromName/contigName' - in the latter case, adjust the match position
// from position in contig to position in file
struct StringIndex
{
  StringIndex( const string& squashDirName ) :
    squashDirName_(squashDirName)
  { 
    cerr << "Building string index for " << squashDirName << endl;
    // Need to make sure the vector is not reallocated as all the pointers
    // then become invalid. Assuming each entry refers to a different file
    // is the most conservative assumption.
    names_.reserve(batchSize); 
  } // ~ctor

  ~StringIndex()
  {
    for ( vector<ContigIndex*>::iterator i(contig_.begin()); 
	  i!=contig_.end(); ++i )
      delete *i;
  } // ~dtor

  struct LessThanString
  {
    bool operator()( const char* a, const char* b ) const
    {
      return (strcmp(a, b)<0);
    } // ~bool lessThanGenomeFrag
  };

  
  //  int getIndex(const char* name)
  void getIndex(char* name, uint& chromNum, uint& chromPos)
  {
    //    cerr << "getIndex " << name << endl;
    char* contigName(NULL);
    if ((contigName=strchr(name,'/'))!=NULL)
    {
      *contigName++='\0';
    } // ~if

    if (index_.find(name)==index_.end())
    { // add name to index if not already seen
      names_.push_back(string(name));
      index_[names_.back().c_str()]=names_.size()-1;
      contig_.push_back(NULL);
    } // ~if
    chromNum=index_[name];
    if (contigName!=NULL)
    { // specified chrom/contig, adjust contig pos to file pos
      //    cerr << "found contig name " << contigName << endl;
      if (contig_[chromNum]==NULL)
      {
	contig_[chromNum]= new ContigIndex
	  ( squashDirName_, name );
      } // ~if
      contig_[chromNum]->adjustPos( contigName, chromPos );
    } // ~if
  } // ~getIndex 

  const string& squashDirName_;
  vector<string> names_;
  vector<ContigIndex*> contig_;
  map<const char*, int, LessThanString > index_;

}; // ~struct StringIndex



struct SeqRequest
{
  SeqRequest( int requestNum, 
	      int readNum,
	      int fileIndex, 
	      int filePos, 
	      char strand ) :
    requestNum_(requestNum),
    readNum_(readNum),
    fileIndex_(fileIndex),
    filePos_(filePos),
    strand_(strand)
  {}
  int requestNum_;
  int readNum_;
  int fileIndex_;
  int filePos_;
  char strand_;
}; // ~struct SeqRequest

bool lessThanRequest (const SeqRequest& a, const SeqRequest& b)
{
  return ((a.fileIndex_<b.fileIndex_)
	  ||((a.fileIndex_==b.fileIndex_)&&(a.filePos_<b.filePos_)));
} // ~bool lessThanRequest;



// SquashFile: provide interface to a 'squashed' chromosome file. Assumes:
// First line might be a comment starting with '>', rest of lines are sequence.
// All sequence lines contain same number of bases, except possibly the last.
// NB Numbering of bases in file starts at 1.
class SquashFile
{
public:
  SquashFile( const char* name );
  ~SquashFile() { munmap( pMap, fileSize ); }

  uint getNextBase( uint i)
  {
    return ( (pWord[i>>4])>>(2*((i&0xF)^0xF)) ) & 0x3; 
  } // ~getNextBase

  char operator[]( int pos )
  {
    if (pos<0) return 'N';
    //    assert(pos<=numChars);
    //  --pos;
    //  pos+=(pos/lineLength);
    // return pStart[pos];
    return(baseNames[getNextBase((uint)pos)]);
  } // ~operator[]

  int getNumChars( void ) const { return numChars; }
private:
  void *pMap;
  const char* pStart;
  const char* pEnd;
  const Word* pWord;
  size_t fileSize;
  int lineLength;
  int numChars;
}; // ~class SquashFile

SquashFile::SquashFile( const char* name )
{
  int fd;
  // memory map sequence file
  fd = open(name, O_RDONLY, S_IRUSR );
  if (fd==-1)
  {
    cerr << "Error: could not open file " 
         << name << endl;
    exit (1);
  } // ~if
  fileSize = lseek(fd, 0, 2);

  pMap=mmap(0, fileSize, PROT_READ, MAP_SHARED, fd, 0);
  if (pMap==NULL)
  {
    cerr << "Error: could not memory map file " 
         << name << endl;
    exit (1);
  } // ~if
  close (fd);

  pStart=(char*)pMap;
  pEnd=pStart+fileSize;

  if (*pStart=='>') 
  {
    while (*pStart!='\n') ++pStart;
    ++pStart;
  } // ~if
  pWord = (const Word*) pStart;

} // ~ctor

struct FragmentFinder
{
 
  FragmentFinder ( const string& squashDirName,
		   const int readLength,
		   const int reverseStrandStartOffset ) :
  squashDirName_(squashDirName),
  readLength_(readLength),
  reverseStrandStartOffset_(reverseStrandStartOffset),
  pSquash_(NULL)
  {} // ~c'tor
  virtual ~FragmentFinder()
  {
    delete pSquash_;
  } // ~d'tor
  void operator()
  ( vector<SeqRequest>& requests, 
    const vector<char*>& reads,
    vector<char*>& frags,
    const StringIndex& files ); // TBD probably const
  virtual void fetchFragment( const SeqRequest& i,
			      const vector<char*>& reads,
			      char* buf );


  const string squashDirName_;
  const int readLength_;
  const int reverseStrandStartOffset_;
  SquashFile* pSquash_;
}; // ~struct FragmentFinder

// default (= base class) behaviour of fetchFragment is to pull
// genomic fragments from the squash files, orienting them in the 
// same direction as the read
void FragmentFinder::fetchFragment
( const SeqRequest& i, const vector<char*>& reads, char* buf )
{
      buf[readLength_]=0;
      if (i.strand_=='F')
      {
	for (int j(0);j<readLength_;j++)
	{
	  buf[j]=(*pSquash_)[i.filePos_+j-1];
	} // ~for j
      } // ~if
      else
      {
	for (int j(0);j<readLength_;j++)
	{
	  buf[readLength_-1-j]
	    =reverseCharASCII[(*pSquash_)
			      [i.filePos_+j-reverseStrandStartOffset_-1]];
	} // ~for
      } // ~else

} // ~FragmentFinder::fetchFragment


void FragmentFinder::operator()
( vector<SeqRequest>& requests, 
  const vector<char*>& reads,
  vector<char*>& frags,
  const StringIndex& files )
{
  if (!requests.empty())
  {
    //    SquashFile* pSquash(NULL);
    string squashFileName;
    int lastFile=requests[0].fileIndex_+1; // ensure false on first go
    sort( requests.begin(), requests.end(), lessThanRequest );

    for ( vector<SeqRequest>::iterator i(requests.begin());
	  i!=requests.end();++i)
    {
      if (i->fileIndex_!=lastFile)
      {
	squashFileName=squashDirName_
	  +(string)"/"
	  +files.names_[i->fileIndex_]
	  +(string)".2bpb";
	delete pSquash_;
	pSquash_=new SquashFile(squashFileName.c_str());
      } // ~if
      lastFile=i->fileIndex_;

      fetchFragment( *i, reads, frags[i->requestNum_] );

    } // ~for i

    delete pSquash_;
    pSquash_=NULL; //ensures don't delete twice if destructor called
  } // ~if

} // ~FragmentFinder::operator()

// struct FragmentFinderAlign : public FragmentFinder
// Instead of printing the genomic fragment, replace bases that match
// the read with '-' characters
struct FragmentFinderAlign : public FragmentFinder
{
 
  FragmentFinderAlign ( const string& squashDirName,
		   const int readLength,
		   const int reverseStrandStartOffset ) :
  FragmentFinder( squashDirName,
		  readLength,
		  reverseStrandStartOffset )
  {} // ~c'tor
  virtual void fetchFragment( const SeqRequest& i,
			      const vector<char*>& reads,
			      char* buf );
}; // struct ~FragmentFinderAlign

void FragmentFinderAlign::fetchFragment
( const SeqRequest& i, const vector<char*>& reads, char* buf )
{
  FragmentFinder::fetchFragment( i, reads, buf );

  for (int j(0);j<readLength_;j++)
  {
    if (toupper(reads[i.readNum_][j])==buf[j])
      buf[j]='-';
      //      frags[i->requestNum_][j]=(*pSquash_)[i->filePos_+j-1];
  } // ~for j

} // ~FragmentFinderAlign::fetchFragment


// struct FragmentFinderAlignCompact : public FragmentFinderAlign
// Replace runs of matching bases ('-' characters) with an integer,
// eg '---A-----C' becomes '3A5C'
// Perl 1 liner to convert back:
// cat s_3_frag.txt | perl -lane '{ while (/(\d+)/) { $a="-" x $1; s/$1/$a/ } print }'
struct FragmentFinderAlignCompact : public FragmentFinderAlign
{
 
  FragmentFinderAlignCompact ( const string& squashDirName,
		   const int readLength,
		   const int reverseStrandStartOffset ) :
  FragmentFinderAlign( squashDirName,
		  readLength,
		  reverseStrandStartOffset )
  {} // ~c'tor
  virtual void fetchFragment( const SeqRequest& i,
			      const vector<char*>& reads,
			      char* buf );
  char buf_[bufSize];
}; // struct ~FragmentFinderAlign

void FragmentFinderAlignCompact::fetchFragment
( const SeqRequest& i, const vector<char*>& reads, char* buf )
{
  FragmentFinderAlign::fetchFragment( i, reads, buf_ );
  int exactRun(0);
  char* p(buf);
  char c;
  for (int j(0);j<readLength_;j++)
  {
    if (buf_[j]=='-') 
    {
      exactRun++;
    } // ~if
    else 
    {
      c=buf_[j];
      if (exactRun!=0) 
      { 
	sprintf(p,"%u%c",exactRun,c);
	exactRun=0;
      } // ~if
      else 
      {
	sprintf(p,"%c",c);
      } // ~else
      p+=strlen(p);
    } // ~else
  } // ~for j
  if (exactRun!=0) sprintf(p,"%u",exactRun);
} // ~FragmentFinderAlign::fetchFragment




struct ResultsParser
{
  virtual void operator()( const char* buf, 
			   StringIndex& files,
			   vector<char*>& reads, 
			   vector<SeqRequest>& requests ) const=0;
}; // ~struct ResultsParser

struct ResultsParserStandard : public ResultsParser
{
  virtual void operator()( const char* buf, 
			   StringIndex& files,
			   vector<char*>& reads, 
			   vector<SeqRequest>& requests ) const
  {
    char chromName[bufSize];
    char readBuf[bufSize];
    uint chromNum;
    uint chromPos;
    uint readLength;
    char strand;    
    if (sscanf(buf, "%*s\t%s\t%*s\t%*s\t%*s\t%*s\t%s\t%d\t%c", 
	       readBuf, chromName, &chromPos, &strand )==4)
    {
      files.getIndex(chromName, chromNum, chromPos );

      requests.push_back
	(SeqRequest(requests.size(),
		    reads.size(),
		    chromNum,
		    chromPos,
		    strand));
      readLength=strlen(readBuf);
      reads.push_back( new char[readLength+1] );
      strcpy(reads.back(),readBuf);
      // cout << reads.back() << endl;
    } // ~if
    return;
  } // ~operator()

}; // ~struct ResultsParserStandard : public ResultsParser

struct ResultsParserMulti : public ResultsParser
{
  virtual void operator()( const char* buf, 
			   StringIndex& files,
			   vector<char*>& reads, 
			   vector<SeqRequest>& requests ) const
  {
    char chromName[bufSize];
    const char *p1, *p2;
    uint chromNum;
    uint pos;
    char strand;    

    p1=strchr(buf,'\t'); // scan past read ID
    assert(p1!=NULL);
    p1++;
    p2=strchr(p1,'\t'); // scan past read
    assert(p2!=NULL);
    uint readLength(p2-p1);
    p2++; // now first character of match type field
    if (isalpha(*p2)) return; // NM,QC,RM...always a letter if no match found

    reads.push_back( new char[readLength+1] );
    strncpy(reads.back(),p1,readLength);
    reads.back()[readLength]='\0'; // TBD needed? maybe done by strncpy?

    p1=strchr(p2,'\t'); // scan past match type
    //    cout << buf << endl;
    if (p1==NULL) return; // too many repeats for matches to be stored

    chromName[0]='\0';
    while (1)
    {
      p1++;
      p2=strpbrk(p1,":,");
      if ((p2==NULL)||(*p2==','))
      {
	assert (sscanf(p1,"%u%c%*c", &pos, &strand)==2);
	assert ((strand=='F')||(strand=='R'));
	assert (chromName[0]!='\0');
	files.getIndex(chromName, chromNum, pos );
	requests.push_back
	  (SeqRequest(requests.size(),
		      reads.size()-1,
		      chromNum,
		      pos,
		      strand));
	if (p2==NULL) return;
      } // ~if
      else
      { // must be a colon, therefore it's a chrom name
	strncpy(chromName,p1,p2-p1);
	chromName[p2-p1]='\0';
      } // ~else
      p1=p2;
    } // ~while

  }   // ~operator()

}; // ~struct ResultsParserMulti : public ResultsParser


void printFragments( const vector<char*>& frags,
		     const vector<int>& requestsPerLine )
{
      vector<char*>::const_iterator j(frags.begin());
      for ( vector<int>::const_iterator i(requestsPerLine.begin());
	    i!=requestsPerLine.end(); ++i)
      {
	for (int k(0); k<*i; k++)
	{  
	  if (k!=0) printf(":");
	  assert(j!=frags.end());
	  printf("%s",*j++);
	} // ~for k
	printf("\n");
      } // ~for i

} // ~void printFragments


void usage( const char** args )
{
    cerr << "Usage: cat elandOutput.txt | " << args[0] 
	 << " squashedFileDir readLength [readLengthUsedForELAND] [--multi]" 
	 << " [--read|--align|--compact]" 
	 << endl;
    exit (1);
}

int main( int numArgs, const char** args )
{
  

  if (numArgs<3) usage(args);

  int thisArg(1);

  string squashDir(args[thisArg++]);
  uint readLength(atoi(args[thisArg++]));
  assert(readLength>0);

  ResultsParser* parseResults(NULL);
  FragmentFinder* getFragments(NULL);

  int readLengthELAND;
  if ((numArgs>thisArg)&&(atoi(args[thisArg])!=0))
  {
    readLengthELAND=atoi(args[thisArg++]);
  } // ~if
  else
  { // make a sensible guess
    readLengthELAND=(readLength>maxReadLengthELAND)
      ?maxReadLengthELAND:readLength;
  } // ~else

  // In the event of a reverse complement match need to grab bases that
  // come before the read in the forward strand
  const int reverseStrandStartOffset(readLength-readLengthELAND);

  if ((numArgs>thisArg)&&(strcmp(args[thisArg],"--multi")==0))
  {
    parseResults=new ResultsParserMulti();
    thisArg++;
  } // ~if

  if (numArgs>thisArg)
  {
    if (strcmp(args[thisArg],"--read")==0)
    {
      getFragments
	=new FragmentFinder( squashDir, 
			     readLength, 
			     reverseStrandStartOffset);
      thisArg++;
    } // ~if
    else if (strcmp(args[thisArg],"--align")==0)
    {
      getFragments
	=new FragmentFinderAlign( squashDir, 
			     readLength, 
			     reverseStrandStartOffset);
      thisArg++;
    } // ~if
    else if (strcmp(args[thisArg],"--compact")==0)
    {
      getFragments
	=new FragmentFinderAlignCompact( squashDir, 
					 readLength, 
					 reverseStrandStartOffset);
      thisArg++;
    } // ~else if
    else usage (args);
  } // ~if

  if (numArgs>thisArg) usage(args);

  if (parseResults==NULL) parseResults=new ResultsParserStandard();

  if (getFragments==NULL) 
    getFragments =new FragmentFinder( squashDir, 
				      readLength, 
				      reverseStrandStartOffset);


  // set up request buffer
  vector<SeqRequest> requests;
  requests.reserve(batchSize);
  // set up fragment buffer
  vector<char*> frags(batchSize);
  for (vector<char*>::iterator i(frags.begin());i!=frags.end();i++)
  {
    *i=new char[readLength+1];
  } // ~for
  vector<char*> reads;

  //  FragmentFinder* getFragments
  //  =new FragmentFinder( squashDir, 
  //		 readLength, 
  //		 reverseStrandStartOffset);

  StringIndex files(squashDir);

  char buf[bufSize];
  int prevSize;
  vector<int> requestsPerLine;

  while( fgets(buf, bufSize, stdin )!=NULL)
  {
    prevSize=requests.size();
    (*parseResults)( buf, files, reads, requests );
    requestsPerLine.push_back(requests.size()-prevSize);

    if (requests.size()>=batchSize)
    {
      // increase size of frag buffer slightly if necessary
      while (frags.size()<requests.size())
      {
	frags.push_back(new char[readLength+1]);
      }	// ~while

      (*getFragments)(requests, reads, frags, files);
      printFragments(frags,requestsPerLine);
      requests.clear();
      requestsPerLine.clear();
      for (vector<char*>::iterator i(reads.begin());i!=reads.end();i++)
      {
	delete [] *i;
      } // ~for
      reads.clear();
      //      requestNum=0;
    } // ~if
  } // ~while
  
  (*getFragments)(requests, reads, frags, files);
  printFragments(frags,requestsPerLine);

  // deallocate memory
  for (vector<char*>::iterator i(frags.begin());i!=frags.end();i++)
  {
    delete [] *i;
  } // ~for
  for (vector<char*>::iterator i(reads.begin());i!=reads.end();i++)
  {
    delete [] *i;
  } // ~for
  delete parseResults;
} // ~main

