/*
 * Higher level routines used as wrappers for different FFT implementations.
 * High-performance FFT libraries come with different data layouts and
 * optimisations to achieve optimal performance. These routines are used to
 * provide a uniform interface to the FFT routines. Only single-precision
 * routines are provided.
 *
 * Currently 4 FFT libraries are supported:
 *  - "fftn":  FFT routines for aribtrary dimensions
 *             License: GPL, http://www.jjj.de/fft/fftpage.html 
 *             not very fast, but easily portable and thus included as a 
 *             reference 
 *  - FFTW:    FFT routines for arbitrary dimensions
 *             License: GPL, http://www.fftw.org
 *             extremely fast over a large range of architectures
 *  - IMKL:    Intel Math Kernel Library
 *             fast, accessed through the FFTW3 wrapper provided by Intel
 *  - ESSL:    IBM Engineering and Scientific libraries, contains
 *             FFT routines of dimensions 1-3, lengths have to be such
 *             they can be factorised by 2, 3, 5, 7, 11
 *             optimised for IBM's Power architecture
 *
 * The different transform backends use different representation for the
 * Fourier data; these are cast to a single-precision array if they
 * are made accessible to the caller.
 * The phase centre of the transform (the DC component) is in the 
 * (0,0) array element.
 *             
 * Copyright (c) Solexa 2003, 2004, 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

//---------------------------------------------------------------------------

#if defined(__BORLANDC__)
#include <string.h>
#else
#include <cstring>
#endif
#include <cstdio>
#include <cmath>

#if defined(HAVE_FFTN)
#include "fftn.h"
#endif
#if defined(HAVE_FFTW)
#include <fftw3.h>
#endif
#if defined(HAVE_IMKL)
#include "fftw3.h"
#endif
#if defined(HAVE_ESSL)
#include <complex>
#include <essl.h> 
#endif

#define MAX(x, y) ((x) > (y) ? (x) : (y))
#define MIN(x, y) ((x) < (y) ? (x) : (y))

//---------------------------------------------------------------------------

/*
 * Arrays for Fourier coefficient need to be allocated through this wrapper.
 * The allocation routines ensure that the memory is aligned properly 
 * to improve FFT performance (depending on the platform, usually on 
 * a double word (8-byte) boundary). In fact, without alignment the FFT
 * routines may fail to work properly.  
 */
float *new_fourier_array(int nx, int ny)
{
#if defined(HAVE_ESSL)
    return new float [2*nx*(ny/2+1)];
#elif defined(HAVE_FFTW)
    fftwf_complex *a = (fftwf_complex *)fftwf_malloc(sizeof(fftwf_complex) 
                                                     * nx * (ny/2+1));
    return (float *)a;
#elif defined(HAVE_IMKL)
    /* the MKL wrappers do not implement the memory alignment; they 
     * recommend implementing it directly. TODO: implement this.
     * current version only tested on Linux, where it works without
     * memory alignment.
     */
    return new float [2*nx*(ny/2+1)];
#elif defined(HAVE_FFTN)
    return new float [2*nx*ny];
#endif
}

void copy_fourier_array(float *src, float *dest, int nx, int ny)
{
#if defined(HAVE_ESSL)
    memcpy(dest, src, 2*nx*(ny/2+1)*sizeof(float));
#elif defined(HAVE_FFTW)  || defined(HAVE_IMKL)
    memcpy(dest, src, nx * (ny/2+1)*sizeof(fftwf_complex));
#elif defined(HAVE_FFTN)
    memcpy(dest, src, 2*nx*ny*sizeof(float));
#endif
}

/*
 * clear up aligned memory for FFT data
 */
void delete_fourier_array(float *a)
{
#if defined(HAVE_ESSL)
    delete [] a;
#elif defined(HAVE_FFTW)
    fftwf_free((fftwf_complex *)a);
#elif defined(HAVE_IMKL)
    delete [] a;
#elif defined(HAVE_FFTN)
    delete [] a;
#endif
}


#if defined(HAVE_FFTW) || defined(HAVE_IMKL)

/*
 * convert FFTW complex format to separate real and imaginary images
 */
void convert_fft2realimag_fftw(float *fft, int nx, int ny, 
                               float *real, float *imag, int halfplane)
{
    fftwf_complex *fftw = (fftwf_complex *)fft;
    int ny2 = ny/2+1;
    int i, j;
    for (i = 0; i < nx; i++) {
        for (j = 0; j < ny2; j++) {
            int iw = i*ny2+j;
            int ir = i*ny+j;
            real[ir] = fftw[iw][0]; imag[ir] = fftw[iw][1];
        }
    }
    /* reconstruct redundant part of Fourier plane that is not stored:
     * for a real image, the Fourier transform is Hermitian
     */
    if (halfplane)
        return;
    for (i = 1; i < nx/2; i++)
        for (j = 1; j < ny/2; j++) {
            int iw = i*ny2+j;
            int ii = (nx-i)*ny+(ny-j);
            real[ii] = fftw[iw][0]; imag[ii] = -fftw[iw][1];
        }
    for (i = 1; i < nx/2; i++)
        for (j = 1; j < ny/2; j++) {
            int iw = (nx-i)*ny2+j;
            int ii = i*ny+(ny-j);
            real[ii] = fftw[iw][0]; imag[ii] = -fftw[iw][1];
        }
}

/*
 * convert real and imaginary images to FFTW's complex format
 */
void convert_realimag2fft_fftw(float *real, float *imag, int nx, int ny,
                               float *fft)
{
    fftwf_complex *fftw = (fftwf_complex *)fft;
    int ny2 = ny/2+1;
    int i, j;
    for (i = 0; i < nx; i++) {
        for (j = 0; j < ny2; j++) {
            int iw = i*ny2+j;
            int ir = i*ny+j;
            fftw[iw][0] = real[ir]; fftw[iw][1] = imag[ir];
        }
    }
}

/*
 * FFTW real to complex (forward) transform
 */
void fft_2d_r2c_fftw(float *in, float *out, int nx, int ny)
{
    fftwf_plan plan;
    fftwf_complex *fftw_out = (fftwf_complex *)out;
    plan = fftwf_plan_dft_r2c_2d(nx, ny, in, fftw_out, FFTW_ESTIMATE);
    fftwf_execute(plan);
    fftwf_destroy_plan(plan);
}

/*
 * FFTW complex to real (backward) transform
 */
void fft_2d_c2r_fftw(float *in, float *out, int nx, int ny)
{
    fftwf_plan plan;
    fftwf_complex *fftw_in = (fftwf_complex *)in;
    plan = fftwf_plan_dft_c2r_2d(nx, ny, fftw_in, out, FFTW_ESTIMATE);
    fftwf_execute(plan);
    fftwf_destroy_plan(plan);
}

/*
 * multiply a complex image in FFTW format with a real image 
 */
void multiply_fft_real_fftw(float *fft, float *mask, int nx, int ny)
{
    fftwf_complex *fftw_in = (fftwf_complex *)fft;
    int nt = nx*ny;
    int ny2 = ny/2+1;
    int xi;
    float val;
    int x, y;
    for (x = 0; x < nx; x++) {
        xi = x*ny2;
        for (y = 0; y < ny2; y++) {
            val = mask[x*ny+y]/nt;
            fftw_in[xi+y][0] *= val;
            fftw_in[xi+y][1] *= val;
        }
    } 
}

/*
 * multiply two complex images in FFTW format
 */
void  multiply_fft_fft_fftw(float *image1, float *image2, float *result, 
                            int nx, int ny)
{
    float re1, im1, re2, im2;
    fftwf_complex *fftw_image1 = (fftwf_complex *)image1; 
    fftwf_complex *fftw_image2 = (fftwf_complex *)image2; 
    fftwf_complex *fftw_dummy = (fftwf_complex *)result;
    int i; 
    for (i = 0; i < nx*(ny/2+1); i++) {
        re1 = fftw_image1[i][0]; im1 = fftw_image1[i][1];
        re2 = fftw_image2[i][0]; im2 = fftw_image2[i][1];
        fftw_dummy[i][0] = re1*re2+im1*im2;
        fftw_dummy[i][1] = re2*im1-re1*im2;
    }
}
#endif // defined(HAVE_FFTW)

#if defined(HAVE_FFTN)

/*
 * convert a complex image into two images signifying real and imaginary parts
 */
void convert_fft2realimag_fftn(float *fft, int nx, int ny, 
                               float *real, float *imag, int halfplane)
{
    int i;
    for (i = 0; i < nx*ny; i++) {
        real[i] = fft[2*i];
        imag[i] = fft[2*i+1];
    }
}

/*
 * convert two images signifying real and imaginary parts
 * into a complex image
 */
void convert_realimag2fft_fftn(float *real, float *imag, int nx, int ny,
                               float *fft)
{
    int i;
    for (i = 0; i < nx*ny; i++) {
        fft[2*i]   = real[i];
        fft[2*i+1] = imag[i];
    }
}

/*
 * FFTN real-to-complex (forward) transform
 */
void fft_2d_r2c_fftn(float *in, float *out, int nx, int ny)
{
    int ndim = 2;
    int dims[2];
    double scaling = 1.;
    float *re = new float [nx*ny];
    float *im = new float [nx*ny];

    int i;
    for (i = 0; i < nx*ny; i++) {
        re[i] = in[i];
        im[i] = 0.;
    }
    // empirically, the x- and y-dimensions seem to be swapped.
    // perhaps some remnant from fftn's Fortran heritage??
    dims[0] = ny; dims[1] = nx;
    fftnf(ndim, dims, re, im, -1, scaling);
    convert_realimag2fft_fftn(re, im, nx, ny, out);
    delete [] re; delete [] im;
}

/*
 * FFTN complex-to-real (backward) transform
 */
void fft_2d_c2r_fftn(float *in, float *out, int nx, int ny)
{
    int ndim = 2;
    int dims[2];
    double scaling = 1.;
    float *re = new float [nx*ny];
    float *im = new float [nx*ny];

    convert_fft2realimag_fftn(in, nx, ny, re, im, 0);
    // empirically, the x- and y-dimensions seem to be swapped.
    // perhaps some relic from fftn's Fortran heritage??
    dims[0] = ny; dims[1] = nx;
    fftnf(ndim, dims, re, im, 1, scaling);
    int i;
    for (i = 0; i < nx*ny; i++) {
        out[i] = re[i];
    }
    delete [] re;
    delete [] im;
}

/*
 * multiply a complex image by a real image
 */
void multiply_fft_real_fftn(float *fft, float *mask, int nx, int ny)
{
    int nt = nx*ny;
    float val;
    int j;
    for (j = 0; j < nx * ny; j++) {
        val = mask[j]/nt;
        fft[2*j+0] *= val;
        fft[2*j+1] *= val;
    }
}

/*
 * multiply two complex images
 */
void  multiply_fft_fft_fftn(float *image1, float *image2, float *result, 
                            int nx, int ny)
{
    int i;
    float re1, im1, re2, im2;
    for (i = 0; i < nx*ny; i++) {
        re1 = image1[2*i]; im1 = image1[2*i+1];
        re2 = image2[2*i]; im2 = image2[2*i+1];
        result[2*i] = re1*re2+im1*im2;
        result[2*i+1] = re2*im1-re1*im2;
    }
}
#endif /* defined(HAVE_FFTN) */


#if defined(HAVE_ESSL)

static int essl_last_c2r_xdim = 0;
static int essl_last_c2r_ydim = 0;

/*
 * ESSL complex-to-real (backward) transform
 */
void fft_2d_c2r_essl(float *in, float *out, int nx, int ny)
{
    int n1 = nx;
    int n2 = ny;
    static int naux1 = 0;
    static int naux2 = 0;
    static double *aux1;
    static double *aux2;
    int isign = 1;
    float scale = 1.;
    int inc2x = n1/2+1;
    int inc2y = n1+2;
    complex<float> *x = new complex<float> [ inc2x * n2 ];
    float *y = new float [ inc2y * n2 ];
    int i, j;

    for (i = 0; i < inc2x; i++)
        for (j = 0; j < n2; j++)
            x[j*inc2x+i] = complex<float>(in[2*(i*n2+j)], in[2*(i*n2+j)+1]);

    // initialise Fourier transform if no previous transform of 
    // this size has been called.
    if ((nx != essl_last_c2r_xdim) || (ny != essl_last_c2r_ydim)) {
    
        // clear up memory if routine has been called with different 
        // arguments
        if ((essl_last_c2r_xdim != 0) || (essl_last_c2r_ydim != 0)) {
            delete [] aux1;
            delete [] aux2;
        }
        essl_last_c2r_xdim = nx;
        essl_last_c2r_ydim = ny;

        if (MAX(n1/2, n2) <= 8192)
            naux1 = 45000;
        else
            naux1 = 40000 + 0.82*n1 + 1.14*n2;

        if (n1 <= 16384)
            naux2 = 20000;
        else
            naux2 = 20000+0.57*n1;
        if (n2 >= 252)
            naux2 += (n2+256) * (1.14+MIN(64, 1+n1/2));

        aux1 = new double [ naux1 ];
        aux2 = new double [ naux2 ];
        scrft2(1, x, inc2x, y, inc2y, n1, n2, isign, scale,
               aux1, naux1, aux2, naux2, NULL, 0); 
    }
    scrft2(0, x, inc2x, y, inc2y, n1, n2, isign, scale,
           aux1, naux1, aux2, naux2, NULL, 0); 

    for (i = 0; i < n1; i++)
        for (j = 0; j < n2; j++)
            out[i*n2+j] = y[j*inc2y+i];

    delete [] x;
    delete [] y;
}

static int essl_last_r2c_xdim = 0;
static int essl_last_r2c_ydim = 0;

/*
 * ESSL real-to-complex (forward) transform
 */
void fft_2d_r2c_essl(float *in, float *out, int nx, int ny)
{
    int n1 = nx;
    int n2 = ny;
    static int naux1 = 0;
    static int naux2 = 0;
    static double *aux1;
    static double *aux2;
    int isign = -1;
    float scale = 1.;
    int inc2x = n1;
    int inc2y = n1/2+1;
    float *x = new float [ inc2x * n2];
    complex<float> *y = new complex<float> [ inc2y * n2 ]; 
    int i, j;

    for (i = 0; i < n1; i++)
        for (j = 0; j < n2; j++)
            x[j*inc2x+i] = in[i*n2+j];

    // initialise Fourier transform if no previous transform of 
    // this size has been called.
    if ((nx != essl_last_r2c_xdim) || (ny != essl_last_r2c_ydim)) {
    
        // clear up memory if routine has been called with different 
        // arguments
        if ((essl_last_r2c_xdim != 0) || (essl_last_r2c_ydim != 0)) {
            delete [] aux1;
            delete [] aux2;
        }
        essl_last_r2c_xdim = nx;
        essl_last_r2c_ydim = ny;

        if (MAX(n1/2, n2) <= 8192)
            naux1 = 45000;
        else
            naux1 = 40000 + 0.82*n1 + 1.14*n2;

        if (n1 <= 16384)
            naux2 = 20000;
        else
            naux2 = 20000+0.57*n1;
        if (n2 >= 252)
            naux2 += (n2+256) * (1.14+MIN(64, 1+n1/2));

        aux1 = new double [ naux1 ];
        aux2 = new double [ naux2 ];
        srcft2(1, x, inc2x, y, inc2y, n1, n2, isign, scale,
               aux1, naux1, aux2, naux2, NULL, 0); 
    }
    srcft2(0, x, inc2x, y, inc2y, n1, n2, isign, scale,
           aux1, naux1, aux2, naux2, NULL, 0); 


    for (i = 0; i < inc2y; i++)
        for (j = 0; j < n2; j++) {
            out[2*(i*n2+j)] = y[j*inc2y+i].real();
            out[2*(i*n2+j)+1] = y[j*inc2y+i].imag();
        }
    delete [] x;
    delete [] y;
}

/*
 * multiply complex image in ESSL format with real image
 */
void multiply_fft_real_essl(float *fft, float *mask, int nx, int ny)
{
    int nt = nx*ny;
    int nx2 = nx/2+1;
    int xi;
    float val;
    int i, j;
    for (i = 0; i < nx2; i++) {
        xi = i*ny;
        for (j = 0; j < ny; j++) {
            val = mask[i*ny+j]/nt;
            fft[2*(i*ny+j)] *= val;
            fft[2*(i*ny+j)+1] *= val;
        }
    } 
}

/*
 * multiply two complex images
 */
void multiply_fft_fft_essl(float *fft1, float *fft2, float *result,
                           int nx, int ny)
{
    fprintf(stderr, "multiply_fft_fft_essl not implemented.\n");
    exit(0);
}
#endif /* defined(HAVE_ESSL) */

/*
 * real-to-complex transform. the semantics are based on
 * FFTW's real-to-complex routines. 
 * no rescaling of coefficients is being done.
 * the output array holds both real and imaginary parts of the complex
 * coefficients (in alternating order).
 */ 
void fft_2d_r2c(float *in, float *out, int nx, int ny)
{
#if defined(HAVE_ESSL)
    fft_2d_r2c_essl(in, out, nx, ny);
#elif defined(HAVE_FFTW) || defined(HAVE_IMKL)
    fft_2d_r2c_fftw(in, out, nx, ny);
#elif defined(HAVE_FFTN)
    fft_2d_r2c_fftn(in, out, nx, ny);
#endif
}

/*
 * complex-to-real transform. the semantics are based on
 * FFTW's complex-to-real routines. 
 * no rescaling of coefficients is being done.
 * the input array holds both real and imaginary parts of the complex
 * coefficients (in alternating order).
 */ 
void fft_2d_c2r(float *in, float *out, int nx, int ny)
{
#if defined(HAVE_ESSL)
    fft_2d_c2r_essl(in, out, nx, ny);
#elif defined(HAVE_FFTW) || defined(HAVE_IMKL)
    fft_2d_c2r_fftw(in, out, nx, ny);
#elif defined(HAVE_FFTN)
    fft_2d_c2r_fftn(in, out, nx, ny);
#endif
}

/*
 * complex multiplication A B^*
 */
void multiply_fft_fft(float *fft1, float *fft2, float *result, int nx, int ny)
{
#if defined(HAVE_ESSL)
    multiply_fft_fft_essl(fft1, fft2, result, nx, ny);
#elif defined(HAVE_FFTW) || defined(HAVE_IMKL)
    multiply_fft_fft_fftw(fft1, fft2, result, nx, ny);
#elif defined(HAVE_FFTN)
    multiply_fft_fft_fftn(fft1, fft2, result, nx, ny);
#endif
}

/*
 * multiply complex array with a real array
 */
void multiply_fft_real(float *fft, float *mask, int nx, int ny)
{
#if defined(HAVE_ESSL)
    multiply_fft_real_essl(fft, mask, nx, ny);
#elif defined(HAVE_FFTW) || defined(HAVE_IMKL)
    multiply_fft_real_fftw(fft, mask, nx, ny);
#elif defined(HAVE_FFTN)
    multiply_fft_real_fftn(fft, mask, nx, ny);
#endif
}

/*
 * unpack the FFT storage format into real and imaginary arrays
 */
void convert_fft2realimag(float *fft, int nx, int ny, 
                          float *real, float *imag, int halfplane)
{
#if defined(HAVE_FFTW) || defined(HAVE_IMKL)
    convert_fft2realimag_fftw(fft, nx, ny, real, imag, halfplane);
#elif defined(HAVE_FFTN)
    convert_fft2realimag_fftn(fft, nx, ny, real, imag, halfplane);
#endif
}

/*
 * pack real and imaginary arrays into the relevant FFT storage format
 */
void convert_realimag2fft(float *real, float *imag, int nx, int ny,
                          float *fft)
                          
{
#if defined(HAVE_FFTW) || defined(HAVE_IMKL)
    convert_realimag2fft_fftw(real, imag, nx, ny, fft);
#elif defined(HAVE_FFTN)
    convert_realimag2fft_fftn(real, imag, nx, ny, fft);
#endif
}

/*
 * Given a the Fourier transform of an image, apply a filter mask
 * that is also given in Fourier space.
 * In other words, simpy perform a multiplication of two complex images
 * and transform the result back into real space.
 */
void apply_fft_mask(float *in, float *out, float *mask, int nx, int ny)
{
    // multiply the complex image by a real mask
    multiply_fft_real(in, mask, nx, ny);

    // transform result back to real plane
    fft_2d_c2r(in, out, nx, ny);
}

/*
 * Apply a filter function <mask> given in Fourier space to a real image.
 * This is essentially a fast way of applying an arbitrary filter function.
 * The filter is given by its Fourier transform rather than the real-space
 * filter mask to allow the caller to avoid redundant transforms.
 */
void filter_fft_mask(float *real, float *out, float *mask, int nx, int ny)
{
    // FFT the image
    float *fourier = new_fourier_array(nx, ny);
    fft_2d_r2c(real, fourier, nx, ny);
    // apply the mask
    apply_fft_mask(fourier, out, mask, nx, ny);

    delete_fourier_array(fourier);
}

/*
 * Perform cross-correlation of two images that have been Fourier transformed.
 * (The input images are given in the Fourier domain.)
 */
void correlate_fft_images(float *image1, float *image2, int nx, int ny,
                          float *correlation)
{
    float *dummy = new_fourier_array(nx,ny);

    // perform complex multiplications A B^*
    multiply_fft_fft(image1, image2, dummy, nx, ny);

    // transform back to pixel space
    fft_2d_c2r(dummy, correlation, nx, ny);

    delete_fourier_array(dummy);
}

/*
 * Perform cross-correlation of two real images by FFT.
 */
void correlate_images(float *image1, float *image2, int nx, int ny,
                      float *correlation)
{
    float *fourier1 = new_fourier_array(nx, ny);
    float *fourier2 = new_fourier_array(nx, ny);
    fft_2d_r2c(image1, fourier1, nx, ny);
    fft_2d_r2c(image2, fourier2, nx, ny);
    correlate_fft_images(fourier1, fourier2, nx, ny, correlation);
    delete_fourier_array(fourier1);
    delete_fourier_array(fourier2);
}

/*
 * The speed of the FFT depends crucially on the number of data samples
 * in each dimension. Lengths that can be factorised into small prime 
 * numbers work best. 
 * Furthermore, the IBM ESSL FFT routines only allow certain values
 * for FFT lengths. This function provides a lookup table that finds the
 * next largest optimal size for a given transform length. There are
 * some intermediate values that factorise into low primes, but to keep
 * things simple this list is restricted to the values allowed by ESSL.
 */  
int get_optimal_fft_length(int n)
{
    int ndim = 165;
    int dim[165] = {
        2, 4, 6, 8, 10, 12, 14, 16, 18,
       20,    22,    24,    28,    30,    32,    36,    40,    42,    44,
       48,    56,    60,    64,    66,    70,    72,    80,    84,    88,
       90,    96,   110,   112,   120,   126,   128,   132,   140,   144,
      154,   160,   168,   176,   180,   192,   198,   210,   220,   224,
      240,   252,   256,   264,   280,   288,   308,   320,   330,   336,
      352,   360,   384,   396,   420,   440,   448,   462,   480,   504,
      512,   528,   560,   576,   616,   630,   640,   660,   672,   704,
      720,   768,   770,   792,   840,   880,   896,   924,   960,   990,
     1008,  1024,  1056,  1120,  1152,  1232,  1260,  1280,  1320,  1344,
     1386,  1408,  1440,  1536,  1540,  1584,  1680,  1760,  1792,  1848,
     1920,  1980,  2016,  2048,  2112,  2240,  2304,  2310,  2464,  2520,
     2560,  2640,  2688,  2772,  2816,  2880,  3072,  3080,  3168,  3360,
     3520,  3584,  3696,  3840,  3960,  4032,  4096,  4224,  4480,  4608,
     4620,  4928,  5040,  5120,  5280,  5376,  5544,  5632,  5760,  6144,
     6160,  6336,  6720,  6930,  7040,  7168,  7392,  7680,  7920,  8064,
     8192,  8448,  8960,  9216,  9240,  9856 };
    /*
    10080 10240 10560 10752 11088 11264 11520 12288 12320 12672 13440 13860 14080 14336 14784 15360 15840 16128 16384 16896 17920 18432 18480 19712 20160 20480 21120 21504 22176 22528 23040 24576 24640 25344 26880 27720 28160 28672 29568 30720 31680 32256 32768 33792 35840 36864 36960 39424 40320 40960 42240 43008 44352 45056 46080 49152 49280 50688 53760 55440 56320 57344 59136 61440 63360 64512 65536 67584 71680 73728 73920 78848 80640 81920 84480 86016 88704 90112 92160 98304 98560
    */

    if (n <= dim[ndim-1]) {
        int i = 0;
        while ((i < ndim) && (n > dim[i])) {
            i++;
        }
        return dim[i];
    }
    else 
        return n;
}
