/*  Time-stamp: <26 Jun 2006 16:51 BST ImageUtil.cpp km> */
/*
 * General-purpose image utility functions.
 *
 * Copyright (c) Solexa 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <string.h>
#include <stdio.h>
#else
#include <cstring>
#include <cstdio>
#endif
#include <vector>
#include "Maths.h"
#include "Stats.h"
#include "ImageUtil.h"

//---------------------------------------------------------------------------
// pixel conversion routines
//---------------------------------------------------------------------------
st_fposition *fposition_from_fnoisepixel(st_fnoisepixel *pixel, int n)
{
    int i;
    st_fposition *pos = new st_fposition [n];
    for (i = 0; i < n; i++) {
        pos[i].x = (float)pixel[i].x;
        pos[i].y = (float)pixel[i].y;
    }
    return pos;
}

st_fposition *fposition_from_pixel(st_pixel *pixel, int n)
{
    int i;
    st_fposition *pos = new st_fposition [n];
    for (i = 0; i < n; i++) {
        pos[i].x = (float)pixel[i].x;
        pos[i].y = (float)pixel[i].y;
    }
    return pos;
}

/*
 * compatibility mode: conversion routines from floating point to
 * unsigned short pixels
 */
st_pixel *pixel_from_fposition(st_fposition *pos, int n)
{
    int i;
    st_pixel *pixel = new st_pixel [n];
    for (i = 0; i < n; i++) {
        pixel[i].x = (unsigned int)(pos[i].x+0.5);
        pixel[i].y = (unsigned int)(pos[i].y+0.5);
        pixel[i].value = 1;
    }
    return pixel;
}

st_pixel *pixel_from_fnoisepixel(st_fnoisepixel *fpos, int npos)
{
    int i;
    st_pixel *position = new st_pixel [npos];
    for (i = 0; i < npos; i++) {
        position[i].x = (unsigned short)(fpos[i].x + 0.5);
        position[i].y = (unsigned short)(fpos[i].y + 0.5);
        position[i].value = (unsigned short)(fpos[i].value + 0.5);
    }
    return position;
}

st_fnoisepixel *fnoisepixel_from_pixel(st_pixel *pixel, int n, 
                                       float *noise_image, int nx, int ny)
{
    int i;
    st_fnoisepixel *pos = new st_fnoisepixel [n];
    for (i = 0; i < n; i++) {
        pos[i].x = (float)pixel[i].x;
        pos[i].y = (float)pixel[i].y;
        pos[i].value = (float)pixel[i].value;
        pos[i].noise = 0.;
    }
    if (noise_image && (nx > 0) && (ny > 0)) {
        for (i = 0; i < n; i++) {
            pos[i].noise = noise_image[pixel[i].x * ny + pixel[i].y];
        }
    }
    return pos;
}

st_fnoisepixel *fnoisepixel_from_position(st_fposition *pixel, int n,
                                          float *image, 
                                          float *noise_image, int nx, int ny)
{
    int i;
    st_fnoisepixel *pos = new st_fnoisepixel [n];
    for (i = 0; i < n; i++) {
        pos[i].x = (float)pixel[i].x;
        pos[i].y = (float)pixel[i].y;
        pos[i].value = 0.;
        pos[i].noise = 0.;
    }
    if (image && (nx > 0) && (ny > 0)) {
        for (i = 0; i < n; i++) {
            pos[i].value = noise_image[round(pixel[i].x) * ny + round(pixel[i].y)];
        }
    }
    if (noise_image && (nx > 0) && (ny > 0)) {
        for (i = 0; i < n; i++) {
            pos[i].noise = noise_image[round(pixel[i].x) * ny + round(pixel[i].y)];
        }
    }
    return pos;
}

/*
 * auxiliary function to dump an array of positions to disk
 */
int write_positions(const char filename[], st_fposition *positions, int n)
{
    FILE *posfile = fopen(filename, "wt");
    if (!posfile) {
        fprintf(stderr, "Failed to open position file %s\n", filename);
        return 1;
    }
    int code = 0;
    int x;
    for (x = 0; x < n; x++)
        if (fprintf(posfile, "%7.2f %7.2f\n", positions[x].x, positions[x].y)
            < 15) {
            fprintf(stderr, "Failed to write to %s\n", filename);
            code = 2;
        }
            
    if (fclose(posfile)) {
        fprintf(stderr, "Failed to close position file %s\n", filename);
        return 3;
    }
    return code;
}

/*
 * auxiliary function to read an array of positions from disk.
 * if <nmax> is negative, count the number of lines in the file
 */
int read_positions(const char filename[], st_fposition *positions, int nmax)
{
    int nread = 0;
    FILE *posfile = fopen(filename, "r");
    if (posfile) {
        char buffer[200+1];
        while (fgets(buffer, 200, posfile)) {
            float x, y;
            if (buffer[0] == '#') // discard lines marked as comments
                continue;
            if (sscanf(buffer, "%f %f", &x, &y) == 2) {
                if (nmax > 0) { // read positions from file
                    if (nread < nmax) {
                        positions[nread].x = x;
                        positions[nread].y = y;
                        nread++;
                    }
                }
                else {         // count number of lines in file
                    nread++;
                }
            }
        }
        fclose(posfile);
        return nread;
    }
    return -1;
}

int read_positions(const char filename[], st_fposition **positions, int *npos)
{
    int nread = 0;
    *npos  = 0;
    std::vector<st_fposition> values;
    values.reserve(MAX_OBJECTS);
    FILE *posfile = fopen(filename, "r");
    if (posfile) {
        char buffer[200+1];
        while (fgets(buffer, 200, posfile)) {
            st_fposition position;
            if (buffer[0] == '#') // discard lines marked as comments
                continue;
            if (sscanf(buffer, "%f %f", &(position.x), &(position.y)) == 2) {
                nread++;
                values.push_back(position);
            }
            else {
                fprintf(stderr, "Error: Bad line in %s\n", filename);
                return 2;
            }
        }
        fclose(posfile);
        *positions = new st_fposition [ nread ];
        *npos = nread;
        memcpy(*positions, &(values[0]), nread * sizeof(st_fposition));
        return 0;
    }
    else {
        fprintf(stderr, "Error: Failed to open %s\n", filename);
        return 1;
    }
    return 0;
}

void transpose_ushort_image(unsigned short *image, 
                            unsigned short *transpose, int nx, int ny)
{
    int x, y;
    for (x = 0; x < nx; x++)
        for (y = 0; y < ny; y++)
            transpose[x*ny+y] = image[y*nx+x];
}

//---------------------------------------------------------------------------
// image conversion routines
//---------------------------------------------------------------------------

/*
 * take the raw image data from the camera (given as a character array
 * with two bytes per pixel, little-endian, and convert to a 2-d array
 * of unsigned shorts
 */
void convert_image_uchar2ushort(unsigned char cameradata[],
                                unsigned short image[],
                                int nx, int ny, int transpose)
{
    unsigned short *camera = (unsigned short *)cameradata;

    if (transpose)
        transpose_ushort_image(camera, image, nx, ny);
    else
        memcpy(image, camera, sizeof(unsigned short)*nx*ny);
}

/*
 * simulate a character array holding image pixels as produced by a camera
 * from a 2-d unsigned short image
 */
void convert_image_ushort2uchar(unsigned char cameradata[],
                                unsigned short image[],
                                int nx, int ny, int transpose)
{
    unsigned short *camera = (unsigned short *)cameradata;

    if (transpose)
        transpose_ushort_image(image, camera, nx, ny);
    else
        memcpy(camera, image, sizeof(unsigned short)*nx*ny);
}

void convert_image_ushort2float(unsigned short image[], float *fimage,
                                int nx, int ny)
{
    int i;
    for (i = 0; i < nx*ny;  i++)
        fimage[i] = (float)image[i];
}



//---------------------------------------------------------------------------
// determine optimal dimensions of image subregions
//---------------------------------------------------------------------------
/*
 * find smallest image size greater or equal nmin that can be 
 * obtained from <ntotal> by a division by a power of 2
 */
int get_local_image_size(int ntotal, int nmin)
{
    int size = ntotal;
    while (size/2 >= nmin)
        size /= 2;
    return size;
}

//---------------------------------------------------------------------------
// routines for saturated pixels
//---------------------------------------------------------------------------
/*
 * return image saturation
 */
float get_saturation(float *array, int n, unsigned short nbits)
{
    double sat = 0.;
    int i;
    float max_val = (float)((1<<nbits)-1);
    for (i = 0; i < n; i++)
        sat += array[i];
    sat /= n;
    return 100. * (sat / max_val);
}

float get_saturation(unsigned short *array, int n, unsigned short nbits)
{
    int i;
    float *farray = new float [ n ];
    for (i = 0; i < n; i++)
        farray[i] = array[i];
    float sat = get_saturation(farray, n, nbits);
    delete [] farray; 
    return sat;
}

float get_saturation_fraction(float *array, int n, unsigned short nbits)
{
    int i;
    int count = 0;
    float max_val = (float)((1<<nbits)-1);
    for (i = 0; i < n; i++)
        if (array[i] >= max_val)
            count++;
    return ((float)count)/n;
}

/*
 * remove saturated pixels by replacing them with an average of the adjacent
 * pixels
 */
void remove_saturated_pixels(unsigned short *array, int nx, int ny,
                             unsigned short nbits)
{
    int x, y;
    unsigned short saturated = 1<<nbits-1;

    for (x = 0; x < nx; x++)
        for (y = 0; y < ny; y++)
            if (array[x*ny+y] == saturated) {
                long sum = 0, num = 0;
                int xmin, xmax, ymin, ymax;
                int single_pixel = 1;
                xmin = MAX(0, x-1); xmax = MIN(nx-1, x+1);
                ymin = MAX(0, y-1); ymax = MIN(ny-1, y+1);
                int x2, y2;
                for (x2 = xmin; x2 <= xmax; x2++)
                    for (y2 = ymin; y2 <= ymax; y2++)
                        if ((x2 != x) || (y2 != y)) {
                            if (array[x2*ny+y2] != saturated) {
                                sum += array[x2*ny+y2];
                                num++;
                            }
                            else
                                single_pixel = 0;
                        }
                        if ((single_pixel) && (num > 0))
                            array[x*ny+y] = sum/num;
            }
}

//---------------------------------------------------------------------------
// operations on images and subregions
//---------------------------------------------------------------------------
void copy_image_region(float *image, int nx, int ny,
                       float *region, int sx, int sy, int x, int y)
{
    int i, j;
    for (i = 0; i < sx; i++) {
        int idx = (i+x)*ny+y;
        for (j = 0; j < sy; j++) {
            region[i*sy+j] = image[idx+j];
        }
    }
}
                       
/*
 * add pixel mask to image region at image position (xpos, ypos)
 */ 
void add_image_region(float *image, int nx, int ny,
                      float *region, int sx, int sy, 
                      int xpos, int ypos, float amp)
{
    int x, y;
    int xmin = MAX(0, xpos); int xmax = MIN(nx, xpos+sx);
    int ymin = MAX(0, ypos); int ymax = MIN(ny, ypos+sy);
    for (x = xmin; x < xmax; x++)
        for (y = ymin; y < ymax; y++)
            image[x*ny+y] += (amp * region[(x-xpos)*sy+(y-ypos)]);
}

void add_image_region_constant(float *image, int nx, int ny,
                               int x, int y, int sizex, int sizey,
                               float offset)
{
    int a, b;
    for (a = x; a < (x+sizex); a++)
        for (b = y; b < (y+sizey); b++)
            image[a*ny+b] += offset;
}
                              
void set_image_region_value(float *image, int nx, int ny,
                            int x, int y, int sizex, int sizey,
                            float value)
{
    int a, b;
    for (a = x; a < (x+sizex); a++)
        for (b = y; b < (y+sizey); b++)
            image[a*ny+b] = value;
}
                              
/*
 * insert image region at image position (xpos, ypos)
 */ 
void insert_image_region(float *image, int nx, int ny,
                         float *region, int sx, int sy, 
                         int xpos, int ypos)
{
    int x, y;
    int xmin = MAX(0, xpos); int xmax = MIN(nx, xpos+sx);
    int ymin = MAX(0, ypos); int ymax = MIN(ny, ypos+sy);
    for (x = xmin; x < xmax; x++)
        for (y = ymin; y < ymax; y++)
            image[x*ny+y] = region[(x-xpos)*sy+(y-ypos)];
}

/*
 * insert image region at image position (xpos, ypos)
 */ 
void insert_image_region(float *image, int nx, int ny,
                         unsigned short *region, int sx, int sy, 
                         int xpos, int ypos)
{
    int x, y;
    int xmin = MAX(0, xpos); int xmax = MIN(nx, xpos+sx);
    int ymin = MAX(0, ypos); int ymax = MIN(ny, ypos+sy);
    for (x = xmin; x < xmax; x++)
        for (y = ymin; y < ymax; y++)
            image[x*ny+y] = (float)region[(x-xpos)*sy+(y-ypos)];
}


//---------------------------------------------------------------------------
// debugging aid: write out image region as a table of pixel values
//---------------------------------------------------------------------------
void display_ushort_image_region(unsigned short *image, int nx, int ny,
                                 int x, int y, int nsx, int nsy, 
                                 int offx, int offy)
{
    int i, j;
    printf("     |");
    for (i = 0; i < nsx; i++)
        printf(" %4d", i+x+offx);
    printf("\n");
    printf("-----+");
    for (i = 0; i < nsx; i++)
        printf("-----");
    printf("\n");
    for (j = 0; j < nsy; j++) {
        printf("%4d |", j+y+offy);
        for (i = 0; i < nsx; i++)
            printf(" %4d", image[(i+x)*ny+(j+y)]);
        printf("\n");
    }
}

void display_float_image_region(float *image, int nx, int ny,
                                int x, int y, int nsx, int nsy, 
                                int offx, int offy)
{
    int i, j;
    printf("     |");
    for (i = 0; i < nsx; i++)
        printf(" %4d", i+x+offx);
    printf("\n");
    printf("-----+");
    for (i = 0; i < nsx; i++)
        printf("-----");
    printf("\n");
    for (j = 0; j < nsy; j++) {
        printf("%4d |", j+y+offy);
        for (i = 0; i < nsx; i++)
            printf(" %4.0f", image[(i+x)*ny+(j+y)]);
        printf("\n");
    }
}
