/*
 * Routines to handle cluster intensities in different frequency channels
 * and cycles.
 * 
 * Copyright (c) Solexa   2005, 2006
 *               Illumina 2007
 * Author:       Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#else
#include <cstdio>
#include <cstdlib>
#include <cstring>
#endif
#include <cassert>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <sys/stat.h>
#include "IntensityData.h"
#include "ImageUtil.h"
#include "Stats.h"
#include "IOProcess.h"
#include "Message.h"

static int format_positions(char buffer[], int buffersize,
                            int lane, int tile, const Position &position)
{
    return snprintf(buffer, buffersize, "%i\t%i\t%.0f\t%.0f",
                    lane + 1, tile + 1,
                    int(round(position.X * 100.)) / 100.,
                    int(round(position.Y * 100.)) / 100.);
}

/*
 * Write an index file.
 * This could later be used to paste correct identifiers
 * for each cluster to a different file.
 */
int write_index_positions(const char       filename[],
                          const int        nobj,
                          const int        lane,
                          const int        tile,
                          const Position * positions)
{
    int          errorno  = 0;
    std::string  errormsg = "";
    FILE        *fp       = fopen(filename, "w");
    if (!fp) {
        errormsg = "Failed to open file " + std::string(filename) + ".";
        errorno  = 1; goto terminate;
    }

    int  i, idx;
    char buffer[256];
    for (i = 0; i < nobj; i++) {
        idx = format_positions(buffer, 20, lane, tile, positions[i]);
        sprintf(buffer+idx, "\n");
        if (fprintf(fp, buffer) < (int)strlen(buffer)) {
            errormsg = "Failed to write to " + std::string(filename) + ".";
            errorno  = 2; goto terminate;
        }
    }
    if (fclose(fp)) {
        errormsg = "Failed to close file " + std::string(filename) + ".";
        errorno  = 3; goto terminate;
    }
terminate:
    if (errorno != 0)
        error(errormsg.c_str());
    return errorno;
}

/*
 * Write an intensity file.
 * The intensity file is a text file with one line per cluster.
 * Intensities are arranged in groups of <ncol>, each intensity
 * separated by a blank. The number of such groups is the cycle number;
 * each group is separated by a tab.
 * If more than one cycle is present, there may be 4 additional, tab-separated
 * columns at the beginning of each line, identifying the cluster position.
 */
int write_signal_array(const char                    filename[],
                       const std::vector<float * > & signal,
                       const int                     nobj,
                       const int                     ncol,
                       const int                     compression,
                       const int                     lane,
                       const int                     tile,
                       const Position              * positions)
{
    int   i, j, k;
    int   idx, totalidx, count;
    int   code            = 0;
    int   ncycles         = int(signal.size());
    int   buffersize      = (22+ncycles*ncol*11);
    char *buffer          = new char [buffersize];
    int   totalbuffersize = nobj * buffersize;
    char *totalbuffer     = new char [totalbuffersize]; 
    assert(buffer != 0);
    assert(totalbuffer != 0);

    /*
     * format output and write to a char buffer
     */
    totalidx = 0;
    for (i = 0; i < nobj; i++) {
        idx = 0;
        if (positions != NULL) {
            idx = format_positions(buffer, 20, lane, tile, positions[i]);
            idx += sprintf(buffer + idx, "\t");
        }
        for (j = 0; j < ncycles; j++) {
            for (k = 0; k < ncol; k++) {
                count = snprintf(buffer+idx, 10, "%6.1f", signal[j][i*ncol+k]);
                if (count >= 10) // last character written is '\0'
                    count = 9;
                idx += count;
                if (k != ncol-1)
                    idx += sprintf(buffer+idx, " ");
                else if (j != ncycles-1)
                    idx += sprintf(buffer+idx, "\t");
                else
                    idx += sprintf(buffer+idx,"\n");
            }
        }
        memcpy(totalbuffer+totalidx, buffer, strlen(buffer));
        totalidx += strlen(buffer);
    }

    /*
     * write formatted buffer to file
     */
    if (compression == SIGNAL_COMPRESSION_NONE) {
        code = write_file(filename, totalbuffer, totalidx);
    }
    else if (compression == SIGNAL_COMPRESSION_GZIP) {
        std::string filename2(filename);
        filename2 += ".gz";
        code = write_gzip_file(filename2.c_str(), totalbuffer, totalidx);
    }
    else if (compression == SIGNAL_COMPRESSION_BZIP2) {
        std::string filename2(filename);
        filename2 += ".bz2";
        code = write_bzip2_file(filename2.c_str(), totalbuffer, totalidx);
    }

    delete [] buffer;
    delete [] totalbuffer;
    return code;
}

int write_signal_array(const char   filename[],
                       float       *intensity,
                       const int    numClusters,
                       const int    numChannels,
                       const int    compression)
{
    std::vector<float * > intensities;
    intensities.push_back(intensity);
    return write_signal_array(filename, intensities, numClusters, numChannels,
                              compression);
}


std::vector<std::string > read_signal_file(const char *filename, 
                                           const int   compression,
                                           int        *error)
{
    std::vector<std::string > lines;
    char                  *read_buffer = 0;
    long                   read_buffersize = 0;
    if (compression == SIGNAL_COMPRESSION_NONE) {
        *error = read_file(filename, &read_buffer, &read_buffersize);
    }
    else if (compression == SIGNAL_COMPRESSION_GZIP) {
        std::string filename2(filename);
        filename2 += ".gz";
        *error = read_gzip_file(filename2.c_str(),
                                &read_buffer, &read_buffersize);
    }
    else if (compression == SIGNAL_COMPRESSION_BZIP2) {
        std::string filename2(filename);
        filename2 += ".bz2";
        *error = read_bzip2_file(filename2.c_str(),
                                 &read_buffer, &read_buffersize);
    }

    if (! (*error)) {
        lines.reserve(MAX_OBJECTS);
        const int  buffersize = 16384;
        char      *s          = new char [buffersize];
        long       l, l1;
        l1 = 0;
        for (l = 0; l < read_buffersize; l++)
            if (read_buffer[l] == '\n') {
                if (read_buffer[l1] != '#') {
                    memcpy(s, &(read_buffer[l1]), l-l1+1);
                    s[l-l1+1] = '\0';
                    lines.push_back(std::string(s));
                }
                l1 = l+1;
            }
        delete [] s; 
    }
    if (read_buffersize > 0)
        delete [] read_buffer;
    return lines;
}


/*
 * read intensity data from a full text file, as described in the
 * run-folder spec.
 */
std::vector<float * > read_signal_array(const char *filename,
                                        int        *nobj,
                                        int        *ncol,
                                        int        *errorno,
                                        const int   compression,
                                        const int   offset)
{
    std::vector<float * >    intensities;
    std::vector<std::string> lines;
    std::string              errormsg = "";
    
    lines = read_signal_file(filename, compression, errorno);
    if (*errorno)
        return intensities;
    std::string buffer;
    *errorno = 0;
    *ncol = 0;
    *nobj = 0;
    buffer.reserve(10000);

    /*
     * count number of objects, cycles and columns
     */
    *nobj = int(lines.size());
    int ncycles = 0;
    if (*nobj > 0) {
        std::stringstream ss(lines[0]);
        while (getline(ss, buffer, '\t')) {
            ncycles++;
            if (ncycles > offset) {
                float dummy;
                int n = sscanf(buffer.c_str(),
                                  "%f %f %f %f", 
                                  &dummy, &dummy, &dummy, &dummy);
                if (*ncol == 0)
                    *ncol = n;
                if (*ncol != n) {
                    errormsg =  std::string(filename) + ", line 0"+
                                 ": Wrong number of intensities.";   
                    *errorno = 2;
                }
                
             }
        }
    }
    ncycles -= 4;

    /* 
     * allocate memory
     */
    int i, j, k;
    for (j = 0; j < ncycles; j++) {
        float *array = new float [*nobj * *ncol];
        for (i = 0; i < *nobj * *ncol; i++)
             array[i] = 0.;
        intensities.push_back(array);
    }

    for (i = 0; i < *nobj; i++) {
#if 0   // slow C++ version
        std::stringstream ss(lines[i]);
        int cycle = 0;
        while (getline(ss, buffer, '\t')) {
            j = cycle - offset;
            if (j >= 0) {
                float dummy[4];
                int n = sscanf(buffer.c_str(), "%f %f %f %f", 
                               &(dummy[0]), &(dummy[1]),
                               &(dummy[2]), &(dummy[3]));
                for (k = 0; k < n; k++)
                    intensities[j][i * *ncol + k] = dummy[k];
                if (*ncol != n) {
                    std::cerr << "Error: " << filename << ", line " << i <<
                                 ": Wrong number of intensities" << std::endl;
                    *errorno = 2;
                }
            }
            cycle++;
            
            if (j >= ncycles) {
                std::cerr << "Error: " << filename << ", line " << i << 
                             ": Wrong number of cycles" << std::endl;
                *errorno = 3;
            }
        }
#else
        j = -offset;
        int ncolmax = ncycles * *ncol;
        int cycle;
        const char *buffer = lines[i].c_str();
        char *begin, *end;
        float value;
        end = (char *)buffer;
        do {
            begin = end;
            value = (float)strtod(begin, &end);
            if (j >= 0) {
                cycle = j / *ncol;
                k     = j % *ncol;
                intensities[cycle][i * *ncol + k] = value;
            }
            j++;
        } while ((j < ncolmax) && (end != NULL) && (end != begin));
        if (j != ncolmax) {
            char msg[1024];
            snprintf(msg, sizeof(msg),
                     "%s, line %i: File corrupted. Expecting %i columns, "
                     "found %i.", filename, i, ncolmax+offset, j);
            errormsg = msg;
            *errorno = 2;
        }
#endif
    }
    
    if (*errorno != 0)
        error(errormsg.c_str());
    return intensities;
}

/*
 * auxiliary function to read intensity data; compare previous function
 */
int read_signal_array(const char  filename[],
                      float      *signal,
                      int        *nobj,
                      int        *ncol,
                      const int   compression,
                      const int   skipcols)
{
    int ncolmax          = *ncol;
    int nobjmax          = *nobj;
    *nobj                = 0;
    *ncol                = 0;
  
    int   errorno        = 0;
    std::string errormsg = "";
    std::vector<std::string > lines = read_signal_file(filename, compression,
                                                       &errorno);
    if (errorno)
        return errorno;


    int ncolfirst = 0;
    int lineno = 0;
    int i;
    for (i = 0; i < int(lines.size()); i++) {
        std::string line = lines[i];
        char *buffer = (char *)line.c_str();
        lineno ++;
        // ignore comments
        if (buffer[0] == '#')
            continue;
        // read molecule intensities
        if (*nobj < nobjmax) {
            int j = 0;
            char *begin, *end;
            float value;
            end = buffer;
            do {
                begin = end;
                value = (float)strtod(begin, &end);
                if (j >= skipcols)
                    signal[(*nobj)*ncolmax+j-skipcols] = value;
                j++;
            } while ((j < ncolmax+skipcols) && (end != NULL) && (end != begin));
            *ncol = j - skipcols;
            if (ncolfirst == 0) {
                ncolfirst = *ncol;
            }
            if (*ncol != ncolfirst) {
                char msg[1024];
                snprintf(msg, sizeof(msg),
                         "%s, line %i: File corrupted. Expecting %i columns, "
                         "found %i.", filename, lineno, ncolfirst, *ncol);
                errormsg = msg;
                errorno  = 2;
            }
            for (j = *ncol; j < ncolmax; j++)
                signal[(*nobj)*ncolmax+j] = 0.;
            (*nobj) ++;
        }
    }
    *ncol = ncolfirst;
    
    if (errorno != 0)
        error(errormsg.c_str());
    return errorno;
}

