
/*
 *
 * Copyright (c) 2006-2007 Solexa
 * Author: Lorenzo Milazzo
 *
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 *
 */

//
// CVS - $Id: matrix.h,v 1.5 2007/01/30 10:00:28 lm Exp $
//


//
// library name: 'matrix_lib'
// version:       0.3.2
//             
// A library for matrix operations.
//



#ifndef __MX_LIB_H__
#define __MX_LIB_H__


#include <iostream>
#include <cmath>

#include "LinAlg.h"


#define DEFAULT_MX_DIM 16


template < class T > class matrix;

template < class T >
    const matrix < T > doMultiply(const matrix < T > &mx_s, const T & scalar);

template < class T > class matrix {

  public:
    T * data;
    int rows, cols, dim;

    matrix();
    matrix(int r, int c);
    matrix(T * mx_a, int r, int c);
    ~matrix();
    matrix(const matrix < T > &mx_s);
    matrix & operator=(const matrix < T > &mx_s);
    friend matrix < T > operator*(const matrix < T > &mx_s, const T & scalar) {
        return doMultiply(mx_s, scalar);
    }
    friend matrix < T > operator*(const T & scalar, const matrix < T > &mx_s) {
        return doMultiply(mx_s, scalar);
    }
    matrix < T > operator*(const matrix < T > &mx_r);
    void setEl(const int r, const int c, const T & value);
    T getEl(const int r, const int c);
    void identity_matrix();
    float det();
    matrix < T > transpose();
    matrix < T > inverse();
    void display();

};

template < class T > matrix < T >::matrix()
{
    dim = DEFAULT_MX_DIM;
    rows = cols = (int)sqrt(DEFAULT_MX_DIM);

    data = new T[DEFAULT_MX_DIM];

    for (int i = 0; i < dim; i++) {
        data[i] = 0;
    }

}

template < class T > matrix < T >::matrix(int r, int c)
{
    dim = r * c;
    rows = r;
    cols = c;


    data = new T[dim];

    for (int i = 0; i < dim; i++) {
        data[i] = 0;
    }

}

template < class T > matrix < T >::matrix(T * mx_a, int r, int c)
{
    dim = r * c;
    rows = r;
    cols = c;

    data = new T[dim];

    for (int i = 0; i < dim; i++) {
        data[i] = mx_a[i];
    }

}

template < class T > matrix < T >::~matrix()
{
    delete [] data;
}

template < class T > matrix < T >::matrix(const matrix < T > &mx_s)
{
    dim = mx_s.dim;
    rows = mx_s.rows;
    cols = mx_s.cols;

    data = new T[dim];

    for (int i = 0; i < dim; i++) {
        data[i] = mx_s.data[i];
    }

}

template < class T >
    matrix < T > &matrix < T >::operator=(const matrix < T > &mx_s)
{
    if (this != &mx_s) {
        delete [] data;

        dim = mx_s.dim;
        rows = mx_s.rows;
        cols = mx_s.cols;

        data = new T[dim];

        for (int i = 0; i < dim; i++) {
            data[i] = mx_s.data[i];
        }
    }

    return *this;

}

template < class T >
    matrix < T > matrix < T >::operator*(const matrix < T > &mx_r)
{
    matrix < T > result(rows, mx_r.cols);

    int z = 0;
    for (int i = 0; i < rows; ++i) {
        for (int j = 0; j < mx_r.cols; ++j) {
            for (int k = 0; k < cols; ++k) {
                result.data[z] +=
                    data[k + i * cols] * mx_r.data[j + k * mx_r.cols];
            }
            z++;
        }
    }

    return result;

}

template < class T >
    void matrix < T >::setEl(const int r, const int c, const T & value)
{
    data[(c - 1) + (r - 1) * cols] = value;

}

template < class T > T matrix < T >::getEl(const int r, const int c)
{
    return data[(c - 1) + (r - 1) * cols];

}

template < typename T > void matrix < T >::identity_matrix()
{
    int i = 0;
    do {
        data[i] = 1;
        i += (cols + 1);
    } while (i < dim);

}

template < class T > float matrix < T >::det()
{
    float det;
    int *index = new int[rows];
    matrix < T > mx_temp(cols, rows);

    for (int i = 0; i < dim; i++) {
        mx_temp.data[i] = data[i];
    }

    lu_decompose_gauss_jordan_partial(mx_temp.data, rows, index, &det);

    for (int i = 0; i < rows; i++)
        det *= mx_temp.data[i + i * cols];

    delete [] index;

    return det;

}

template < class T > matrix < T > matrix < T >::transpose()
{
    matrix < T > mx_tr(cols, rows);

    int i = 0;
    for (int c = 0; c < cols; ++c) {
        for (int r = 0; r < rows; ++r) {
            mx_tr.data[i] = data[c + r * cols];
            i++;
        }
    }

    return mx_tr;

}

template < typename T > matrix < T > matrix < T >::inverse()
{
    matrix < T > mx_inv(cols, rows);

    matinvert(data, mx_inv.data, rows);

    return mx_inv;

}

template < typename T > void matrix < T >::display()
{
    std::cout.setf(std::ios::fixed, std::ios::floatfield);
    std::cout.precision(2);

    for (int r = 0; r < rows; ++r) {
        for (int c = 0; c < cols; ++c) {
            std::cout << data[c + r * cols] << " ";
        }
        std::cout << "\n";
    }

}


template < class T >
    const matrix < T > doMultiply(const matrix < T > &mx_s, const T & scalar)
{
    matrix < T > result(mx_s.rows, mx_s.cols);

    for (int i = 0; i < result.dim; i++) {
        result.data[i] = scalar * mx_s.data[i];
    }

    return result;

}


#endif // __MX_LIB_H__
