/*
 * Copyright (c) 2006-2007 Solexa
 * Author: Lorenzo Milazzo
 *
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

// A program to estimate the cross-talk matrix for the 'color-correction'
// in four-dye florescence-based DNA sequencing.


#include <iostream>
#include <fstream>
#include <cassert>

#include "display_xtalk_mx.h"
#include "estimate_xtalk_mx.h"
#include "savinio_common.h"

#include "input_process.h"
#include "output_process.h"
#include "savinio_exceptions.h"

#include "IntensityData.h"



static void usage()
{
    std::cout << "Savinio version 0.8.0\n" 
                 "Usage:\n"
                 "    savinio [option] <input file> \n\n" " Options:\n"
                 "    -h                help\n"
                 "    -o <output file>  writing to output file\n"
                 "    -z <compression>  compression level (none, gzip, bzip2)\n"
                 "    -c <cycle>        sequencing cycle\n"
              << std::endl;
}


int main(int argC, char *argV[])
{
    char * input_file  = "";
    char * output_file = "";
    bool   output      = false;
    int    compression = SIGNAL_COMPRESSION_NONE;
    int    cycle       = -1;

    if (argC < 2) {
        usage();
        return (1);
    }

    int i;
    for (i = 1; i < argC; i++) {
        if (!strcmp(argV[i], "-h")) {
            usage();
            return (0);
        }
        else if (!strcmp(argV[i], "-o")) {
            output = true;
            assert(i+1 < argC);
            output_file = argV[i+1];
            i++;
        }
        else if (!strcmp(argV[i], "-z")) {
            assert(i+1 < argC);
            if (!strcmp(argV[i+1], "gzip")) {
                compression = SIGNAL_COMPRESSION_GZIP;
            }
            else if (!strcmp(argV[i+1], "bzip2")) {
                compression = SIGNAL_COMPRESSION_BZIP2;
            }
            i++;
        }
        else if (!strcmp(argV[i], "-c")) {
            assert(i+1 < argC);
            cycle = atoi(argV[i+1]) - 1;
            i++;
        }
        else {
            input_file = argV[i];
        }
    }


    int    no_objects  = MAX_NO_CLUSTERS;
    int    no_cols     = NO_BASES;
    float *intensities = new float[no_objects * no_cols];

    // cross-talk matrix    
    Matrix mx;

    InputProcess  in(input_file);
    OutputProcess out(output_file);

    // reading input file
    in.fromInput_Intensities(intensities, no_objects, no_cols,
                             compression, cycle);


    try {
        // estimating the cross-talk matrix  
        estimateXTalkMx(intensities, no_objects, no_cols, mx);
    }
    catch(Exception01 e1) {
        Exception01 e1new((const char *)input_file, e1.getNumberObjects());
        e1new.display();

        // initial estimate of the cross-talk matrix
        mx.identity_matrix();
    }


    // printing out the cross-talk matrix
    displayXTalkMx(mx);

    // writing output file
    if (output == true) {
        out.toOutput_Matrix(mx);
    }

    delete [] intensities;

    return (0);
}
