/*
 * Class to write image analysis QC output to XML
 *
 * (c) Copyright 2006 Solexa
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */
#include "QCWriter.h"

QCWriter::QCWriter()
{
}

QCWriter::~QCWriter()
{
}

void QCWriter::openFile(const char filename[],
                        const char software[], const char software_version[])
{
    setFilename(filename);
    startDocument();
    startElement("QC");
    writeElementString("Software", software);
    writeElementString("SoftwareVersion", software_version);
}

void QCWriter::closeFile() 
{
    endElement("QC");
    endDocument();
}

void QCWriter::writeImage(Image *image,
                          int lane, int tile, int cycle, int colour)
{
    startElement("Image");
    writeElementInt("CycleNumber",   cycle);
    writeElementInt("ChannelNumber", lane);
    writeElementInt("TileNumber",    tile);
    writeElementInt("FilterNumber",  colour);
    
    writeElementChar("Base",      (char)image->data->sequencing->base);
    writeElementInt("Deblocked",        image->data->sequencing->is_deblocked);
    writeElementString("FilePath",      image->get_filename().c_str());
    writeElementFloat("ExposureTime",   image->properties->ExposureTime);
    writeElementFloat("StagePositionX", image->properties->StageXPosition);
    writeElementFloat("StagePositionY", image->properties->StageYPosition);
    writeElementInt("ObjectCount", image->data->nobj);
    writeElementFloat("EstimatedFWHM", image->data->estimated_fwhm);
    
    // global image properties
    startElement("Intensity");
    writeElementFloat("Average", image->data->image_stats->Average);
    writeElementFloat("StandardDeviation",
                                 image->data->image_stats->StandardDeviation);
    writeElementFloat("Minimum", image->data->image_stats->Minimum);
    writeElementFloat("Maximum", image->data->image_stats->Maximum);
    writeElementFloat("SaturationFraction",
                                 image->data->image_stats->FractionSaturation);
    int i;
#if 0
    for (i = 0; i < image->data->nlocal_stats; i++) {
        ImageStatsRecord *t = &(image->data->local_image_stats[i]);
        startElement("ImageRegion");
        writeElementString("ImageRegionType", "Rectangle");
        writeElementInt("ImageRegionX1", t->x1);
        writeElementInt("ImageRegionY1", t->y1);
        writeElementInt("ImageRegionX2", t->x1 + t->nx);
        writeElementInt("ImageRegionY2", t->y1 + t->ny);
        writeElementFloat("Average", t->average);
        writeElementFloat("StandardDeviation", t->standard_deviation);
        writeElementFloat("Minimum", t->minimum);
        writeElementFloat("Maximum", t->maximum);
        writeElementFloat("SaturationFraction", t->fraction_saturation);
        endElement("ImageRegion");
    }
#endif

    endElement("Intensity");

	// cluster statistics (first cycle only)
	if (cycle == 1) { 
	  startElement("ClusterStatistics");
	  writeElementInt("Count", image->data->cluster_stats.Num);
	  startElement("FWHM");
	  writeElementFloat("Mean", image->data->cluster_stats.FwhmMean);
	  writeElementFloat("Median", image->data->cluster_stats.FwhmMedian);
	  writeElementFloat("StandardDeviation",
						image->data->cluster_stats.FwhmStandardDeviation);
	  endElement("FWHM");
	  endElement("ClusterStatistics");
	}

    
    // image alignment
    startElement("UsedAlignment");
    writeElementFloat("OffsetX", image->data->usedTransform->XOffset);
    writeElementFloat("OffsetY", image->data->usedTransform->YOffset);
    writeElementFloat("StretchX", image->data->usedTransform->XScale);
    writeElementFloat("StretchY", image->data->usedTransform->YScale);
    endElement("UsedAlignment");

    startElement("Alignment");
    writeElementFloat("OffsetX", image->data->transform->XOffset);
    writeElementFloat("OffsetY", image->data->transform->YOffset);
    writeElementFloat("StretchX", image->data->transform->XScale);
    writeElementFloat("StretchY", image->data->transform->YScale);
    writeElementFloat("CorrelationMinimum",
                      image->data->transform->CorrelationMinimum);
    writeElementFloat("CorrelationMaximum",
                      image->data->transform->CorrelationMaximum);
    writeElementFloat("CorrelationWidth",
                      image->data->transform->CorrelationWidth);
    for (i = 0; i < image->data->transform->nlocal; i++) {
        AlignmentRecord *t = &(image->data->transform->local_alignment[i]);
        startElement("ImageRegion");
        writeElementString("ImageRegionType", "Rectangle");
        writeElementInt("ImageRegionX1", t->Region.XMin);
        writeElementInt("ImageRegionY1", t->Region.YMin);
        writeElementInt("ImageRegionX2", t->Region.XMin + t->Region.XSize);
        writeElementInt("ImageRegionY2", t->Region.YMin + t->Region.YSize);
        writeElementFloat("OffsetX", t->XOffset);
        writeElementFloat("OffsetY", t->YOffset);
        writeElementFloat("CorrelationMinimum", t->CorrelationMinimum);
        writeElementFloat("CorrelationMaximum", t->CorrelationMaximum);
        writeElementFloat("CorrelationWidth",   t->CorrelationWidth);
        endElement("ImageRegion");
    }
    endElement("Alignment");

    if (image->data->transform2) {
        startElement("Alignment");
        writeElementFloat("OffsetX", image->data->transform2->XOffset);
        writeElementFloat("OffsetY", image->data->transform2->YOffset);
        writeElementFloat("StretchX", image->data->transform2->XScale);
        writeElementFloat("StretchY", image->data->transform2->YScale);
        writeElementFloat("CorrelationMinimum",
                          image->data->transform2->CorrelationMinimum);
        writeElementFloat("CorrelationMaximum",
                          image->data->transform2->CorrelationMaximum);
        writeElementFloat("CorrelationWidth",
                          image->data->transform2->CorrelationWidth);
        endElement("Alignment");
    }

    endElement("Image");
}
