/* 
 * Copyright (c) 2007 Illumina 
 * Authors: Klaus Maisinger, Mirian Karbelashvili
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */
//#include "stdafx.h"
#include <string>
#include <cassert>
#include "TileAnalyser.h"
#include "ParameterTypes.h"
#include "Process.h"
#include "Tiling.h"
#include "IntensityData.h"

TileAnalyser::TileAnalyser(	const TileInfo			  &	tileInfo,
							const Position			  * positions,
							int							numPositions,
						    const RunInfo			  & runInfo,
							const AnalysisParameters  & parameters,
							const Alignment			  * defaultAlignment
								/* num Alignment elements = NUM_CHANNELS */ ) :
	m_TileInfo(0), m_Positions( positions ), m_NumPositions( numPositions ),
	m_RunInfo(0), m_Parameters(0), m_DefaultAlignment(0),
    m_RefIndices(0), m_Reference(0), m_Images(0)
{
    m_TileInfo          = new TileInfo( tileInfo );
    m_RunInfo           = new RunInfo( runInfo );
    m_Parameters        = new AnalysisParameters( parameters );
    m_DefaultAlignment	= new Alignment[m_RunInfo->NumChannels];
    m_Processor         = new DataProcessor (*m_RunInfo);
    SetAlignment( defaultAlignment );
    m_Processor->SetParameters(*m_Parameters);
    m_Processor->SetTile(*m_TileInfo);
    if (m_TileInfo->Cycle > 0)
        m_Processor->SetReference(m_Positions, numPositions);
    m_Images = new Image * [m_RunInfo->NumChannels];
    memset(m_Images, 0, sizeof(Image *) * (m_RunInfo->NumChannels));
}

TileAnalyser::TileAnalyser( const RunInfo & runInfo ) :
	m_TileInfo(0), m_Positions( 0 ), m_NumPositions( 0 ),
	m_RunInfo(0), m_Parameters(0), m_DefaultAlignment(0),
    m_RefIndices(0), m_Reference(0), m_Images(0)
{
    m_RunInfo          = new RunInfo( runInfo );
    m_DefaultAlignment = new Alignment[runInfo.NumChannels];
    m_Processor        = new DataProcessor (*m_RunInfo);
    m_Images = new Image * [m_RunInfo->NumChannels];
    memset(m_Images, 0, sizeof(Image *) * (m_RunInfo->NumChannels));
}

TileAnalyser::~TileAnalyser()
{
    int i;
    for (i = 0; i < m_RunInfo->NumChannels; i++)
        if (m_Images[i])
            delete m_Images[i];
    delete    m_TileInfo;
    delete    m_RunInfo;
    delete [] m_DefaultAlignment;
    delete    m_Parameters;
    delete    m_Processor;
    delete [] m_RefIndices;
    if (m_Reference)
        delete [] m_Reference;
    delete [] m_Images;
}

// Used in the offline version.
void TileAnalyser::SetParameters( const AnalysisParameters & parameters )
{
    delete m_Parameters;
    m_Parameters = new AnalysisParameters( parameters );
    m_Processor->SetParameters( *m_Parameters );
}

void TileAnalyser::SetReference( const Position * positions, int numPositions )
{
	m_Positions = positions;
	m_NumPositions = numPositions;
    m_Processor->SetReference( positions, numPositions );
}

void TileAnalyser::SetTile( const TileInfo & tileInfo )
{
    delete m_TileInfo;
    m_TileInfo = new TileInfo( tileInfo );
    m_Processor->SetTile( *m_TileInfo );
}

void TileAnalyser::SetAlignment( const Alignment * defaultAlignment )// num elements = NumChannels
{
    memcpy( m_DefaultAlignment, defaultAlignment, m_RunInfo->NumChannels * sizeof(Alignment) );
    m_Processor->SetAlignment(m_DefaultAlignment);
}

void TileAnalyser::AnalyseImage( unsigned short * imageValues, int colour, char base )
{
    unsigned short *pixels = new unsigned short [ m_RunInfo->ImageWidth *
                                                  m_RunInfo->ImageHeight ];
    m_Images[colour] = new Image();

    transpose_ushort_image(imageValues, pixels, 
                           m_RunInfo->ImageWidth, m_RunInfo->ImageHeight);
    m_Images[colour]->setImage(pixels,
                               m_RunInfo->ImageWidth, m_RunInfo->ImageHeight);
    AnalyseImage(m_Images[colour], colour, base);
}

void TileAnalyser::AnalyseImage( Image * image, int colour, char base )
{
    image->set_threshold(m_Parameters->DetectionThreshold);
    image->set_fwhm     (m_Parameters->ClusterFwhm);
    image->set_distance (m_Parameters->ClusterDistance);
    image->analyse();
    if (m_TileInfo->Cycle == 0)
        image->data->analyseObjectStats();
    m_Processor->AnalyseImage(image, colour, base);
}

void TileAnalyser::BuildReference( Position ** positions, int * numClusters,
                                   Position *oldPositions, int numOldClusters  )
{
    Position *reference;
    m_Processor->BuildReference(&reference, numClusters, 
                                oldPositions, numOldClusters);
    m_Reference = new Position [ *numClusters ];
    memcpy(m_Reference, reference, *numClusters * sizeof(Position));
    delete [] reference;
    *positions = m_Reference;
	m_NumPositions = *numClusters;

}

void TileAnalyser::CompleteTile(
	float	     * intensities,	 // [out]. 'NumChannels' values per cluster.
	float	     * noise,		 // [out]. 'NumChannels' values per cluster.
	Position     * position,	 // [out]. One value per cluster.
	int		     & numClusters,	 // [in].  Number of clusters (used elements) in the above arrays.
	int		       numElements,	 // [in].  Total number of elements in the above arrays, numElements >= numClusters.
	Alignment    * alignment,	 // [out]. Array of 'NumChannels' elements, allocated by the caller.
	TileQCValues * qcValues,	 // [out]. Array of 'NumChannels' elements, allocated by the caller.
	const char  ** errorMessage	)// [out]. If nonzero pointer is supplied, returns an internally allocated error message.
{
	m_NumPositions = numClusters;
	int flag = 0;
    if (m_RefIndices != 0)
        delete [] m_RefIndices;
    m_RefIndices = new int [ numClusters * m_RunInfo->NumChannels ];
    m_Processor->ClearReference();
    m_Processor->CompleteTile(intensities, noise, position, numClusters,
                              numElements, errorMessage, &flag, m_RefIndices);
    int i, j;
    Image **images = m_Processor->GetImages();
    for (i = 0; i < m_RunInfo->NumChannels; i++)
        memcpy(&alignment[i], (images[i]->data->transform), sizeof(Alignment));

    for (i = 0; i < m_RunInfo->NumChannels; i++) {
        qcValues[i].EstimatedFWHM    = images[i]->data->estimated_fwhm;
        qcValues[i].OffsetsValidated = flag;
        memcpy(&(qcValues[i].IntensityStats), (images[i]->data->image_stats),
               sizeof(ImageStats));
        memcpy(&(qcValues[i].UsedAlignment),  (images[i]->data->usedTransform),
               sizeof(Alignment));
        memcpy(&(qcValues[i].RawAlignment),   (images[i]->data->transform),
               sizeof(AlignmentStats));
		qcValues[i].Fwhm = images[i]->data->cluster_stats;

        assert(NumAlignmentRegions >= images[i]->data->transform->nlocal);
        for (j = 0; j < images[i]->data->transform->nlocal; j++)
            memcpy(&(qcValues[i].RegionAlignments[j]),
                   &(images[i]->data->transform->local_alignment[j]),
                   sizeof(AlignmentStats));
        for (j = images[i]->data->transform->nlocal; j < NumAlignmentRegions;
             j++)
            memset(&(qcValues[i].RegionAlignments[j]), 0, 
                   sizeof(AlignmentStats));
    }
}

void TileAnalyser::SaveClusterDetails( const TileInfo & tileInfo, const char * runFolder,
                                       int numClusters )
{
    int j;
    Image **images = m_Processor->GetImages();
    Tiling *name = GetName(tileInfo, runFolder);
    for (j = 0; j < m_RunInfo->NumChannels; j++) {
        images[j]->getData()->usedTransform->transformReverse(
                                images[j]->get_objects()->getData(),
                                images[j]->get_objects()->getSize());
                    
        images[j]->data->writeObjectStats(
                                name->getIndexFilename(CLUSTER_FILE, 0,
                                                       tileInfo.Cycle, j).c_str(),
                                images[j]->get_objects()->getData(),
                                m_RefIndices+j*numClusters, numClusters);
    }
    delete name;   
}

void TileAnalyser::SaveQCFile( const TileInfo & tileInfo, const char * runFolder)
{
    int j;
    Tiling *name = GetName(tileInfo, runFolder);

    QCWriter *qc_handler;
    qc_handler = new QCWriter();
    qc_handler->setIndentLevel(0);

    qc_handler->openFile(name->getIndexFilename(QC_FILE, 0, tileInfo.Cycle).c_str(),
                         software, software_version);
    for (j = 0; j < m_RunInfo->NumChannels; j++)
        qc_handler->writeImage(m_Processor->GetImages()[j], 
                               name->getLane(0)+1,
                               name->getTile(0)+1, tileInfo.Cycle+1, j+1);
   qc_handler->closeFile();
   delete qc_handler;
   delete name; 
}

void TileAnalyser::SaveIntensity( const TileInfo & tileInfo, const char * runFolder,
								  int numClusters, int numChannels, float * intensity )
{
    Tiling *name = GetName(tileInfo, runFolder);
    write_signal_array(name->getIndexFilename(INTENSITY_FILE, 0, 
                                              tileInfo.Cycle).c_str(),
                       intensity, numClusters, numChannels);
    delete name;    
}

void TileAnalyser::SaveNoise( const TileInfo & tileInfo, const char * runFolder,
							  int numClusters, int numChannels, float * noise )
{
    Tiling *name = GetName(tileInfo, runFolder);
    write_signal_array(name->getIndexFilename(NOISE_FILE, 0, 
                                              tileInfo.Cycle).c_str(),
                       noise, numClusters, numChannels);
    delete name;    
}

void TileAnalyser::SavePositions( const TileInfo & tileInfo, const char * runFolder,
								  int numClusters, Position * positions )
{
    Tiling *name = GetName(tileInfo, runFolder);
    write_positions(name->getIndexFilename(POSITION_FILE, 0) .c_str(),
                    (st_fposition* )positions, numClusters);
    delete name;  
}

void TileAnalyser::SaveIndexPositions( const TileInfo & tileInfo, const char * runFolder,
                                       int numClusters, Position * positions )
{
    Tiling *name = GetName(tileInfo, runFolder);
    write_index_positions(name->getIndexFilename(INDEX_FILE, 0) .c_str(),
                          numClusters, tileInfo.Lane, tileInfo.TileIndex,
                          positions);
    delete name;  
}

DataProcessor * TileAnalyser::GetAnalyser()
{
    return m_Processor;
}

Tiling *TileAnalyser::GetName(const TileInfo & tileInfo, const char * folder)
{
    std::string path = std::string(folder) + "/" + "s";
    Tiling *name = new Tiling(path.c_str());
    name->setLane(tileInfo.Lane, tileInfo.TileIndex, 1);
    return name;
}
