//  Time-stamp: <26 Jun 2006 16:52 BST Tools.cpp km>
/*
 * Mathematical and other auxiliary methods (e.g. for average,
 * standard deviation, histograms).
 *
 * Copyright (c) Solexa 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <math.h>
#include <stdio.h>
#else
#include <cmath>
#include <cstdio>
#endif
#include <cassert>
#include "Maths.h"
#include "Tools.h"

void initialise_array(float *a, int n, float value)
{
    int i;
    for (i = 0; i < n; i++)
        a[i] = value;
}

void add_array(float *result, float *array1, float *array2, int n)
{
    int i;
    for (i = 0; i < n; i++)
        result[i] = array1[i] + array2[i];
    return;
}

void subtract_array(float *result, float *array1, float *array2, int n)
{
    int i;
    for (i = 0; i < n; i++)
        result[i] = array1[i] - array2[i];
    return;
}

void threshold_array(float *result, float *array, float *threshold,
                     float factor, int n, float default_value)
{
    int i;
    for (i = 0; i < n; i++) {
        if (array[i] > factor * threshold[i])
            result[i] = array[i];
        else
            result[i] = default_value;
    }
}

float *identity_matrix(int n)
{
    int i;
    float *matrix = new float [n*n];
    for (i = 0; i < n*n; i++)
        matrix[i] = 0.;
    for (i = 0; i < n; i++)
        matrix[i*n+i] = 1.;
    return matrix;
}


//---------------------------------------------------------------------------
// statistics
//---------------------------------------------------------------------------

/*
 * determine average and standard deviation of the <n> values in 
 * the <image> array
 */
void get_array_stats(float *image, int n, float *average, float *sdev)
{
    if (n <= 0) {
        *average = 0.;
        *sdev = 0.;
        return;
    }
    int i;
    double a = 0., s = 0.;
    for (i = 0; i < n; i++) {
        a += image[i];
        s += image[i]*image[i];
    }
    a /= n; s /= n; s = sqrt(fabs(s - a*a));
    *average = (float)a; *sdev = (float)s;
}

void get_array_min_max(float *array, int n, float *min, float *max)
{
    int i;
    *min = array[0];
    *max = array[0];
    for (i = 0; i < n; i++) {
        if (array[i] > *max)
            *max = array[i];
        if (array[i] < *min)
            *min = array[i];
    }
}

/*
 * Some image statistics, mainly of historical interest
 * (to keep compatibility with MetaMorph journals).
 * Count the average signal value above and below a local threshold
 * given by <thresholds> * <factor>.
 */
void get_threshold_average(float *array, int n,
                           float *thresholds, float factor, 
                           float *above, float *below)
{
    int i;
    int count_above = 0, count_below = 0;
    double sum_above = 0., sum_below = 0.;
    for (i = 0; i < n; i++) {
        if (array[i] <= thresholds[i] * factor) {
            sum_below += array[i];
            count_below++;
        }
        else {
            sum_above +=  array[i];
            count_above++;
        }
    }
    if (count_above > 0)
        sum_above /= count_above;
    if (count_below != 0)
        sum_below /= count_below;
    *above = (float)sum_above;
    *below = (float)sum_below;
}

/*
 * Autoscale an image using the Genopsys algorithm,
 * i.e. the minimum is average minus one standard deviation
 * and the maximum is average plus four standard deviations.
 */
void get_genopsys_autoscale(float *im, int nx, int ny,
                            float *min, float *max)
{
    *min = *max = im[0];
    double mx = 0.;
    double mx2 = 0.;
    int i;
    for (i = 0; i < nx*ny; i++) {
        float x = im[i];
        mx += x;
        mx2 += x*x;
        *min = (*min < x ? *min : x);
        *max = (*max > x ? *max : x);
    }
    mx /= (nx*ny);
    mx2 /= (nx*ny);
    double S = sqrt(fabs(mx2-mx*mx));
    *min = (float)(mx - S);
    *max = (float)(mx + 4.*S);
}

/*
 * Autoscale an image using the Genopsys algorithm
 */
void get_genopsys_autoscale(unsigned short *im, int nx, int ny,
                            unsigned int *min, unsigned int *max)
{
    float *image = new float [nx*ny];
    float fmin, fmax;
    int i;
    for (i = 0; i < nx*ny; i++)
        image[i] = (float)im[i];
    get_genopsys_autoscale(image, nx, ny, &fmin, &fmax);
    delete [] image;
    *min = (unsigned int)MAX(0., fmin);
    *max = (unsigned int)MIN(65535., fmax);
}

//---------------------------------------------------------------------------
// histograms
//---------------------------------------------------------------------------
/*
 * bin an <n>-dim array <image> to a histogram <hist> with a given number 
 * <ntotbins> of bins of the specified <width>.
 * The first bin interval starts at <minval>.
 */
void bin(float *image, int n, int *hist, int ntotbins,
         float minval, float width)
{
    int i;
    for (i = 0; i < ntotbins; i++)
        hist[i] = 0;
#if 0
    int idx;
    for (i = 0; i < n; i++) {
        idx = (int)((image[i]-minval)/width);
        if ((idx >= 0) && (idx < ntotbins))
            hist[idx]++;
    }
#else // functionally identical version, but appears to produce faster object code
    int *index = new int [n];
    for (i = 0; i < n; i++) 
        index[i] = (int)((image[i]-minval)/width);
    for (i = 0; i < n; i++) {
        if ((index[i] >= 0) && (index[i] < ntotbins))
            hist[index[i]]++;
    }
    delete [] index;
#endif
}

//---------------------------------------------------------------------------
// diagnostics
//---------------------------------------------------------------------------
void snprint_matrix(char s[], int nmax, float *matrix, int n1, int n2)
{
    int i, j, pos = 0;
    assert(n1*n2*8+10 < nmax);
    pos += sprintf(s+pos, "[ ");
    for (i = 0; i < n1; i++) {
        pos += sprintf(s+pos, "[ ");
        for (j = 0; j < n2; j++)
            pos += sprintf(s+pos, "%4.2f ", matrix[i*n2+j]);
        assert(pos < nmax);
        pos += sprintf(s+pos, "]");
        if (i != n1 -1)
            pos += sprintf(s+pos, ", ");
    }
    pos += sprintf(s+pos, " ]");
}

