/*
 * Set of plotting routines using the PGPLOT graphics library
 * See www.astro.caltech.edu/~tjp/pgplot/
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */
/*  Time-stamp: <26 Jun 2006 17:03 BST plot.c km> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <cpgplot.h>
#include <string.h>
#include "plot.h"

char xlabel[80] = "x", ylabel[80] = "y", tlabel[80] = " ";

/*
 * set PGPLOT labels for axes and title
 */
void set_plot_labels(char xl[], char yl[], char tl[])
{
    if (strlen(xl) < 80)
        sprintf(xlabel, "%s", xl);
    if (strlen(yl) < 80)
        sprintf(ylabel, "%s", yl);
    if (strlen(tl) < 80)
        sprintf(tlabel, "%s", tl);
    return;
}

/* 
 * set the PGPLOT color table; this gives yellowish-red images 
 */
void ctab(float x[], float r[], float g[], float b[])
{
    x[0] = 0.;
    r[0] = 0.;
    g[0] = 0.0;
    b[0] = 0.;
    x[1] = 0.25;
    r[1] = 0.5;
    g[1] = 0.0;
    b[1] = 0.0;
    x[2] = 0.5;
    r[2] = 1.0;
    g[2] = 0.5;
    b[2] = 0.0;
    x[3] = 0.75;
    r[3] = 1.0;
    g[3] = 1.0;
    b[3] = 0.5;
    x[4] = 1.0;
    r[4] = 1.0;
    g[4] = 1.0;
    b[4] = 1.0;

    return;
}

/* 
 * determine the PGPLOT viewport 
 */
void viewpt(int sizex, int sizey, FTYPE cellx, FTYPE celly, float trans[],
            float *xmin, float *xmax, float *ymin, float *ymax)
{
    int centrex, centrey;

    centrex = sizex / 2 + 1;
    centrey = sizey / 2 + 1;
    trans[0] = -(float)(centrex) * cellx;
    trans[1] = (float)cellx;
    trans[2] = 0.;
    trans[3] = -(float)(centrey) * celly;
    trans[4] = 0.;
    trans[5] = (float)celly;
    *xmin = -((float)(centrex) - 0.5) * cellx;
    *xmax = ((float)(centrex - 1) - 0.5) * cellx;
    *ymin = -((float)(centrey) - 0.5) * celly;
    *ymax = ((float)(centrey - 1) - 0.5) * celly;

    return;
}

#define _NC 10

/* 
 * std value: +1., negative sign inverts color table.
 * useful for PostScript generation.
 */
float _map_plot_contrast = 1.;

void set_plot_contrast(FTYPE contra)
{
    _map_plot_contrast = (float)contra;
    return;
}

void set_plot_window(FTYPE xmin, FTYPE xmax, FTYPE ymin, FTYPE ymax)
{
    cpgenv((float)xmin, (float)xmax, (float)ymin, (float)ymax, 0, 1);
}

/*
 * subroutine to plot a 2-d array
 */
void plot_2d_array_helper(FTYPE * image, int nx, int ny, int icolor,
                          FTYPE min, FTYPE max, float trans[])
{
    int nc = _NC;
    float c[_NC], r[_NC], g[_NC], b[_NC];
    float *data;
    int i, j;

#ifdef HAVE_PGPLOT
    /* transpose image array for C-Fortran call */
    data = (float *)malloc(nx * ny * sizeof(float));
    for (i = 0; i < nx; i++)
        for (j = 0; j < ny; j++)
            data[j * nx + i] = (float)image[i * ny + j];

    if (icolor) {
        /* set the color table */
        ctab(c, r, g, b);
        cpgctab(c, r, g, b, nc, _map_plot_contrast, .5);
        /* plot the image */
        cpgimag(data, nx, ny, 1, nx, 1, ny, min, max, trans);
    }
    else {
        cpggray(data, nx, ny, 1, nx, 1, ny, max, min, trans);
    }
    free(data);
#endif

    return;
}

/* 
 * plot a 2-dimensional hypersurface of a 3-dimensional image
 */
void plot_2d_array_slice(FTYPE * image, int nx, int ny,
                         FTYPE cellx, FTYPE celly, int zval, int icolor,
                         FTYPE min, FTYPE max, int wedge)
{
    float trans[6];
    float xmin, xmax, ymin, ymax;

    if ((nx <= 1) || (ny <= 1)) {
        printf("plot_2d_slice_map: Map not 2-dimensional. Cannot plot.\n");
        return;
    }
    viewpt(nx, ny, cellx, celly, trans, &xmin, &xmax, &ymin, &ymax);
#if 1
    printf
        (" Plotting image (%ix%i) from (%5.1f, %5.1f) to (%5.1f, %5.1f)...\n",
         nx, ny, xmin, ymin, xmax, ymax);
    printf(" Cell size: (%5.2f, %5.2f)\n", cellx, celly);
#endif

#ifdef HAVE_PGPLOT
    cpgenv(xmin, xmax, ymin, ymax, 1, -2);
    plot_2d_array_helper(image + zval * nx * ny, nx, ny, icolor, min, max,
                         trans);
    cpgbox("BCTSN", 0.0, 0, "BCTSN", 0.0, 0);
    if (wedge) {
        cpglab(xlabel, ylabel, tlabel);
        if (icolor)
            cpgwedg("RI", 1., 4., (float)min, (float)max, " ");
        else
            cpgwedg("RI", 1., 4., (float)max, (float)min, " ");
    }
#endif

    return;
}

/* 
 * plot an  array, giving the minimum and maximum values of the colour scale
 */
void plot_2d_array(FTYPE * image, int nx, int ny, FTYPE cellx, FTYPE celly,
                   int icolor, FTYPE min, FTYPE max)
{
    plot_2d_array_slice(image, nx, ny, cellx, celly, 0, icolor, min, max, 1);
    return;
}

/* 
 * find useful statistics of an array 
 */
void stats(FTYPE * image, int size, FTYPE * min, FTYPE * max,
           FTYPE * mean, FTYPE * sigma)
{
    int i;
    FTYPE sum, sum2, temp;

    sum = 0.;
    sum2 = 0.;
    *max = image[0];
    *min = image[0];
    for (i = 0; i < size; i++) {
        sum += image[i];
        sum2 += (image[i] * image[i]);
        if (image[i] > *max)
            *max = image[i];
        if (image[i] < *min)
            *min = image[i];
    }
    *mean = sum / ((FTYPE) size);
    temp = sum2 / ((FTYPE) size) - ((*mean) * (*mean));
    if (temp < 0.) {
        printf("Warning: sigma %f < 0\n", temp);
        temp = 0.;
    }
    *sigma = sqrt(temp);
    return;
}


/*
 * plot several maps in one PGPLOT window, using nx rows and ny columns
 * the colour scale can be chosen
 *   -2   use the minimum and maximum of all images to determine dynamic 
 *        range; plot all maps on the same colour scale
 *   -1   use different scales for all maps
 *    n   use the minimum and maximum of the nth map to plot all maps on
 *        the same scale  
 */
void plot_3d_array(FTYPE * image, int nz, int nx, int ny, int samescale,
                   int color)
{
    int vx, vy;
    int k;
    FTYPE min, max, mean, sdev;
    float trans[6];
    float xmin, xmax, ymin, ymax;

    /* determine arrangement of panels */
    vy = (int)(floor(sqrt(nz)));
    vx = nz / vy;
    if (vx * vy < nz)
        vx++;

    min = max = image[0];
    if (samescale == -2)
        stats(image, nx * ny * nz, &min, &max, &mean, &sdev);
    if ((samescale >= 0) && (samescale < nz))
        stats(image + samescale * nx * ny, nx * ny, &min, &max, &mean, &sdev);

    printf("Using %ix%i panel\n", vx, vy);
    cpgsubp(vx, vy);
    cpgsvp(0.0, 1., 0., 1.);
    viewpt(nx, ny, 1., 1., trans, &xmin, &xmax, &ymin, &ymax);
    cpgswin(xmin, xmax, ymin, ymax);
    for (k = 0; k < nz; k++) {
        if (samescale == -1)
            stats(image + k * nx * ny, nx * ny, &min, &max, &mean, &sdev);
        cpgpage();
        plot_2d_array_helper(image + k * nx * ny, nx, ny, color,
                             min, max, trans);
    }
    cpgsubp(1, 1);
    cpgvstd();
}

void plot_line(FTYPE * x, FTYPE * y, int n, int color)
{
    int i, oldcolor, oldls;
    float *xval = (float *)malloc(n * sizeof(float));
    float *yval = (float *)malloc(n * sizeof(float));

    for (i = 0; i < n; i++) {
        xval[i] = x[i];
        yval[i] = y[i];
    }
    cpgqci(&oldcolor);
    cpgqls(&oldls);
    cpgsls(1);
    cpgsci(color);
    cpgline(n, xval, yval);
    cpgsls(oldls);
    cpgsci(oldcolor);
    free(xval);
    free(yval);
}

void plot_points(FTYPE * x, FTYPE * xerr, FTYPE * y, FTYPE * yerr,
                 int n, int color)
{
    float xv, yv, ev, xmin, xmax, ymin, ymax;
    int oldcolor, i;

    cpgqwin(&xmin, &xmax, &ymin, &ymax);
    printf("Coordinates: %f %f %f %f\n", xmin, xmax, ymin, ymax);
    cpgqci(&oldcolor);
    cpgsci(color);
    for (i = 0; i < n; i++) {
#if 0
        printf("%3i (%f %f) (%f %f)\n", i, x[i], xerr[i], y[i], yerr[i]);
#endif
        xv = (float)(x[i]);
        yv = (float)(y[i]);
        if ((xv <= xmax) && (xv >= xmin) && (yv <= ymax) && (yv >= ymin)) {
            ev = (float)(xerr[i]);
            cpgerr1(5, xv, yv, ev, 0.);
            ev = (float)(yerr[i]);
            cpgerr1(6, xv, yv, ev, 0.);
        }
    }
    cpgsci(oldcolor);
}

void plot_histogram(FTYPE * data, int n_data, FTYPE dmin, FTYPE dmax,
                    int n_bins, int color, int usewindow)
{
    int i;
    float *fdata = (float *)malloc(n_data * sizeof(float));

    for (i = 0; i < n_data; i++)
        fdata[i] = (float)data[i];
    cpgsci(color);
    cpghist(n_data, fdata, (float)dmin, (float)dmax, n_bins, usewindow);
    free(fdata);
}

FTYPE poisson(FTYPE mean, int n)
{
    int i;
    FTYPE r = exp(-mean);
    for (i = 1; i <= n; i++)
        r *= mean / i;
    return r;
}

void plot_poisson(FTYPE mean, FTYPE scale, int color)
{
    float xmin, xmax, ymin, ymax;
    FTYPE *x, *y;
    int i, i1, i2, n;

    cpgqwin(&xmin, &xmax, &ymin, &ymax);
    i1 = (int)ceil(xmin);
    i2 = (int)floor(xmax);
    n = i2 - i1 + 1;
    x = (FTYPE *) malloc(n * sizeof(FTYPE));
    y = (FTYPE *) malloc(n * sizeof(FTYPE));
    x[0] = (FTYPE) (i1);
    y[0] = (FTYPE) (scale * poisson(mean, i1));
    for (i = 1; i < n; i++) {
        x[i] = (FTYPE) (i + i1);
        y[i] = y[i - 1] * mean / (i + i1);
    }
    plot_line(x, y, n, color);
    free(x);
    free(y);
}

void plot_gaussian(FTYPE mean, FTYPE sigma, FTYPE scale, FTYPE shift,
                   int color)
{
    float xmin, xmax, ymin, ymax, xcell;
    FTYPE *x, *y;
    int i;
    int n = 100;

    cpgqwin(&xmin, &xmax, &ymin, &ymax);
    xcell = (xmax - xmin) / n;
    x = (FTYPE *) malloc((n + 1) * sizeof(FTYPE));
    y = (FTYPE *) malloc((n + 1) * sizeof(FTYPE));
    for (i = 0; i <= n; i++) {
        x[i] = (FTYPE) (xmin + i * xcell);
        y[i] =
            (FTYPE) (scale *
                     exp(-(x[i] - mean) * (x[i] - mean) /
                         (2. * sigma * sigma)) + shift);
#if 0
        printf("%f %f\n", x[i], y[i]);
#endif
    }
    plot_line(x, y, n, color);
    free(x);
    free(y);
}

/*
 * plot lines to show boundaries of wavelet domains.
 */
void plot_2d_wt_domains(int nx, int ny, FTYPE cellx, FTYPE celly,
                        int *i1, int *j1, int *i2, int *j2, int n)
{
#ifdef HAVE_PGPLOT
    int i;
    float x1, x2, y1, y2;
    float trans[6], xmin, xmax, ymin, ymax;
    int oldc, olds;

    cpgqci(&oldc);
    cpgqls(&olds);
    cpgsls(1);
    cpgsci(1);
    viewpt(nx, ny, cellx, celly, trans, &xmin, &xmax, &ymin, &ymax);
    for (i = 0; i < n; i++) {
        x1 = trans[0] + trans[1] * (i1[i] + 1.5);
        y1 = trans[3] + trans[5] * (j1[i] + 1.5);
        x2 = trans[0] + trans[1] * (i2[i] + 1.5);
        y2 = trans[3] + trans[5] * (j2[i] + 1.5);
        cpgmove(x2, y1);
        cpgdraw(x2, y2);
        cpgdraw(x1, y2);
    }
    cpgsls(olds);
    cpgsci(oldc);
#endif

    return;
}
