#!/usr/bin/env perl

# File: plotErrorRate.pl
# Copyright (c) 2005, 2006 Solexa
# Author: A. J. Cox
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

# Description:
# Produces plots of error rates vs cycle from files output by score.pl
# Assumes filenames end in .txt.
# For each fileName.txt, puts an associated fileName.pdf in same directory

use warnings;
use strict;

use FindBin;
use lib "$FindBin::Bin/../";
use Gerald::Common;

my $version;
my $id;
{
    my $cvs_id='@(#) $Id: plotErrorRate.pl,v 1.9 2007/09/13 14:14:22 rshaw Exp $';
    my $cvs_version='$Revision: 1.9 $';
    # Make CVS tag into a displayable string
    $version=$cvs_version;
    $version=~s/\$//g;
    $version=~s/Revision//;
    $id=$cvs_id;
    $id=~s/\$//g;
} #scope

#------------------------------------------------------------------------------

sub draw_graphs($$$)
{
    my ($name, $tmpName, $outName) = @_;

    open (OUT, ">$tmpName.txt") 
	|| die "Problem opening $tmpName.txt for writing: $!";

# 1. "set output" doesn't work in multiplot mode, have to pipe
# the output to a file instead
# 2. Would like to autoscale the x axes, probably need a newer version
# of gnuplot
    print OUT << "END";
#    set terminal png color;
    set terminal postscript portrait color solid;
    set nokey;
#    set yrange [0:];
    set yrange [0:20];
    set boxwidth 0.6;
    set size 1.0, 1.0;
    set origin 0.0, 0.0;
    set multiplot;
    set size 1.0, 0.5;
    set origin 0.0, 0.5;
    set title "Error rates by cycle for file $name";
#    set xlabel "Cycle Number" -1;
    set ylabel "Percent";
    plot '$tmpName.dat' using 1:2 title 'error rate' with impulses lt 1 lw 24;
    set origin 0.0, 0.0;
    set title "Occurrence of blank bases by cycle for file $name";
    set xlabel "Cycle Number" -1;
    plot '$tmpName.dat' using 1:3 title 'blank rate' with impulses lt 3 lw 24;
    set nomultiplot;

END

    close (OUT);
    gnuplotImage($tmpName.".txt", $outName);

    unlink "$tmpName.dat", "$tmpName.txt";
}

#------------------------------------------------------------------------------

my $name;
my $outName;
my $tmpName;
my $tagString="Breakdown of errors by cycle"; # look for this in file
#$tagString=~s/\ /\\\ /g;
#print $tagString, "\n"; die;

die "Usage: $0 scoreFile.txt [ outFile.png ] "
    unless ((@ARGV==1)||(@ARGV==2));

$name=$ARGV[0];

if (@ARGV==2)
{
    $outName=$ARGV[1];
} # if
else
{
    $outName=$name;
    $outName=~s/\.txt/\.png/;
    die if ($outName eq $name);
} # else

$tmpName="/tmp/tmp-plotError-$>-$$";
open (IN, $name) 
    || die "Problem opening $name: $!";

# FIXME : now even more duplication of plot_Error_Curves.pl - one script could 
#         produce both _rescore.png and _errors.png from _rescore.txt.

my $cycle_num = 0;
my $num_cycles = 0;
my $bad_num_cycles_found = 0;

my $read_num = 1; # default
my $num_reads = 1; # default

# @cycle_offsets is indexed by 1-offset $read_num.
# Initialise to zero for Read 1.
my @cycle_offsets = (0, 0);
my $cycle_offset = 0;

my @err_info_by_cycle;
my $in_errs_by_cycle = 0;
my @num_unique_aligns_by_read;

while (<IN>) {
    if ($_ =~ m/\#/) {
	if ($_ =~ /^\# Lane.*Read/) {
            if ($_ =~ /^\# Lane.* : Read (\d+)/) {
		# Header for numbered individual Read
                $num_reads = $read_num = $1; # assumes Reads in order

		if ($read_num > 1) {
		    if ($num_cycles == 0) {
			$bad_num_cycles_found = 0;
		    }

                    # We now know the number of cycles in the previous read.
                    updateCycleOffsets(\@cycle_offsets,
                                       $read_num - 1, $num_cycles);
                    $cycle_offset = $cycle_offsets[$read_num];
                }
	    } else {
                # All Reads header - assumed after the individual Reads.
                last; # do not need any All Reads statistics
            }
        }

        next;
    } elsif ($_ =~ /unique alignments : (\d+) \(total score \d+\)/) {
	$num_unique_aligns_by_read[$read_num] = $1;
    } elsif ($_ =~ /cycles : (\d+)/) {
	$num_cycles = $1;
    }

    chomp;

    if ($in_errs_by_cycle == 1) {
	my @fields = split;

	if (@fields != 3) {
	    # First line after an errors-by-cycle data line (if any).
	    $in_errs_by_cycle = 0;
	    next;
	}

	$cycle_num = $fields[0];
	@{$err_info_by_cycle[$cycle_offset + $cycle_num - 1]} = @fields;
    } elsif ($_ eq $tagString) {
	# Skip the column headings line.
	die "Expected line after $tagString" unless defined(<IN>);
	$in_errs_by_cycle = 1;
    }
} # loop over input file

close (IN);

# We now know the number of cycles in the last individual Read.

if ($num_cycles == 0) {
    $bad_num_cycles_found = 0;
}

updateCycleOffsets(\@cycle_offsets, $read_num, $num_cycles);


# If no data found, create a blank file and continue 
# (rather than stop pipeline)
if ($bad_num_cycles_found) {
    warn "No clusters found in one or more reads, creating blank $outName";
    open (OUT, ">$outName") 
	|| die "Problem creating blank file $outName: $!";
    close (OUT);
    exit (0);
}

# Write gnuplot data file.

open (OUT, ">$tmpName.dat")
    || die "Problem opening $tmpName.dat for writing: $!";

for (my $read_ind = 1; $read_ind <= $num_reads; ++$read_ind) {
    my $num_unique_aligns = $num_unique_aligns_by_read[$read_ind];
    my $next_read_start_cycle = $cycle_offsets[$read_ind + 1];

    for (my $cycle_ind = $cycle_offsets[$read_ind];
         $cycle_ind < $next_read_start_cycle; ++$cycle_ind) {

	printf OUT "%d\t", $cycle_ind + 1;

	# If there are no unique alignments for a read, set err% 0, blank% 100.
	if ($num_unique_aligns == 0) {
	    print OUT "0.0\t100.0\n";
	    next;
	}

	for (my $field_ind = 1; $field_ind <= 2; ++$field_ind) {
	    print OUT "$err_info_by_cycle[$cycle_ind][$field_ind]\t";
	}

	print OUT "\n";
    }
}

close (OUT);

draw_graphs($name, $tmpName, $outName);

#------------------------------------------------------------------------------

