#! /usr/bin/env python
#
# Copyright (c) Solexa 2007
# Author: Klaus Maisinger
#
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

import sys, os, glob, re, types, time
import runfolder, XMLSimple

"""Common routines between goat_pipeline and bustard."""

PairedEndConfigFile    = "PairedEndInfo.xml"
MaxSequencingAutoCycle = 1000 

class Params:
    def __init__(self):
        self.options = []

    def getDict(self):
        """Return variable settings as a dictionary."""
        dic = getDict(self)
        del dic["options"]
        return dic

    def parse(self, options):
        error("parse() function not implemented!")


class Read:
    def __init__(self, RunFolder, Index=-1, FirstCycle = -1,
                 LastCycle = -1):
        self.FirstCycle = FirstCycle
        self.LastCycle  = LastCycle
        self.Index      = Index
        self.RunFolder  = RunFolder

def readcmp(x, y):
    if x.FirstCycle > y.FirstCycle:
        return 1
    else:
        return -1
    return 0

class GeneralParams(Params):
    def __init__(self):
        self.Compression       = "gzip"
        self.CompressionSuffix = ".gz"
        self.SelectedTiles     = []
        self.SelectedFolders   = []
        self.MakeFlag          = 0
        self.BasecallFlag      = 1
        self.MinCycle          = -1
        self.MaxCycle          = -1
        self.Deblocked         = 0
        self.SelectedCycles    = []
        self.AutoCycleFlag     = 0
        self.ImagingReads      = [ Read("", Index=1, FirstCycle=1) ]
        self.Reads             = [ Read("", Index=1, FirstCycle=1) ]
        self.Instrument        = ""
        self.InstrumentFolder  = os.getenv('INSTRUMENT_DIR')
        self.RunFolder         = ""
        self.DebugFlag         = 0
        self.options           = [ "new-read-cycle=", "compression=",
                                   "make", "tiles=", "folders=",
                                   "nobasecall", "deblocked",
                                   "maxcycle=", "mincycle=",
                                   "cycles=", "debug",
                                   "INSTRUMENT_DIR="]

    def parseImageFolders(self, args):
        """Identify total number of reads.
        Different reads can be recognised in two ways:
        1. They are in different run-folders
        2. From the configuration file \"PairedEndInfo.xml\""""
        folders = []
        for a in args:
            folder = runfolder.get_run_folder(a)
            if not os.path.exists(folder):
                warning("Folder %s not found. Ignoring." % folder)
                continue
            if folder not in folders:
                folders.append(folder)
        if (len(folders) > 0):
            self.ImagingReads = []
            self.addRead(folders[0], FirstCycle=1)
        if len(folders) > 1: # more than one run-folder
            for i in range(1, len(folders)):
                folder = folders[i]
                self.addRead(folder, FirstCycle=1)
        elif (len(folders) == 0):
            error("You need to specify at least one "
                  "Run-Folder for analysis!")
            raise "Error"
        elif len(folders) == 1: # one run-folder: look for config file
            folder = folders[0]
            pairedfile = os.path.join(folder, PairedEndConfigFile)
            if os.path.exists(pairedfile):
                print "Found paired end config file %s." % pairedfile
                LastCycle = 0
                try:
                    data = XMLSimple.XMLin(pairedfile, KeepRoot=1)
                    LastCycle = int(data['FirstRead']['Length'])
                    print "First read has length %i." % LastCycle
                except:
                    warning("Failed to parse %s. Ignoring." % pairedfile)
                #print data
                if LastCycle > 0:
                    self.addRead(FirstCycle=LastCycle+1)
        self.RunFolder = folders[0]
        self.Instrument = get_instrument(self.RunFolder)

    def parseImageAnalysis(self, args):
        """Read run parameters from the .params file in a Firecrest folder."""
        if (len(args)) < 1:
            error("Need to specify image analysis folder!")
            raise "Error"
        analysis_folder = os.path.abspath(args[0])
        paramsfile = os.path.join(runfolder.get_data_folder(analysis_folder),
                                  ".params")
        if paramsfile:
            params = runfolder.get_params(paramsfile,
                                          ["ImageAnalysis", "data"])
            run = runfolder.get_run(params, analysis_folder)
            if not run:
                error("Could not find run in .params file")
                raise "Error"
            if run.has_key("RunParameters"):
                data = run["RunParameters"]
                for key in data.keys():
                    self.__dict__[key] = data[key]
                if type(self.ImagingReads) == types.DictType:
                    self.ImagingReads = [ self.ImagingReads ]
                for i in range(len(self.ImagingReads)):
                    read = Read("")
                    for key in self.ImagingReads[i].keys():
                        read.__dict__[key] = self.ImagingReads[i][key]
                    self.ImagingReads[i] = read

                if type(self.Reads) == types.DictType:
                    self.Reads = [ self.Reads ]
                for i in range(len(self.Reads)):
                    read = Read("")
                    for key in self.Reads[i].keys():
                        read.__dict__[key] = self.Reads[i][key]
                    self.Reads[i] = read

            self.Tiles        = runfolder.get_tiles_from_tile_selection(run)
            self.MaxCycle     = run["Cycles"]['Number']
            self.MakeFlag     = 0
            self.BasecallFlag = 1
            self.DebugFlag    = 0


    def printInstrument(self):
        print "Instrument:    %s" % self.Instrument
        
    def printReads(self):
        if (len(self.ImagingReads) > 1):
            for read in self.ImagingReads:
                print "Read %i:        Imaging cycles %3i-%3i" % \
                      (read.Index, read.FirstCycle, read.LastCycle)

    def addRead(self, RunFolder = "", FirstCycle = -1,
                LastCycle = -1):
        for read in self.ImagingReads:
            if FirstCycle > 1 and FirstCycle == read.FirstCycle:
                error("Cannot add new read at cycle %i: "
                      "Already have read starting at cycle %i!\n"
                      % (FirstCycle, read.FirstCycle))
                sys.exit(1)
        Index = len(self.ImagingReads)
        for i in range(len(self.ImagingReads)):
            if FirstCycle > 1:
                if self.ImagingReads[i] <= FirstCycle:
                    Index = i+1
        for i in range(Index, len(self.ImagingReads)):
            self.ImagingReads[i].Index += 1
        #if len(self.Reads) > 1:
        #    error("Cannot add new read: Already have more than one read!")
        #    sys.exit(1)
        if FirstCycle > -1 and len(self.ImagingReads) > 0:
            self.ImagingReads[Index-1].LastCycle = FirstCycle - 1
        self.ImagingReads.insert(Index, Read(RunFolder, Index = len(self.ImagingReads)+1,
                               FirstCycle=FirstCycle,
                               LastCycle=LastCycle))
 
    def completeReads(self, LastCycle, RestartCycles=[]):
        """Fill in missing cycle information and ensure logical order
        of reads."""
        count = 0
        for read in self.ImagingReads:
            if read.FirstCycle == 1:
                count += 1
        if (count != 1 + len(RestartCycles)):
            error("GeneralParams::completeRead(): Inconsistent read info.")
            sys.exit(1)
        count = 0
        for read in self.ImagingReads:
            if read.FirstCycle == 1:
                if (count > 0):
                    read.FirstCycle = RestartCycles[count-1]
                count += 1

        # now sort reads and fill in last cycles
        self.ImagingReads.sort(readcmp)
        for i in range(len(self.ImagingReads)):
            self.ImagingReads[i].Index = i+1
        self.ImagingReads[-1].LastCycle = LastCycle
        for i in range(len(self.ImagingReads)-1):
            self.ImagingReads[i].LastCycle = self.ImagingReads[i+1].FirstCycle-1
        # create list of logical cycle numbers
        self.Reads = []
        count = 0
        for read in self.ImagingReads:
            newread = Read(read.RunFolder, read.Index)
            newread.FirstCycle = count+1
            cycles = range(read.FirstCycle, read.LastCycle+1)
            newcycles = []
            if self.SelectedCycles:
                for cycle in cycles:
                    if cycle-read.FirstCycle+1 in self.SelectedCycles:
                        newcycles.append(cycle)
            else:
                newcycles = cycles
            count += len(newcycles)
            newread.LastCycle = count
            self.Reads.append(newread)

                
    def parse(self, options):
        for o, a in options:
            if o == "--new-read-cycle":
                self.addRead(FirstCycle=int(a))
            if o == "--debug":
                self.DebugFlag = 1
            if o == "--compression":
                comp = a.lower()
                compression_options = { "none"  : "",
                                        "gzip"  : ".gz",
                                        "bzip2" : ".bz2" }
                if comp in compression_options.keys():
                    self.Compression = comp
                    self.CompressionSuffix = compression_options[comp]
            if o == "--make":
                self.MakeFlag = 1
            if o == "--tiles":
                self.SelectedTiles = a.split(",")
            if o == "--folders":
                self.SelectedFolders = a.split(",")
            if o == "--nobasecall":
                self.BasecallFlag = 0
            if o == "--maxcycle":
                self.MaxCycle = int(a)
            if o in [ "--mincycle", "-m"]:
                self.MinCycle = int(a)
            if o == "--deblocked":
                self.Deblocked = 1
            if o == "--cycles":
                a = a.lower()
                if (a.find("auto") >= 0):
                    self.AutoCycleFlag = 1
                if (a.find("auto") > 0):
                    a = a.replace('auto', "%i" % MaxSequencingAutoCycle)
                if not (a.find('auto') == 0):
                    self.SelectedCycles = expand_ranges(a)
            if o == "--INSTRUMENT_DIR":
                self.InstrumentFolder = os.path.abspath(a)


def getDict(obj):
    """Serialise an object to simple types (dictionaries),
    which can later be processed by XMLSimple."""
    result = None
    if type(obj) == types.ListType:
        result = obj[:]
        for i in range(len(result)):
            result[i] = getDict(result[i])
    elif type(obj) == types.DictType:
        result = obj.copy()
        for key in result.keys():
            result[key] = getDict(result[key])
    elif type(obj) == types.InstanceType:
        result = obj.__dict__.copy()
        for key in result.keys():
            result[key] = getDict(result[key])
    else:
        result = obj
    return result


def unique_list(a):
    """Return a reduced list whose elements are unique."""
    dic = {}
    for element in a:
        dic[element] = 1
    return dic.keys()


def expand_ranges(a):
    """Expand various short forms of lists of ordered numbers."""
    numbers = []
    if (type(a) == types.StringType):
        for arg in a.split(","):
            arg2 = map(int, arg.split("-"))
            if len(arg2) == 1:
                arg2.append(arg2[0])
            numbers += range(int(arg2[0]), int(arg2[1])+1)
    elif (type(a) == types.ListType):
        for arg in a:
            if type(arg) == types.IntType:
                numbers.append(arg)
            else:
                numbers.append(arg[0], arg[1]+1)
    numbers = unique_list(numbers)
    numbers.sort()
    return numbers

def collapse_ranges(numbers):
    """Given a list of numbers, collapse runs of consecutive numbers
    to a list denoting the first and the last number."""
    sortednumbers = unique_list(numbers)
    sortednumbers.sort()
    newnumbers = []
    rangelist = []
    first, last = sortednumbers[0], sortednumbers[-1]
    prev = first
    for item in sortednumbers[1:]:
        if item != prev+1:
            rangelist.append([ first, prev ])
            first = item
        prev = item
    rangelist.append([first, last])
    folderlist = []
    for pair in rangelist:
        if pair[0] == pair[1]:
            newnumbers.append(pair[0])
        else:
            newnumbers.append([pair[0], pair[1]])
    return newnumbers
        
def collapse_ranges_to_string():
    pass

def select_tiles(tiledic, selected_tiles =[]):
    """Select tiles from a dictionary whose keys refer to tiles. Only part
    of the key needs to match the selected tiles."""
    if selected_tiles:
        a = re.compile("(%s)" % "|".join(selected_tiles))
        for key in tiledic.keys():
            if not a.search(key):
                del tiledic[key]
    return tiledic

def count_intensity_files(analysis_folder):
    """Count the number of cycles per intesntiy file in the image analysis
    working directory."""
    searchfolder = os.path.join(analysis_folder, "*_int.txt")
    filenames = glob.glob(searchfolder)
    tiledic = {}
    if (len(filenames) == 0):
        warning("No intensity files found in %s." % analysis_folder)
        searchfolder = os.path.join(analysis_folder, "*_int.txt.*")
        if len(glob.glob(searchfolder)) > 0:
            warning("Found compressed files in %s. Cannot deduce "
                    "cycle number from compressed files." % analysis_folder)
    for filename in filenames:
        tile  = runfolder.FileName.tilestr(filename)
        line  = open(filename).readline()
        count = len(line.split("\t"))-4
        print filename, tile, line, count
        if count > 0:
            tiledic[tile] = count
    return tiledic

def group_dictionary(function, list):
    """Create a dictionary where each key corresponds to a function
    value and the corresponding values are all list elements with the
    same function value.""" 
    dict = {}
    for item in list:
        id = function(item)
        dict.setdefault(id, []).append(item)
    return dict
        

def group(function, list):
    """Create a list of lists each containing elements that yield the
    same result when the function is applied. This creates groups
    of elements that have the same function value."""
    # sort by key value and return array
    dic = group_dictionary(function, list)
    keys = dic.keys()
    keys.sort()
    return [ dic[key] for key in keys ]


def get_instrument(analysis_folder):
    """Return name of instrument from runfolder parameter file"""
    paramsfile = runfolder.get_run_params_filename(analysis_folder)
    instrument = ""
    if os.path.isfile(paramsfile):
        params = runfolder.get_params(paramsfile)
        try:
            instrument = params["Run"][0]["instrument"]
        except:
            print params
            warning("No instrument id found in parameter file %s."
                    % paramsfile)
    else:
        warning("Parameter file %s not found!\n" % paramsfile)
    return instrument

def get_matrix(matrixfilename, instrument_folder, analysis_folder):
    """Return path to matrix file"""
    matrixfile = matrixfilename or \
                 runfolder.find_instrument_matrix_file(analysis_folder,
                                                       instrument_folder,
                                                       get_instrument(analysis_folder))
    if not os.path.exists(matrixfile):
        warning("No matrix file %s found!" % matrixfile)
        return ""
    else:
        return matrixfile

def logline(level, message):
    tm = time.gmtime(time.time())
    mytime = time.strftime("%H:%M:%S", tm)
    mydate = time.strftime("%d/%m/%Y", tm)
    return ("%s\t%s\t%s\t%s\t%s\t%s\t%s\n"
            % (level, mydate, mytime, sys.argv[0],
               os.getenv('LOGNAME') or os.getenv('USERNAME'),
               os.getenv('HOSTNAME'), message))

def log(filename, level, message):
    try:
        open(filename, "a").writelines([ logline(level, message)])
    except:
        sys.stderr.write("Warning: Failed to write to %s.\n" %
                         os.path.abspath(filename))
    

def error(message):
    """Central routine for handling error messages"""
    sys.stderr.write("Error: %s\n" % message)
    log("error.txt", 7, message)

def warning(message):
    """Central routine for handling warning messages"""
    sys.stderr.write("Warning: %s\n" % message)
    log("warning.txt", 6, message)
