#! /usr/bin/env python

#
# Copyright (c) 2005-2007 Solexa
# Author: Klaus Maisinger, Lorenzo Milazzo
#
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.
#

# program name: 'Goat-Pipeline'
# CVS - $Id: goat_pipeline.py,v 1.101 2007/11/09 11:07:18 rshaw Exp $      
#             
# script to configure the analysis environment;
# generation of the Makefiles for:
#  . image analysis 
#  . base-calling
#  . sequencing

import os, os.path, sys, getopt, glob, re, shutil
import runfolder, makefile, matrix, common, bustard, gerald, XMLSimple
from manteia     import ManteiaFilePath
from swerdlotron import SwerdlotronFileName

"""Script to start the Genome Analyzer Pipeline on
a given experimental run-folder produced by a Genome Analyzer."""

class ImageParams(common.Params):
    def __init__(self): 
        self.Threshold            =  4.
        self.Fwhm                 =  2.7
        self.RemappingDistance    =  1.5
        self.AutoOffsetFlag       =  1
        self.AutoSizeFlag         =  0
        self.SizeFile             = ""
        self.OffsetFile           = ""
        self.DataOffsetFile       = ""
        self.InstrumentOffsetFile = ""
        self.options              = [ "threshold=",
                                      "fwhm=",
                                      "distance=",
                                      "offsets=",
                                      "size="]

    def setSize(self, size):
        """Set an average cluster size."""
        self.Fwhm              = size
        self.RemappingDistance = size/2. + 0.15

    def getParamsXML(self):
        newparams = self.getDict()
        newparams2 = {}
        outputKeys = [ 'Fwhm', 'RemappingDistance' ]
        outputRoot = 'ImageParameters'
        for key in outputKeys:
            newparams2[key] = newparams[key]
        return XMLSimple.XMLout(newparams2, RootName=outputRoot)
        
    def completePath(self, run_params):
        run_folder        = run_params.RunFolder
        instrument_folder = run_params.InstrumentFolder
        instrument        = run_params.Instrument
        data_folder = runfolder.get_data_folder(run_folder)
        self.DataOffsetFile = os.path.join(data_folder,
                                           runfolder.offsetfile)
        if not self.OffsetFile:
            self.OffsetFile = runfolder.find_data_offset_file(data_folder)
        if not self.OffsetFile:
            self.InstrumentOffsetFile = \
                runfolder.find_instrument_offset_file(data_folder,
                                                      instrument_folder,
                                                      instrument)
            self.OffsetFile = self.InstrumentOffsetFile
        if (not self.OffsetFile) or (not os.path.exists(self.OffsetFile)):    
            common.warning("No offset file found!")
            if not self.AutoOffsetFlag:
                common.warning("Now turning on the "
                           "\"--offsets=auto\" option!")
            self.AutoOffsetFlag = 1
            self.InstrumentOffsetFile = \
                create_instrument_offset_file(data_folder,
                                              instrument_folder,
                                              instrument)
            self.OffsetFile = self.InstrumentOffsetFile

    def printOffsets(self):
        if self.OffsetFile and os.path.exists(self.OffsetFile):    
            print "Offset file:   %s" % self.OffsetFile
        if self.AutoOffsetFlag:
            print "  Using automatic offset calibration."
        if self.InstrumentOffsetFile:
            print "  Will update offsets file %s after run!" \
                  % self.InstrumentOffsetFile

    def parse(self, options):
        """Set values according to command-line options obtained by
        getopt.getopt"""
        for o, a in options:
            if o == "--threshold":
                self.Threshold = float(a)
            if o == "--fwhm":
                self.Fwhm = float(a)
            if o == "--distance":
                self.RemappingDistance = float(a)
            if o == "--offsets":
                if (a.lower().find("auto") == 0):
                    self.AutoOffsetFlag = 1
                elif (a.lower().find("default") == 0):
                    self.AutoOffsetFlag = 0
                else:
                    self.OffsetFile = os.path.abspath(a)
                    self.AutoOffsetFlag = 0
            if o == "--size":
                if (a.lower().find("auto") == 0):
                    self.AutoSizeFlag = 1
                else:
                    self.AutoSizeFlag = 0
                    try:
                        size = float(a)
                        self.setSize(a)
                    except:
                        self.SizeFile = os.path.abspath(a)
                    
        

image_software              = "Firecrest"
image_software_version      = "1.9.2"
image_executable      = "run_image_analysis"
image2_executable     = "analyse_image"
offset_executable     = "parse_offset_scaling.py"
offset2_executable    = "update_offsets.py"
sizes_executable      = "parse_cluster_stats.py"
sizes2_executable     = "select_cluster_size.py"
sizes3_executable     = "fwhm2params.py"
matrixpath            = "Matrix"
matrix1_executable    = os.path.join(matrixpath, "savinio")
matrix3_executable    = os.path.join(matrixpath, "summarise_matrices.py")
matrix4_executable    = os.path.join(matrixpath, "select_matrices.py")


def get_tiff_files(files):
    """Return all tiff files given in a list of files or directories."""
    allnames = []
    for file in files:
        if os.path.isdir(file):
            allnames += glob.glob(os.path.join(file, "*.tif"))
            allnames += glob.glob(os.path.join(file, "*.stk"))
        else:
            allnames += [file]
    allnames = filter(re.compile("\.[(tif)|(stk)]").search, allnames)
    return allnames


def get_bottom_level_folders(filelist):
    """Obtain folder names in which image files reside."""
    folders = []
    for folder in filter(None, map(runfolder.get_bottom_level_folder,
                                   filelist)):
        if folder not in folders:
            folders.append(folder)
    return folders


def get_cycle_params(files, FileNameClass):
    """Given a list of image files, extract number of cycles, colours
    and a string describing the alternation of incorporation/deblock
    cycles."""
    if (len(files) == 0):
        return (0, 0, "")
    firstcycle = FileNameClass.cycle(files[0])
    
    firstdeblock = FileNameClass.is_deblocked(files[0])
    ncolours = 1
    for i in range(1, len(files)):
        if ((FileNameClass.cycle(files[i]) == firstcycle) and
            ((FileNameClass.is_deblocked(files[i]) == firstdeblock))):
            ncolours += 1
        else:
            break
    ncycles = len(files)/ncolours
    
    deblock = ['C'] * ncycles
    for i in range(ncycles):
        if FileNameClass.is_deblocked(files[i*ncolours]):
            deblock[i] = 'D'
    return (ncycles, ncolours, "".join(deblock))
        

def write_image_journal(filename, files, FileNameClass):
    """Write out a journal file for the image analysis. This is used
    to pass on a list of filenames to be analysed to the image
    analysis software. The format is archaic and derived from a legacy
    application. This will be replaced by an XML file."""
    ncycles, ncolours, deblock = get_cycle_params(files, FileNameClass)
    filename = filename.replace("txt", "xml")
    try:
        baselist = map(lambda x: FileNameClass.base(x).upper(),
                       files[:ncolours])
        output = { "Channels": { "Num" :ncolours,
                                 "Order" : ''.join(baselist) },
                   "Cycles":   { "Num" : ncycles },
                   "Compression": "none",
                   "Filename" :   map(lambda x: { "Path": x}, files) }
        XMLSimple.XMLout(output,
                         RootName="JournalFile",
                         OutputFile=filename, 
                         XMLDecl = 1,
                         KeyAttr = ["Path", "Num", "Order"])
    except:
        common.error("Could not create %s." % filename)
        sys.exit(2)


def run_generate_journal(analysis_folder, sampleid, lane, tile,
                         files, FileNameClass):
    """Generate journal file for the image analysis."""
    # generate folder for intermediate files
    tmp_folder = runfolder.get_lane_subfolder(analysis_folder,
                                              runfolder.analysissubdir,
                                              sampleid, lane, tile)
    journalfile = os.path.join(tmp_folder,
                               runfolder.FileName.name(sampleid, "jnl",
                                                       lane, tile))
    write_image_journal(journalfile, files, FileNameClass)


def rename_image_folders(old_analysis_folder, analysis_folder):
    """Rename image folders."""
    try:
        os.rename(old_analysis_folder, analysis_folder)
        for filename in glob.glob(os.path.join(analysis_folder,
                                           "*.old*")):
            os.remove(filename)
        for suffix in [ "_int.txt", "_nse.txt" ]:
            for filename in glob.glob(os.path.join(analysis_folder,
                                                   "*%s*" % suffix)):
                parts = filename.split(suffix)
                newfilename = parts[0] + suffix + ".old"
                if (len(parts) > 1):
                    newfilename += parts[1]
                os.rename(filename, newfilename)
        for filename in glob.glob(os.path.join(analysis_folder,
                                           "*finished.txt")):
            os.remove(filename)
        
    except:
        common.error("Failed to rename %s to %s!"
                     % (old_analysis_folder, analysis_folder))
        sys.exit(1)


def create_image_folders(analysis_folder, tiles):
    lanes = runfolder.get_lane_samples(tiles)
    try:
        os.makedirs(analysis_folder)
        os.mkdir(os.path.join(analysis_folder,
                              runfolder.analysissubdir))
        for lane in lanes.keys():
            folder = runfolder.get_lane_subfolder(analysis_folder,
                                                  runfolder.analysissubdir,
                                                  lanes[lane], lane, 0)
            os.mkdir(folder)
        os.mkdir(os.path.join(analysis_folder, runfolder.offsetdir))
        os.mkdir(os.path.join(analysis_folder, runfolder.matrixdir))
        os.mkdir(os.path.join(analysis_folder, runfolder.focusdir))
        os.mkdir(os.path.join(analysis_folder, runfolder.sizedir))
    except:
        common.error("Encountered error while creating folder %s. "
                     "Perhaps you do not have the permissions?"
                     % analysis_folder)
        sys.exit(1)


def get_cycle_counts_dic(tile_dict, FileNameClass):
    """Count count number of images per cycle per tile"""
    counts = {}
    for tile in tile_dict.keys():
        counts[tile] = get_cycle_counts(tile_dict[tile], FileNameClass)
    return counts
    
    
def get_cycle_counts(files, FileNameClass):
    """Count count number of images per cycle give a list of files"""
    counts = [ [ 0, 0 ] ]
    lastid = 0

    if (len(files)) <= 0:
        return []
    lastcycle = FileNameClass.cycle(files[0])

    for file in files:
        id = FileNameClass.is_deblocked(file)
        cycle = FileNameClass.cycle(file)
        if ((lastid == 1 and id == 0) or (counts[-1][id] >= 4) or
            (cycle != lastcycle)):
            counts.append([0,0])
        counts[-1][id] += 1
        lastid = id
        lastcycle = cycle

    return counts


def print_cycle_overview(tile_dict, FileNameClass):
    """Print a summary of the number of images found in enzymology and
    deblocking scans for each tile."""
    counts = get_cycle_counts_dic(tile_dict, FileNameClass)
    maxcycle = max(map(len, counts.values()))
    keys = tile_dict.keys()
    keys.sort()
    print "Tile " + " " * max([len(keys[0])-4, 0]) +" images  %s" %\
          (" ".join(map(lambda x: "C%02d" % x, range(1, maxcycle+1))))
    for tile in keys:
        print "%4s  %3d    " % (tile, len(tile_dict[tile])),
        for cycleno, deblockedno in counts[tile]:
            print "%1d-%1d" % (cycleno, deblockedno),
        print

        
def validate_cycles(tiledic, FileNameClass, autocycleflag=1):
    """Runs some validation checks on a list of filenames to make sure
    that the directory structure is not corrupted."""
    validated = 1
    counts = get_cycle_counts_dic(tiledic, FileNameClass)
    maxcycle = 0
    mincycle = -1
    for tile in tiledic.keys():
        deblocked = counts[tile][0][1]
        ncycles = len(counts[tile])
        # make sure all tiles have the same number of images
        if maxcycle == 0:
            maxcycle = ncycles
        if mincycle < 0:
            mincycle = ncycles
        if maxcycle != ncycles and not autocycleflag:
            validated = 0
            common.error(
                "Tile %s: "
                "Different number of cycles: %i instead of %i. "
                "The downstream pipeline cannot currently handle this. "
                "You may need to restrict the number of cycles using "
                "the \"--cycles\" option (or use \"--cycles=auto\")."
                % (tile, ncycles, maxcycle))
        if mincycle > ncycles:
            mincycle = ncycles
        if maxcycle < ncycles:
            maxcycle = ncycles
            
        firstcyclecount = counts[tile][0][0]
        for i in range(ncycles):
            cyclecount, deblockedcount = counts[tile][i]
            # check that each cycle has the same number of images as
            # in for the first cycle 
            if (cyclecount != firstcyclecount):
                msg = ("Tile %s cycle %2i: "
                       "Wrong number of images. Your image folders "
                       "are not consistent. Expected "
                       "%i images, but found %i."
                       % (tile, i+1, firstcyclecount, cyclecount))
                if (i == ncycles-1) and autocycleflag:
                    common.warning(msg)
                    if mincycle >  ncycles-1:
                        mincycle = ncycles-1
                else:
                    validated = 0
                    common.error(msg)

            # deblocked images also need to have the same number of
            # images (apart from the last cycle), or no deblocked
            # images at all. 
            if (deblockedcount not in [0,firstcyclecount]):
                if (i == ncycles-1):
                    common.warning(
                        "Tile %s cycle %2i: "
                        "Discarding last %i deblocked images "
                        "to make number of images consistent "
                        "with runfolder spec."
                        % (tile, i+1, deblockedcount))
                    tiledic[tile] = tiledic[tile][:-deblockedcount]
                    counts[tile][i][1] = 0
                else:
                    validated = 0
                    common.error(
                        "Tile %s cycle %2i: Wrong number of "
                        "deblocked images. Expected 0 or %i images, "
                        "but found %i images."
                        % (tile, i+1, firstcyclecount, deblockedcount))
            if ((deblockedcount != deblocked) and (i != ncycles-1)):
                validate = 0
                common.error(
                    "Tile %s cycle %2i: " 
                    "Must have deblocked images in all cycles or none."
                    % (tile, i+1))
    if autocycleflag and maxcycle != mincycle:
        print "Setting uniform read length of %i cycles." % mincycle
        for tile in tiledic.keys():
            tiledic[tile] = filter(lambda x: FileNameClass.cycle(x)
                                   <= mincycle,
                                   tiledic[tile])
            if len(tiledic[tile]) == 0:
                del tiledic[tile]
            #print tile, tiledic[tile]
    
    return validated


def missing_files_error():
    """error message (image files)"""
    common.error("Not enough image files found.")
    sys.exit(1)


def get_qc_filenames(dic, FileNameClass, datafolder):
    """Generate a list of all used lanes in the experiment, and a
    list of the corresponding filenames for QC files."""
    lanes = []
    for tilestr in dic.keys():
        lane = FileNameClass.lane(tiledic[tilestr][0]) 
        if not lane in lanes:
            lanes.append(lane)

    filenames = []
    folder = os.path.join(datafolder, runfolder.analysissubdir)
    tilestr = FileNameClass.tilestr(tiledic.values()[0][0])
    prefix = tilestr.split("_")[0] 
    for lane in lanes:
        filenames.append(os.path.join(folder, "%s_%i_qcm.xml" %
                                      (prefix, lane+1)))
    return lanes, filenames


def get_read_length(tiledic, FileNameClass):
    length = 1
    tiles = tiledic.keys()
    if len(tiles) > 0:
        length = get_cycle_params(tiledic[tiles[0]], FileNameClass)[0]
    return length


def create_instrument_offset_file(analysis_folder, instrument_folder,
                                  instrument):
    print "Auto-generating offset file..."
    path = runfolder.get_instrument_folder(analysis_folder,
                                           instrument_folder)
    
    if not os.path.exists(path):
        try:
            os.mkdir(path)
        except:
            common.error("Failed to create %s." % path)
            sys.exit(1)
    path = os.path.join(path, instrument)
    if not os.path.exists(path):
        try:
            os.mkdir(path)
        except:
            common.error("Failed to create %s." % path)
            sys.exit(1)
    filename = os.path.join(os.path.dirname(os.path.abspath(sys.argv[0])),
                            runfolder.offsetfile)
    try:
        shutil.copyfile(filename,
                        os.path.join(path, runfolder.offsetfile))
    except:
        common.error("Failed to copy %s to %s."
                     % (filename, os.path.join(path, runfolder.offsetfile)))
        sys.exit(1)
    return os.path.join(path, runfolder.offsetfile)
        

def get_lanes(selected_tiles):
    dic = {}
    for tile in selected_tiles:
        if tile[0] == '_':
            for i in range(runfolder.maxlane):
                dic[i] = 1
        else:
            try:
                lane = int(tile.split("_")[1]) - 1
            except:
                lane = -1
            if lane >= 0:
                dic[lane] = 1
    return dic.keys()


def select_lanes(folders, selected_tiles):
    lanes = get_lanes(selected_tiles)
    if lanes:
        folders = filter(lambda x: runfolder.LaneDirectory.lane(x) in
                         lanes or runfolder.LaneDirectory.lane(x) < 0,
                         folders)
    return folders


def run_lawn_analysis(dic, datafolder):
    """Simplified analysis for chips that have a dense mono-template
    instead of clusters."""
    for tile in tiledic.keys():
        outputfile = os.path.join(datafolder, tile + "_out.txt")
        for file in dic[tile]:
            commandline = "%s -v -v %s >> %s" %\
                          (runfolder.get_executable(image2_executable),
                           file, outputfile)
            print commandline
            os.system(commandline)

def processParamsFiles(folder, tiledic, image_params):
    if not image_params.AutoSizeFlag:
        for lane in runfolder.get_lane_samples(tiledic.keys()).keys():
            paramsfile = os.path.join(folder, "params%i.xml" % (lane+1))
            params = { "Fwhm":              image_params.Fwhm,
                       "RemappingDistance": image_params.RemappingDistance }
            XMLSimple.XMLout(params, RootName="ImageParameters",
                             OutputFile = paramsfile)

def mergeCycleRangePair(cycle_range1, cycle_range2, restarts):
    offset = cycle_range1[-1]
    restart = offset + 1
    restarts.append(restart)
    for i in range(len(cycle_range2)):
        cycle_range2[i] = cycle_range2[i] + offset
    cycle_range1 = cycle_range1 + cycle_range2
    return cycle_range1, restarts


def mergeCycleRanges(cycle_ranges, restarts):
    if (len(cycle_ranges) == 1):
        return cycle_ranges[0], restarts
    cycle_range1 = cycle_ranges[0][:]
    for cr in range(1, len(cycle_ranges)):
        cycle_range1, restarts = mergeCycleRangePair(cycle_range1, cycle_ranges[cr], restarts)
    return cycle_range1, restarts


def getCycleRange(tileimagelist):
    """For a given Run-Folder, determine the cycle range."""    
    allfilenames = reduce(lambda x, y: x+y, tileimagelist)    
    folders = map(runfolder.get_bottom_level_folder, allfilenames)    
    cycles = filter(None, map(runfolder.CycleDirectory.cycle, folders))
    
    # remove all multiple references to the same cycle
    unique_cycles = common.unique_list(cycles)
    unique_cycles.sort()
    
    return unique_cycles


def getFullCycleRange(tiledics, restarts):
    """Given two or more Run-Folders, determine the full cycle range."""
    if (len(tiledics) == 1):
        cyclerange = getCycleRange(tiledics[0].values())
        return cyclerange, restarts
    cycle_ranges = []
    for td in range(len(tiledics)):
        cycle_ranges.append(getCycleRange(tiledics[td].values()))
    fullcyclerange, restarts = mergeCycleRanges(cycle_ranges, restarts)
    return fullcyclerange, restarts


def mergeTileDicPair(tiledic1, tiledic2):
    for k in tiledic1.keys():
        tiledic1[k] = tiledic1[k] + tiledic2[k]
    return tiledic1


def linkRunFolders(tiledics):
    """Given two or more Run-Folders, merge the tile dictionaries."""
    if (len(tiledics) == 1):
        return tiledics[0]
    tiledic1 = tiledics[0].copy()
    for td in range(1, len(tiledics)):
        tiledic1 = mergeTileDicPair(tiledic1, tiledics[td])
    return tiledic1


def getTileDic(run_folder, run_params):
    """For a given Run-Folder, define a tile dictionary."""
    folders = runfolder.find_image_folders(run_folder)
    
    if run_params.SelectedTiles:
        # this is just a hack to avoid reading unneeded folders
        # tiles get filtered anyway by the select_tiles function
        folders = select_lanes(folders, run_params.SelectedTiles)
    if run_params.SelectedCycles:
        folders = filter(lambda x: runfolder.CycleDirectory.cycle(x)
                         in run_params.SelectedCycles, folders)
    if not run_params.Deblocked:
        folders = filter(lambda x: not runfolder.CycleDirectory.is_deblocked(x),
                         folders)
        
    folders = runfolder.select_image_folders(folders,
                                             run_params.SelectedFolders)

    filenames = get_tiff_files(folders)
    filenames += map(os.path.abspath,
                     filter(lambda x: x.find(".tif") > 0, args))
    if len(filenames) <= 0:
        missing_files_error()

    # detect filenaming convention
    # (this is used to support legacy instruments)
    FileNameClass = runfolder.FileName
    for nameclass in [ ManteiaFilePath, SwerdlotronFileName,
                       runfolder.RowColumnFileName ]:
        if nameclass.match(filenames[0]):
            print "%s detected." % nameclass.instrument()
            FileNameClass = nameclass
    
    filenames = filter(FileNameClass.match, filenames)

    # exclude filenames with spaces
    spacefilenames = filter(lambda x: x.find(" ") >= 0, filenames)
    if (spacefilenames):
        common.warning("Detected spaces in filenames \"%s\"."
                       "This is not supported by the pipeline and "
                       "the files are removed from the analysis."
                       % "\",\"".join(spacefilenames))
    filenames = filter(lambda x: x.find(" ") < 0, filenames)

    # sort by filename (ignoring the full path)
    filenames.sort(FileNameClass.cmp)

    # group filenames by position on chip
    tiledic = common.group_dictionary(FileNameClass.tilestr, filenames)
    
    # remove unselected positions
    tiledic = common.select_tiles(tiledic, run_params.SelectedTiles)        

    # remove positions with too few files
    if (run_params.MinCycle > 0):
        print "Removing tiles with less than %i files" % run_params.Mincycle
    for key in tiledic.keys():
        if len(tiledic[key]) < run_params.MinCycle:
            del tiledic[key]
    if (len(reduce(lambda x, y: x+y, tiledic.values())) <= 0):
        missing_files_error()

    # discard sequences that start with a deblocked cycle,
    # since there is no clue as to what base has been incorporated
    for key in tiledic.keys():
        if FileNameClass.is_deblocked(tiledic[key][0]) != 0:
            del tiledic[key]
    if (len(tiledic.values()) <= 0):
        missing_files_error()

    # count number of enzymology cycles and
    # deblocked cycles per position
    if not validate_cycles(tiledic, FileNameClass, run_params.AutoCycleFlag):
        sys.exit(1)
    maxcycle = getCycleRange(tiledic.values())
    run_params.SelectedCycles = filter(lambda x: x <= maxcycle[-1],
                                       run_params.SelectedCycles)

    return (tiledic, FileNameClass)

    
def usage():
        print "Usage: goat_pipeline.py [--tiles=<tile1>[,<tile2>[,...]]] \
[--cycles=<cycle1>[-cycle2][,cycle3][...]] \
[--make] [--deblocked] [--nobasecall] \
[--offsets=<offsets filesname>|auto] \
[--matrix=<matrix filename>|auto] \
[--phasing=<phasing rate>|auto|auto1|...] [--prephasing=<prephasing rate>] \
[--reset-phasing-cycle=<restart cycle>] \
[--GERALD=<config file>] \
<Run-Folder1> [...]"


#
# -- main program --
#
if __name__ == "__main__":

    print "Starting Genome Analyzer Pipeline (goat), %s version %s"\
          % (image_software, image_software_version)

    runfolder.validate_executables([ image_executable, image2_executable,
                                     matrix1_executable,
                                     bustard.basecaller_executable,
                                     bustard.phasing1_executable ])
    
    #-------------------------------------------------------------------#
    run_params    = common.GeneralParams()
    image_params  = ImageParams()
    base_params   = bustard.BaseCallParams()
    gerald_params = gerald.GeraldParams()
    
    # parsing command-line options
    try:
        (options, args) = getopt.gnu_getopt(sys.argv[1:], ":m",
                                            ["offsets=",
                                             "lawn",
                                             "directory=" ]
                                            + run_params.options
                                            + image_params.options
                                            + base_params.options
                                            + gerald_params.options)
        run_params.parseImageFolders(args)
    except:
        common.error("Failed to parse command-line options.")
        usage()
        sys.exit(1)
        
    print options, args

    offsetfile          = ""
    old_analysis_folder = ""
    lawnflag            = 0

    # parse command-line options
    run_params.parse(options)
    image_params.parse(options)
    base_params.parse(options)
    gerald_params.parse(options)
    for o, a in options:
        if o == "--lawn":
            lawnflag = 1
        if o == "--directory":
            # choose an existing analysis directory
            old_analysis_folder = os.path.abspath(a)
            if not os.path.exists(old_analysis_folder):
                common.error("Cannot find analysis directory %s." %
                             old_analysis_folder)
                sys.exit(2)
            run_params.BasecallFlag = 0
                
    #-------------------------------------------------------------------#
            
    starttime = runfolder.get_date_time()

    FileNameClass = runfolder.FileName

    # Tile dictionaries
    #
    # case `single Run-Folder` (standard case):
    #       the list 'tiledics' contains only one dictionary
    # case `multi Run-Folders` (e.g. paired reads):
    #       the list 'tiledics' contains two or more dictionaries
    tiledics = []
    
    for run_folder in args:
        tiledic, FileNameClass = getTileDic(run_folder, run_params)
        tiledics.append(tiledic)

    tiledic = linkRunFolders(tiledics) 

    print_cycle_overview(tiledic, FileNameClass)
    # TODO: cycle selection broken for more than one run-folder

    # Cycle range; cycle(s) at which the phasing/prephasing effect is
    # reset
    run_restarts = []
    cyclerange, run_restarts = getFullCycleRange(tiledics,
                                                 run_restarts)
    run_params.completeReads(cyclerange[-1], run_restarts)
    #print common.getDict(run_params)

    # Analysis folder 
    analysis_folder =\
                runfolder.get_new_analysis_folder(os.path.abspath(tiledic.values()[0][0]),
                                                  cyclerange, image_software,
                                                  image_software_version)
    run_folder = runfolder.get_run_folder(analysis_folder)
    # if some analysis already exists, identify the directories
    if old_analysis_folder:
        print  "Original analysis folder: %s" % old_analysis_folder
    print  "Analysis folder: %s" % analysis_folder

    # Sequence folder
    if run_params.BasecallFlag:
        sequence_folder =\
                runfolder.get_new_sequence_folder(old_analysis_folder
                                                  or analysis_folder,
                                                  bustard.basecaller_software,
                                                  bustard.basecaller_software_version)
        sequence_folder = os.path.join(analysis_folder,
                                       os.path.split(sequence_folder)[1])
        print  "Sequence folder: %s" % sequence_folder

    # Instrument
    run_params.printInstrument()
    run_params.printReads()


    # Offset data
    image_params.completePath(run_params)
    image_params.printOffsets()

    # Matrix
    base_params.completePath(run_params)
    base_params.printParams()
    base_params.printRestarts(run_params)
    #print run_params.getDict()
    #print base_params.getDict()
    #print base_params.getMatrixCycles(run_params)
    #print base_params.getPhasingRestarts(run_params)

    # Gerald configuration file
    if not run_params.MakeFlag and run_params.BasecallFlag:
        if (gerald.processGeraldConfig(run_params.MakeFlag,
                                       gerald_params.geraldconfigfiles, 
                                       run_params, sequence_folder) != 0):
            common.error("Failed to process Gerald configuration.")
            sys.exit(1)

    #--
    if not run_params.MakeFlag:
        print ("You can generate the Makefiles by using the \"--make\" "
               "option.")
        sys.exit(0)
    #--

    #-------------------------------------------------------------------#

    # Create output folders
    tiles = tiledic.keys()
    tiles.sort()

    if old_analysis_folder:
        rename_image_folders(old_analysis_folder, analysis_folder)
    else:
        create_image_folders(analysis_folder, tiles)
    processParamsFiles(analysis_folder, tiledic, image_params)

    if run_params.BasecallFlag:
        # Base-caller folders
        bustard.createBasecallerFolders(sequence_folder, tiledic.keys())

        # Matrix
        if base_params.Matrix[0] and not base_params.Matrix[0].AutoFlag:
            for lane in runfolder.get_lane_samples(tiledic.keys()).keys():
                matrixfile2 = os.path.join(sequence_folder,
                                           "matrix%i.txt" % (lane+1))
                shutil.copyfile(base_params.Matrix[0].Path, matrixfile2)

        # Phasing/prephasing effect        
        bustard.processParamsFiles(sequence_folder, tiledic, base_params)

    # run lawn analysis
    if lawnflag:
        run_lawn_analysis(tiledic, analysis_folder)
        sys.exit(0)

    ncolours = 4
    
    print "Generating journals, Makefiles and parameter files ..."
    
    #-------------------------------------------------------------------#

    # generating journal files
    for tilestr in tiles:
        sample, lane, tile = runfolder.FileName.split_tile(tilestr)
        
        files = tiledic[tilestr]        
        firstcycle = 0

        run_generate_journal(analysis_folder, sample, lane, tile,
                             files[ncolours*firstcycle:], FileNameClass)

    #-------------------------------------------------------------------#

    # generating Makefiles    
    makefile.generate_makefile(tiledic, analysis_folder, FileNameClass,
                               run_params, image_params,
                               base_params.getMatrixCycles(run_params),
                               old_analysis_folder,)

    if run_params.BasecallFlag:
        makefile.generate_basecaller_makefile(tiledic, analysis_folder,
                                              sequence_folder,
                                              run_params, base_params)
        
    #-------------------------------------------------------------------#
    
    # generating .params file for intensity remapping
    imagefolders = get_bottom_level_folders(reduce(lambda x, y: x + y,
                                                   tiledic.values()))
    runfolder.write_analysis_params(analysis_folder, tiles,
                                    cyclerange, imagefolders,
                                    image_software, image_software_version,
                                    runfolder.get_login(),
                                    run_params, image_params,
                                    starttime, old_analysis_folder)

    #-
    if not run_params.BasecallFlag:
        sys.exit(0)
    #-

    # generating .params file for base-calls
    runfolder.write_sequence_params(analysis_folder, sequence_folder,
                                    tiles, cyclerange,
                                    bustard.basecaller_software,
                                    bustard.basecaller_software_version,
                                    runfolder.get_login(),
                                    run_params, base_params)

    #-------------------------------------------------------------------#

    # processing Gerald configuration file
    if (gerald.processGeraldConfig(run_params.MakeFlag,
                                   gerald_params.geraldconfigfiles,
                                   run_params, sequence_folder) != 0):
        common.error("Failed to process Gerald configuration.")
        sys.exit(1)
