#! /usr/bin/env python

#  Time-stamp: <17 Sep 2007 14:21 BST runfolder.py km>
#
# Copyright (c) Solexa 2005, 2006
# Author: Klaus Maisinger
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

import sys, os, os.path, time, re, glob,types
import matrix, XMLSimple, common

"""Implementation of the Genome Analyzer runfolder specification."""

imagedir      = "Images"
datadir       = "Data"
logdir        = "Logs"
instrumentdir = "Instruments"
analysissubdir  = "Firecrest"
sequencesubdir  = "Bustard"
phasingdir    = "Phasing"
matrixdir     = "Matrix"
offsetdir     = "Offsets"
sizedir       = "Sizes"
focusdir      = "Focus"
base_order    = [ 'A', 'C', 'G', 'T' ]
offsetfile    = "default_offsets.txt"
matrixfile    = "matrix.txt"
maxlane       = 8
defaultsample = "s"

def validate_executables(executables):
    for file in executables:
        path = get_executable(file)
        if not os.path.exists(path):
            common.error("Missing executable %s.\n"
                         "Your pipeline installation is incomplete. "
                         "Please see \"docs/Pipeline installation.html\" "
                         "and fix the installation.\n" % path)
            sys.exit(1)

def get_platform_executable(filename):
    """Adjust name of executable on Windows operating system."""
    result = filename
    if os.name == "nt":
        (name, ext) = os.path.splitext(result)
        if ext not in ( ".exe", ".py", ".pl" ):
            result = name + ".exe"
        if ext == ".py":
            result = "python " + result
    return result
    

def get_executable(file):
    """Return the full path of an exectuable in the current working
    directory."""
    result = os.path.join(os.path.dirname(os.path.abspath(sys.argv[0])),
                          file)
    return get_platform_executable(result)

def get_bottom_level_folder(filename):
    """Given a full path, return the bottom-level directory in the
    folder hierarchy."""
    parts = os.path.split(os.path.abspath(filename))[0].split(os.sep)
    if len(parts):
        return parts[-1]
    return ""


class LaneDirectory:
    wildcard = "[lL][0-9][0-9][0-9]"
    cd = re.compile("[lL](\d{3})")
    def match(path):
        parts = path.split(os.sep)
        for i in range(1, min(len(parts), 3)):
            b = LaneDirectory.cd.match(parts[-i])
            if b:
                return [ b.group(1) ] 
        return None
    match = staticmethod(match)
    def lane(path):
        b = LaneDirectory.match(path)
        if b:
            return int(b[0])-1
        return -1
    lane = staticmethod(lane)
    def generate(lane, sample=""):
        return "L%03i" % (lane + 1)
    generate = staticmethod(generate)

    
class CycleDirectory:
    wildcard = "[cdCD]*.*"
    cd = re.compile("([cdCD])(\d{1,2})\.(\d{1,2})$")
    def match(filename):
        if os.path.split(filename)[1]:
            filename = os.path.split(filename)[1]
        b = CycleDirectory.cd.match(filename)
        if b:
            return [b.group(1), b.group(2), b.group(3)]
        return None
    match = staticmethod(match)
    def cycle(filename):
        b = CycleDirectory.match(filename)
        if b:
            return int(b[1])
        return 1
    cycle = staticmethod(cycle)
    def revision(filename):
        b = CycleDirectory.match(filename)
        if b:
            return int(b[2])
        return 0
    revision = staticmethod(revision)
    def is_deblocked(filename):
        b = CycleDirectory.match(filename)
        if not b:
            return 0
        if os.path.split(filename)[1]:
            filename = os.path.split(filename)[1]
        return (filename[0] in ['d', 'D'])
    is_deblocked = staticmethod(is_deblocked)


class FileName:
    cd = re.compile("([^_]*)_(\d)_(\d{1,4})")
    
    def match(filename):
        s = os.path.split(filename)[1]
        s = os.path.splitext(s)[0]
        b = FileName.cd.match(s)
        if b:
            return [ b.group(1), b.group(2), b.group(3) ]
        return None
    match = staticmethod(match)
    
    def cmp(filename1, filename2):
        path1 = get_bottom_level_folder(filename1)
        path2 = get_bottom_level_folder(filename2)
        c = cmp(CycleDirectory.cycle(path1), CycleDirectory.cycle(path2))
        if c:
            return c
        c = cmp(CycleDirectory.is_deblocked(path1),
                CycleDirectory.is_deblocked(path2))
        if c:
            return c
        c = cmp(FileName.tile(filename1), FileName.tile(filename2)) 
        if c:
            return c
        name1 = os.path.splitext(os.path.split(filename1)[1])[0]
        name2 = os.path.splitext(os.path.split(filename2)[1])[0]
        elements1 = name1.split("_")
        elements2 = name2.split("_")
        if (len(elements1) > 4) and (len(elements1) > 4):
            # if the filter number is given
            c = cmp(int(elements1[-2]), int(elements2[-2]))
        else:
            # sort by base
            c = cmp(elements1[-1], elements2[-1])
        return c
    cmp = staticmethod(cmp)

    def tilestr(filename):
        elements = os.path.split(filename)[1].split("_")[0:3]
        # work around alpha tile names
        elements[2] = "%04i" % (FileName.tile(filename)+1)
        return "_".join(elements)
    tilestr = staticmethod(tilestr)

    def sample(filename):
        return os.path.split(filename)[1].split("_")[0]
    sample = staticmethod(sample)
        
    def lane(filename):
        b = FileName.match(filename)
        if b and b[1]:
            return int(b[1])-1
        return -1
    lane = staticmethod(lane)

    def tile(filename):
        return (int(os.path.split(filename)[1].split("_")[2]) - 1)
    tile = staticmethod(tile)

    def cycle(filename):
        if (len(os.path.split(filename)[1].split("_")) > 4):
            return (int(os.path.split(filename)[1].split("_")[3]))
        else:
            return CycleDirectory.cycle(get_bottom_level_folder(filename))
    cycle = staticmethod(cycle)

    def base(filename):
        return os.path.splitext(os.path.split(filename)[1])[0].split("_")[-1]
    base = staticmethod(base)

    def split_tile(filename):
        return (FileName.sample(filename), FileName.lane(filename),
                FileName.tile(filename))
    split_tile = staticmethod(split_tile)

    def is_deblocked(filename):
        return CycleDirectory.is_deblocked(\
            os.path.split(filename)[0].split(os.sep)[-1])
    is_deblocked = staticmethod(is_deblocked)

    def name(sample, id, lane, tile, cycle=-1):
        if (cycle >= 0):
            return "%s_%i_%04i_%02i_%s.txt" % (sample, lane+1, tile+1,
                                               cycle+1, id)
        else:
            return "%s_%i_%04i_%s.txt" % (sample, lane+1, tile+1, id)
    name = staticmethod(name)


class RowColumnFileName:
    """New naming convention with rows and columns of tiles."""
    cd = re.compile("([^_]*)_(\d)_(\d)_(\d{1,4})")
    def match(filename):
        s = os.path.split(filename)[1]
        s = os.path.splitext(s)[0]
        b = RowColumnFileName.cd.match(s)
        if b:
            return [ b.group(1), b.group(2), b.group(3), b.group(4) ]
        return None
    match = staticmethod(match)

    def cmp(filename1, filename2):
        path1 = get_bottom_level_folder(filename1)
        path2 = get_bottom_level_folder(filename2)
        c = cmp(CycleDirectory.cycle(path1), CycleDirectory.cycle(path2))
        if c:
            return c
        c = cmp(CycleDirectory.is_deblocked(path1),
                CycleDirectory.is_deblocked(path2))
        if c:
            return c
        c = cmp(RowColumnFileName.tile(filename1),
                RowColumnFileName.tile(filename2)) 
        if c:
            return c
        name1 = os.path.splitext(os.path.split(filename1)[1])[0]
        name2 = os.path.splitext(os.path.split(filename2)[1])[0]
        elements1 = name1.split("_")
        elements2 = name2.split("_")
        if (len(elements1) > 5) and (len(elements1) > 5):
            # if the filter number is given
            c = cmp(int(elements1[-2]), int(elements2[-2]))
        else:
            # sort by base
            c = cmp(elements1[-1], elements2[-1])
        return c
    cmp = staticmethod(cmp)

    def tilestr(filename):
        elements = os.path.split(filename)[1].split("_")[0:3]
        # work around alpha tile names
        elements[2] = "%04i" % (RowColumnFileName.tile(filename)+1)
        return "_".join(elements)
    tilestr = staticmethod(tilestr)

    def sample(filename):
        return RowColumnFileName.match(os.path.split(filename)[1])[0]
    sample = staticmethod(sample)
        
    def lane(filename):
        return (int(RowColumnFileName.match(os.path.split(filename)[1])[1])
                - 1)
    lane = staticmethod(lane)

    def tile(filename):
        return ((RowColumnFileName.tilecolumn(filename)+1) * 1000
                + RowColumnFileName.tilerow(filename))
    tile = staticmethod(tile)

    def tilerow(filename):
        return (int(RowColumnFileName.match(os.path.split(filename)[1])[3])
                - 1)
    tilerow = staticmethod(tilerow)

    def tilecolumn(filename):
        return (int(RowColumnFileName.match(os.path.split(filename)[1])[2])
                - 1)
    tilecolumn = staticmethod(tilecolumn)

    def cycle(filename):
        if (len(os.path.split(filename)[1].split("_")) > 5):
            return (int(os.path.split(filename)[1].split("_")[4]))
        else:
            return CycleDirectory.cycle(get_bottom_level_folder(filename))
    cycle = staticmethod(cycle)

    def base(filename):
        return os.path.splitext(os.path.split(filename)[1])[0].split("_")[-1]
    base = staticmethod(base)

    def split_tile(filename):
        return (RowColumnFileName.sample(filename),
                RowColumnFileName.lane(filename),
                RowColumnFileName.tile(filename))
    split_tile = staticmethod(split_tile)

    def is_deblocked(filename):
        return CycleDirectory.is_deblocked(\
            os.path.split(filename)[0].split(os.sep)[-1])
    is_deblocked = staticmethod(is_deblocked)

    #def name(sample, id, lane, tile, cycle=-1):
    #    if (cycle >= 0):
    #        return "%s_%i_%04i_%02i_%s.txt" % (sample, lane+1, tile+1,
    #                                           cycle+1, id)
    #    else:
    #        return "%s_%i_%04i_%s.txt" % (sample, lane+1, tile+1, id)
    #name = staticmethod(name)

    def instrument():
        return "Genome Analyzer, row-column format"
    instrument = staticmethod(instrument)


def find_data_offset_file(datafolder):
    filename = os.path.join(get_data_folder(datafolder), offsetfile)
    if os.path.exists(filename):
        return filename
    return ""

def find_instrument_offset_file(datafolder, instrumentfolder="",
                                instrument = ""):
    filename = os.path.join(find_instrument_folder(datafolder,
                                                   instrumentfolder,
                                                   instrument),
                            offsetfile)
    if os.path.exists(filename):
        return filename
    return ""

def find_offset_file(datafolder, instrumentfolder="", instrument = ""):
    filename = (find_data_offset_file(datafolder) or
                find_instrument_offset_file(datafolder, instrumentfolder,
                                            instrument))
    return filename

def find_instrument_matrix_file(datafolder, instrumentfolder="",
                                instrument = ""):
    filename = os.path.join(find_instrument_folder(datafolder,
                                                   instrumentfolder,
                                                   instrument),
                            matrixfile)
    if os.path.exists(filename):
        return filename
    return ""

def get_run_folder(path):
    """Given any file inside a runfolder, return the path to the run
    folder."""
    path = os.path.abspath(path)
    hierarchy = path.split(os.sep)
    if os.path.isfile(path):            # keep only directories
        hierarchy = hierarchy[:-1]
    level = len(hierarchy)
    for i in range(len(hierarchy)-1, -1, -1):
        if hierarchy[i] in [ imagedir, logdir ]:
            level = i
            break
        # this extra check is to work around different filenames
        # for the Swerdlotron; there may be a "Data" folder elsewhere
        # in the hierarchy
        if hierarchy[i] in [datadir]:
            if i == len(hierarchy)-1 or hierarchy[i+1][0] in [ 'c', 'C']:
                level = i
                break
    return os.sep.join(hierarchy[:level])
    
def get_image_folder(path):
    """Given any file inside a runfolder, return the path to the image
    folder."""
    return os.path.join(get_run_folder(path), imagedir)
        
def get_instrument_folder(run_folder, instrument_folder, instrument=""):
    """Return the full path to a the instrument folder corresponding to a
    given instrument"""
    #print "Instrument folder:", filename
    if instrument_folder and os.path.exists(instrument_folder):
        path = instrument_folder
    else:
        path = os.path.join(os.path.split(get_run_folder(run_folder))[0],
                            instrumentdir)
    if instrument:
        path = os.path.join(path, instrument)
    return path

def find_instrument_folder(run_folder, instrument_folder, instrument):
    """Return the full path to a the instrument folder corresponding to a
    given instrument if it exists, checking for different capitalisations."""
    path = get_instrument_folder(run_folder, instrument_folder, instrument)
    if os.path.exists(path):
        return path
    dir, name = os.path.split(path)
    if not os.path.exists(dir):
        common.warning(("Instruments folder %s not found!\n" % dir) +
                       ("*"*75+"\n"
                         "It is recommended that you create the folder.\n"
                         + "*"*75))
        return ""
    newpath = os.path.join(dir, name.upper())
    if os.path.exists(newpath):
        return newpath
    newpath = os.path.join(dir, name.capitalize())
    if os.path.exists(newpath):
        return newpath
    newpath = os.path.join(dir, name.lower())
    if os.path.exists(newpath):
        return newpath
    common.warning(("Instrument folder %s not found!\n" % path) +
                   ("*"*75+"\n"
                     "It is recommended that you create the folder "
                     "and copy a matrix file and default offset "
                     "file to the folder.\n"
                     + "*"*75))
    return ""

def get_data_folder(path):
    """Return the full path to a the data folder corresponding to any
    given folder within a runfolder"""
    return os.path.join(get_run_folder(path), datadir)

def get_lane_subfolder(path, subfolder, sample, lane, tile=0):
    folder = os.path.join(path, subfolder)
    folder = os.path.join(folder, LaneDirectory.generate(lane))
    return folder

def find_image_folders(run_folder):
    """Return all image folders in in the folder list args or in
    subfolders of these folders"""
    directories = []
    if os.path.isdir(run_folder):
        path = os.path.abspath(run_folder)
        hierarchy = path.split(os.sep)
        if not imagedir in hierarchy:
            hierarchy.append(imagedir)
            if not os.path.isdir(os.sep.join(hierarchy)):
                directories.append(path)
                
        index = hierarchy.index(imagedir)
        allpaths = []
        if (index < len(hierarchy)-1):
            allpaths = [ os.sep.join(hierarchy[:index+2]) ]
        else:
            # first identify lane directories
            wildcards = LaneDirectory.wildcard
            imagepath = os.sep.join(hierarchy[:index+1] + [wildcards])
            paths = glob.glob(imagepath)
            # if no folder found, try old runfolder scheme
            if not paths:
                paths = [ os.sep.join(hierarchy[:index+1]) ]
            allpaths = []
            for path in paths:
                imagepath = os.path.join(path, CycleDirectory.wildcard)
                allpaths += glob.glob(imagepath)
        #print allpaths
        for path in allpaths:
            if (CycleDirectory.match(path)):
                directories.append(path)
                
    return directories


def select_cycle_folders(folders, selected_folders):
    """Given a list of image folders, select only one folder per
    cycle. this does not distinguish between proper cycles and
    deblocking scans"""
    
    if not len(folders):
        return []
    
    # group folders by lanes if applicable
    lanes = {}
    for folder in folders:
        key = LaneDirectory.lane(folder)
        lanes.setdefault(key, []).append(folder)
    
    # now loop over all lanes and pick last iteration
    results = []
    for key in lanes.keys():
        folders = lanes[key]
        maxcycle = max(map(CycleDirectory.cycle, folders))
        list = [ [] for i in range(maxcycle+1) ]
        for folder in folders:
            list[CycleDirectory.cycle(folder)].append(folder)
        for i in range(len(list)):
            if len(list[i]) > 1:
                for folder in list[i]:
                    id = os.path.split(folder)[1] or folder
                    if id in selected_folders:
                        list[i] = [ folder ]
                        continue
                maxrevision = CycleDirectory.revision(list[i][0])
                maxfolder = list[i][0]
                for folder in list[i]:
                    if CycleDirectory.revision(folder) > maxrevision:
                        maxrevision = CycleDirectory.revision(folder)
                        maxfolder = folder
                list[i] = [maxfolder]
        results += reduce(lambda x, y: x+y, list)
    return results
    
def select_image_folders(folders, selected_folders = []):
    """Returns a reduced list of folders where for each cycle there is
    only one scan. As a default, the folder with the latest revision
    number is chosen. Any folder name appearing in a whitelist
    is automatically selected."""
    return (select_cycle_folders(filter(CycleDirectory.is_deblocked,
                                        folders), selected_folders)
            + select_cycle_folders(filter(lambda x: not
                                          CycleDirectory.is_deblocked(x),
                                          folders), selected_folders))

def read_default_offsets(filename):
    offsets = []
    try:
        for line in open(filename).readlines():
            if line[0] in [ '#', '>' ]:
                continue
            offsets.append(map(float, line.split()))
    except:
        print "Could not load offsets file %s" % filename
    return offsets

def get_unique_path(path):
    """Return a unique name that refers to a new non-existing folder or
    file given by the path""" 
    newpath = os.path.normpath(path)
    index = 1
    while os.path.exists(newpath):
        index += 1
        newpath = os.path.normpath(path) + ".%i" % index
    return newpath

def get_login():
    """Return a unique user id string"""
    name = os.getenv("LOGNAME") or os.getenv("USERNAME") or "unknown"
    return name.replace(" ", "")

def get_new_analysis_folder(path, cyclerange, software, software_version):
    """Create a new unique name for the folder that the image
    analsysis software saves its output to."""
    # create a string to list all the cycles used in the analysis
    cyclelist = common.collapse_ranges(cyclerange)
    folderlist = []
    for cycle in cyclelist:
        if type(cycle) == types.ListType:
            folderlist.append("%i-%i" % (cycle[0], cycle[1]))
        else:
            folderlist.append("%i" % cycle)
    cyclestring = ",".join(folderlist)
            
    # query local date and user name
    times = time.localtime()
    date = "%02i-%02i-%04i" % (times[2], times[1], times[0])
    folder = "C%s_%s_%s_%s" % (cyclestring,
                               software+software_version,
                               date, get_login())  
    folder = os.path.join(get_data_folder(path), folder)
    return get_unique_path(folder)

def get_new_sequence_folder(analysisfolder, software, software_version):
    """Create a new unique name for the folder that the base caller save its
    image analysis output to."""
    times = time.localtime()
    date = "%02i-%02i-%04i" % (times[2], times[1], times[0])
    folder = "%s_%s_%s" % (software+software_version,
                           date, get_login())
    folder = os.path.join(analysisfolder, folder)
    return get_unique_path(folder)
    
def get_date_time():
    return time.strftime("%d-%m-%y %H:%M:%S %Z", time.localtime())

def merge_runs(params1, params2):
    """Merge two different runs in the same experiment"""
    runs = [ params1["Run"], params2["Run"] ]
    for i in range(len(runs)):
        if type(runs[i]) != types.ListType:
            runs[i] = [ runs[i] ]
    return { "Run" : runs[0] + runs[1] }

def get_run_params_filename(folder):
    """Return the name of the parameter file found in the runfolder
    top-level directory"""
    path = get_run_folder(folder)
    filename = os.path.split(path)[1] + ".params"
    return os.path.join(path, filename)

def get_params(paramsfile, expected_tag = "experiment"):
    """Return the parameters found in the top-level runfolder
    parameter file"""
    data = {}
    if type(expected_tag) != types.ListType:
        expected_tag = [ expected_tag ]
    if (os.path.exists(paramsfile)):
        try:
            data = XMLSimple.XMLin(paramsfile, KeepRoot=1)
            tag = data.keys()[0]
            data = data[tag]
            if tag not in expected_tag:
                common.warning("Wrong XML file format! Tag: %s != %s" %
                               (tag, expected_tag))
                data = {}
            # legacy compatibility
            if data.has_key("run") and not data.has_key("Run"):
                data["Run"] = data["run"]
            # force the "Run" key to be an array - makes access easier
            if data.has_key("Run") and (type(data["Run"]) == types.DictType):
                data["Run"] = [ data["Run"] ]
        except:
            common.warning("Error parsing XML file %s." % paramsfile)
    return data

def get_run(params, folder):
    """Return the run corresponding to the analysis folder from
    within a parameter hash"""
    folder = folder.split(os.sep)[-1]
    
    for dic in params.get("Run", []):
        if dic.has_key("Name") and dic["Name"] == folder:
            return dic
    return {}

def get_tile_selection_from_tiles(tiles):
    """Arrange a list of tiles in a format that can be serialised to
    XML and is more similar to the instrument XML output."""
    params = { "Lane": [] }
    lanes = {}
    samples = {}
    for tilestr in tiles:
        sample, lane, tile = FileName.split_tile(tilestr)
        lane += 1
        tile += 1
        lanes.setdefault(lane, []).append(tile)
        samples[lane] = sample
    for lane in lanes.keys():
        dic = { "Index" : lane,
                "Sample": samples[lane]}
        is_range = 1
        ranges = common.collapse_ranges(lanes[lane])        
        if type(ranges[0]) == types.ListType:
            dic["TileRange"] = { "Min": ranges[0][0],
                                 "Max": ranges[0][-1] }
        else:
            dic["Tile"] = ranges
        params["Lane"].append(dic)
    return params

def get_lane_samples(tiles):
    """Given a list of tiles, return a dictionary with the sample ids
    corresponding to the lanes."""
    lanes = {}
    for tilestr in tiles:
        sample, lane, tile = FileName.split_tile(tilestr)
        lanes[lane] = sample
    return lanes

def get_array_key(dic, key):
    if not dic.has_key(key):
        return []
    if type(dic[key]) != types.ListType:
        return [ dic[key] ]
    return dic[key]

def get_tiles_from_tile_selection(dic):
    """Parse the TileSelection XML tags for a list of tiles to be
    analysed."""
    if dic.has_key("TileSelection"):
        dic = dic["TileSelection"]
    if dic.has_key("Incorporation"):
        dic = dic["Incorporation"]
    tiles = []
    for laneitem in get_array_key(dic, "Lane"):
        sample = laneitem["Sample"]
        lane = laneitem["Index"] - 1
        tilelist = get_array_key(laneitem, "Tile")
        for trange in get_array_key(laneitem, "TileRange"):
            tilelist += range(trange["Min"], trange["Max"]+1)
        tilelist = map(lambda x: x-1, tilelist)
        for tile in tilelist:
            tilestr = FileName.tilestr(FileName.name(sample, "", lane,
                                                     tile))
            tiles.append(tilestr)
    return tiles

def write_analysis_params(analysisfolder, tiles,
                          cyclerange, imagefolders,
                          software, software_version, user,
                          run_params, image_params, starttime,
                          oldanalysisfolder = ""):
    offsetfile = image_params.OffsetFile
    offsets = read_default_offsets(offsetfile)
    offsetsx = map(lambda x: x[0], offsets[:4])
    offsetsy = map(lambda x: x[1], offsets[:4])
    offsetsx += [ 0.] * (4-len(offsetsx))
    offsetsy += [ 0.] * (4-len(offsetsy))
    imagedic = image_params.getDict()
    imagedic.update({"Offsets": [ {"X" :offsetsx[i], "Y": offsetsy[i] } for i in
                                  range(len(offsetsx)) ] })
    run = {
        "Name": os.path.split(analysisfolder)[1],
        "Software" : { "Name" : software, "Version":  software_version},
        "User" : {"Name": user},
        "RunParameters": run_params.getDict(),
        "ImageParameters": imagedic,
        "TileSelection" : get_tile_selection_from_tiles(tiles),
        
        "Time" : { "Start": starttime },
        "UsedCycleFolders": ",".join(imagefolders),
        "Cycles": {"First": cyclerange[0],
                   "Last": cyclerange[-1],
                   "Number" : len(cyclerange)},
        }
    if oldanalysisfolder:
        run["Previous"] = {"Name": os.path.split(oldanalysisfolder)[1] }
    data = { "Run" : run }
    
    paramsfile = os.path.join(get_data_folder(analysisfolder), ".params")
    olddata = get_params(paramsfile, ["data", "ImageAnalysis"])
    if olddata:
        data = merge_runs(olddata, data)
    keyattr = ["Name", "Version", "X", "Y", "Index", "Min", "Max",
               "Path", "Number", "First", "Last"]
    XMLSimple.XMLout(data, RootName="ImageAnalysis",
                     OutputFile = paramsfile,
                     KeyAttr=keyattr, XMLDecl=1)


def write_sequence_params(analysisfolder, sequencefolder,
                          tiles, cyclerange,
                          software, software_version, user,
                          run_params, base_params, oldsequencefolder=""):
    frequencymatrix = {}
    matrixfile = base_params.Matrix[0].Path
    if os.path.exists(matrixfile):
        frequencymatrix = matrix.read_matrix(matrixfile)
        frequencymatrix["Path"] = base_params.Matrix[0].Path
    basedic = base_params.getDict(run_params)

    # remove redundant tags
    for tag in [ "Phasing", "Matrix" ]:
        for item in basedic[tag]:
            for tag2 in [ "Duplicates", "PhasingRate",
                          "PrephasingRate"]:
                if item.has_key(tag2):
                    del item[tag2]

    basedic.update({"MatrixElements": frequencymatrix})
    run = {
        "Name": os.path.split(sequencefolder)[1],
        "Software" : { "Name" : software, "Version":  software_version},
        "User" : {"Name": user},
        "RunParameters": run_params.getDict(),
        "BaseCallParameters": basedic,
        "TileSelection" : get_tile_selection_from_tiles(tiles),
        "Time" : { "Start": get_date_time() },
        "Cycles": {"First": cyclerange[0],
                   "Last": cyclerange[-1],
                   "Number" : len(cyclerange)},
        "Input": {"Path": os.path.split(analysisfolder)[1] }
        }
    if oldsequencefolder:
        run["Previous"] = { "Name" : os.path.split(oldsequencefolder)[1]}
    data = { "Run" : run }
    paramsfile = os.path.join(analysisfolder, ".params")
    olddata = get_params(paramsfile, ["data", "BaseCallAnalysis"])
    if olddata:
        data = merge_runs(olddata, data)
    
    keyattr = ["Name", "Version", "X", "Y", "Index", "Min", "Max",
               "Path", "Number", "First", "Last"]
    XMLSimple.XMLout(data, RootName="BaseCallAnalysis",
                     OutputFile=paramsfile,
                     KeyAttr=keyattr, XMLDecl=1)

def paste(filenames, out):
    """Reimplementation of the Unix paste command to create a single
    file from multiple files with the same number of lines."""
    longlines = []
    first = 1
    for filename in filenames:
        try:
            lines = open(filename).readlines()
        except:
            common.error("Unable to read %s." % filename)
            continue
        if first:
            for line in lines:
                longlines.append([ line.rstrip()])
            first = 0
        else:
            if len(lines) != len(longlines):
                common.error("Line numbers do not match: %i != %i." %
                             (len(lines),  len(longlines)))
            for j in range(len(lines)):
                longlines[j].append(lines[j].rstrip())
    for longline in longlines:
        out.write("%s\n" % "\t".join(longline))
