/*****************************************************************************/
// Copyright (c) Illumina 2007
// Author: Richard Shaw
// This source file is covered by the "Solexa Public Source License"
// agreement and bound by the terms therein.
/*****************************************************************************/

#include <iostream>
#include <sstream>

#include "DataSetFactory.h"

/*****************************************************************************/

Data_Set_Factory::Data_Set_Factory(const std::string& data_file_name)
    : my_data_file_name(data_file_name)
{
    ;
}

/*****************************************************************************/

Data_Set_Factory::~Data_Set_Factory()
{
    if (my_data_file.eof()) {
        my_data_file.close();
    }
}

/*****************************************************************************/
// Read a DataSet from the file specified at construction. 
// See comments in the header file and the documentation for the data format.

bool Data_Set_Factory::get_next_data_set(DataSet*& data_set_ptr, bool& at_eof)
{
    at_eof = false;

    if (my_data_file.eof()) {
        at_eof = true;
        return false;
    }

    if (!my_data_file.is_open()) {
        my_data_file.open(my_data_file_name.c_str());

        if (!my_data_file.is_open()) {
            return false;
        }
    }

    data_set_ptr = new DataSet; // Caller must delete.
    const unsigned int max_num_records_per_data_set(DataSet::MAX_NUM_RECORDS);
    const unsigned int alloc_size(DataSet::ALLOC_SIZE);
    unsigned int num_records(0);

    int num_params(0);
    std::string line;
    BaseAlignment read;

    while (getline(my_data_file, line)) {
        if ((line.size() == 0) || (line[0] == '#') || (line[0]=='>')) {
            continue;
        }

        // Determine number of columns.
        if (num_params == 0) {
            std::stringstream ss(line);
            std::string s;

            while (ss >> s)
                ++num_params;

            num_params -= 2;  // last two columns hold base information
            data_set_ptr->values.resize(num_params);
        }

        // Reserve some memory in advance.
        if ((num_records % alloc_size) == 0) {
            for (int i(0); i < data_set_ptr->values.size(); i++)
                data_set_ptr->values[i]
                    .reserve(((num_records / alloc_size) + 1) * alloc_size);
        }

        // Parse line and store its fields.
        std::stringstream ss(line);
        std::vector<std::string> fields;
        std::string s;

        while (ss >> s) {
            fields.push_back(s);
        }

        if (fields.size() != (num_params + 2)) {
            std::cerr << "Invalid record: " << line << std::endl;
        }

        for (int i(0); i < num_params; i++) {
            const float value(atof(fields[i].c_str()));
            data_set_ptr->values[i].push_back(value);
        }

        read.base_read    = fields[fields.size() - 2].c_str()[0];
        read.base_aligned = fields[fields.size() - 1].c_str()[0];
        data_set_ptr->reads.push_back(read);

        if (++num_records == max_num_records_per_data_set) {
            // The DataSet is `full', so return it.
            // The caller should call this method again for further data.
            return true;
        }
    }

    // We are at the end of the file but it is left to the destructor
    // to close it - otherwise if this method is (incorrectly) called
    // again my_data_file.eof() would be false and it would be reopened.
    at_eof = true;
    return true;
}

/*****************************************************************************/
