/*
 * Copyright (c) Solexa 2006
 * Authors: Klaus Maisinger, A. J. Cox 
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

#include <iostream>
#include <cstdio>
#include <string>

#include "DataSetFactory.h"
#include "QualityTable.h"

/*****************************************************************************/

// Identifier to embed in source and binary files
static char* quality_apply_id="@(#) $Id: QualityApply.cpp,v 1.11 2007/12/04 09:45:38 rshaw Exp $";

/*
 * Apply a quality calibration table as calculated by QualityCalibration.cpp
 * to a data set.
 */

const int bufferSize(128);

enum { ERR_OK=0, ERR_ARGS=1, ERR_DATA_FILE=2 };

/*****************************************************************************/

class QualityWriter
{
public:
    virtual void operator()(std::string& s, bool isFirst, int score)=0;
protected:
    char buffer[bufferSize];
}; // ~class QualityWriter

/*****************************************************************************/

// write numerical output as per Klaus' original code
class QualityWriterNumeric : public QualityWriter
{
public:
    virtual void operator()(std::string& s, bool isFirst, int score)
    {
        snprintf(buffer, sizeof(buffer), " %i", score);
        s += (buffer+(isFirst?1:0));
    }
}; // ~class QualityWriterNumeric

/*****************************************************************************/

// write compact ASCII-encoded output
class QualityWriterSymbolic : public QualityWriter
{
public:
    virtual void operator()(std::string& s, bool isFirst, int score)
    {
        snprintf(buffer, sizeof(buffer), "%c", 64+score);
        s+=buffer;
    }
}; // ~class QualityWriterSymbolic

/*****************************************************************************/

int writeQualityOutput(FILE *fp,
                       DataSet *set,
                       std::vector<int> *scores,
                       QualityWriter& writeQuality,
                       std::string& buf_str,
                       int& prev_cycle,
                       bool is_first_data_set,
                       bool is_last_data_set)
{
    int i;
    std::vector<std::vector<float > > *values = set->getParameterValues();
    int cycle = 0;

    for (i = 0; i < scores->size(); i++) {
        cycle = int((*values)[0][i]);

        if (cycle < prev_cycle) {
            fprintf(fp, "%s\n", buf_str.c_str());
            buf_str = "";
        }
        writeQuality(buf_str,
                     (cycle < prev_cycle) || (is_first_data_set && (i==0)),
                     scores->at(i));
        prev_cycle = cycle;

        if (is_last_data_set && (i == scores->size() - 1)) {
            fprintf(fp, "%s\n", buf_str.c_str());
        }
    }
}

/*****************************************************************************/

int writeQualityOutput(const char filename[],
                       DataSet *set,
                       std::vector<int> *scores,
                       QualityWriter& writeQuality,
                       std::string& buf_str,
                       int& prev_cycle,
                       bool is_first_data_set,
                       bool is_last_data_set)
{
    FILE *fp = fopen(filename, "r");
    if (!fp)
        return -1;
    int result = writeQualityOutput(fp, set, scores, writeQuality,
                                    buf_str, prev_cycle,
                                    is_first_data_set, is_last_data_set);
    fclose(fp);
    return result;
}

/*****************************************************************************/

int usage(const std::string app_name, int ret_val)
{
    std::cout << "Usage: " << app_name << " [--symbolic|--numeric] "
              << "<quality-table> <qval-file1>" << std::endl;
    return ret_val;
}

/*****************************************************************************/

int main(int argc, char *argv[])
{
    const std::string app_name(argv[0]);
    bool wantNumeric(false);
    int ret_val(0);

    if (argc==4) {
        if ( (strcmp(argv[1],"--symbolic")!=0)
             && (strcmp(argv[1],"--numeric")!=0) ) {
            exit(usage(app_name, ERR_ARGS));
        } // ~if
        else {
            wantNumeric=(strcmp(argv[1],"--numeric")==0);
            argc--; argv++;
        } // ~else
    } // ~if
    else if (argc!=3) {
        exit(usage(app_name, ERR_ARGS));
    } // ~else if


    QualityTable *table = new QualityTable();
    QualityWriter* writer;
    if (wantNumeric) {
        writer= new QualityWriterNumeric();
    } // ~if
    else {
        writer= new QualityWriterSymbolic();
    } // ~else

    if (table->read(argv[1]) < 0) {
        exit(ERR_DATA_FILE);
    }

#if 0
    table->print();
#endif

    const std::string qval_file_path(argv[2]);
    Data_Set_Factory data_set_factory(qval_file_path);
    DataSet* data_set_ptr(0);
    bool at_eof(false);
    bool is_first_data_set(true);
    int detected_num_params(0);

    std::string buf_str;
    int prev_cycle(0);

    while (data_set_factory.get_next_data_set(data_set_ptr, at_eof)) {
        if (is_first_data_set) {
            detected_num_params = data_set_ptr->getNumberParameters();
            std::cerr << "Detected " << detected_num_params
                      << " parameters.\n";
        }

        std::vector<int> scores;

        table->applySolexa(data_set_ptr, scores);

        writeQualityOutput(stdout, data_set_ptr, &scores, *writer,
                           buf_str, prev_cycle,
                           is_first_data_set, at_eof);

        if (is_first_data_set) {
            is_first_data_set = false;
        }

        /* FIXME : rjs : reuse one DataSet to minimise reallocation? */
        delete data_set_ptr;
        data_set_ptr = 0;

        if (at_eof) {
            break;
        }
    }

    if (!at_eof) {
        // get_next_data_set call failed but not because of EOF
        // -> failed to open file at all.
        std::cerr << "** " << app_name << " : Failed to open file `" 
                  << qval_file_path << "'" << std::endl;
        ret_val = ERR_DATA_FILE;
    }

    delete table;
    delete writer;
    return ret_val;
} // ~main

/*****************************************************************************/
