/*
 * Copyright (c) Solexa 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

#include <cstdio>
#include <iostream>
#include <string>
#include <vector>

#include "DataSetFactory.h"
#include "QualityTable.h"
#include "XMLConfigHandler.h"
#include "XMLSAXParser.h"

/*****************************************************************************/

enum { ERR_OK=0, ERR_ARGS=1, ERR_CFG_FILE=2, ERR_XML_PARSE=3,
       ERR_DATA_FILE=4 };

/*****************************************************************************/

int usage(int ret_val)
{
    printf("Usage: QCalibration <cfg_file.xml> <qvalue-file1> "
           "[<qvalue-file2> ...]\n");
    return ret_val;
}

/*****************************************************************************/

int main(int argc, char *argv[]) 
{
    const std::string app_name(argv[0]);
    int ret_val(ERR_OK);

    if (argc < 3) {
        exit(usage(ERR_ARGS));
    }

    // Open config file.

    const std::string cfg_file_path_str(argv[1]);

    FILE* cfg_file_hndl = fopen(cfg_file_path_str.c_str(), "r");

    if (!cfg_file_hndl) {
        exit(usage(ERR_CFG_FILE));
    }

    // Parse config file.

    XMLConfigHandler cfg_handler;
    XMLSAXParser* sax_parser_ptr = new XMLSAXParser();
    sax_parser_ptr->setContentHandler(&cfg_handler);
    sax_parser_ptr->parse(cfg_file_hndl);
    fclose(cfg_file_hndl);

    Parameter_Config_List param_cfg_list;

    if (!cfg_handler.get_param_cfg_list(param_cfg_list)) {
        std::cerr << "Failed to extract parameter configurations from "
                  << cfg_file_path_str << std::endl;
        exit(ERR_XML_PARSE);
    }

    if (param_cfg_list.empty()) {
        std::cerr << "No parameter configurations found in "
            << cfg_file_path_str << std::endl;
        exit(ERR_XML_PARSE);
    }


    // Populate bin_boundaries.

    const int no_params = param_cfg_list.size();
    unsigned int param_ind(0);

    std::vector<std::vector<float> > bin_boundaries;
    bin_boundaries.resize(no_params);

    for (Parameter_Config_List_CIter param_cfg_citer = param_cfg_list.begin();
         param_cfg_citer != param_cfg_list.end();
         ++param_cfg_citer, ++param_ind) {
        param_cfg_citer->get_bin_boundary_vec(bin_boundaries[param_ind]);
    }

#if 0
    // DEBUG : Dump bin boundaries
    typedef std::vector<float> Float_Vec;
    typedef Float_Vec::const_iterator Float_Vec_CIter;

    for (int param_ind(0); param_ind < no_params; ++param_ind) {
        for (Float_Vec_CIter float_citer(bin_boundaries[param_ind].begin());
             float_citer != bin_boundaries[param_ind].end(); ++float_citer) {
            std::cout << " " << *float_citer;
        }

        std::cout << std::endl << std::endl;
    }
#endif

    // add parameters to table
    QualityTable *table = new QualityTable();

    for (int i(0); i < bin_boundaries.size(); i++)
        table->addParameter(bin_boundaries[i]);

    // fill table with values

    DataSet* data_set_ptr(0);

    for (int i(2); i < argc; i++) {
        fprintf(stderr, "%i %i %s\n", argc, i, argv[i]);

        const std::string curr_qval_file_path(argv[i]);
        Data_Set_Factory data_set_factory(curr_qval_file_path);
        bool at_eof(false);
        bool first_data_set(true);
        int detected_num_params(0);

        while (data_set_factory.get_next_data_set(data_set_ptr, at_eof)) {
            detected_num_params = data_set_ptr->getNumberParameters();

            if (first_data_set) {
                std::cerr << "Detected " << detected_num_params 
                          << " parameters.\n";
                first_data_set = false;
            }

            if (detected_num_params != no_params) {
                std::cerr << "Wrong number of parameters: "<< 
                    data_set_ptr->getNumberParameters() << std::endl;
                continue;
            }

            // This addData adds data in the sense of updating statistics
            // using it rather than allocating additional space for it.
            table->addData(data_set_ptr);

            /* FIXME : rjs : reuse one DataSet to minimise reallocation? */
            delete data_set_ptr;
            data_set_ptr = 0;

            if (at_eof) {
                break;
            }
        }

        if (!at_eof) {
            // get_next_data_set call failed but not because of EOF
            // -> failed to open file at all.
            std::cerr << "** " << app_name << " : Failed to open file `" 
                      << curr_qval_file_path << "'" << std::endl;
            ret_val = ERR_DATA_FILE;
            break;
        }
    }

#if 0
    table->printDataTable();
    table->error_table.create(&(table->data_table));
    table->printErrorTable();
#endif
    table->create();
    table->print();
    delete table;

    return ret_val;
}

/*****************************************************************************/
